@@ -16,6 +16,12 @@
 /* */
 #undef HAVE_SOCKLEN_T
 
+/* */
+#undef HAVE_VOLATILE
+
+/* */
+#undef HAVE_ERRNO
+
 /* */
 #undef RENAME_OVER_EXISTING_FILES
 
@@ -37,15 +43,36 @@
 /* */
 #undef HAVE_STRNCMP
 
+/* */
+#undef HAVE_STRCASECMP
+
+/* */
+#undef HAVE_STRNCASECMP
+
 /* */
 #undef HAVE_STRCSPN
 
+/* */
+#undef HAVE_STRSPN
+
 /* */
 #undef HAVE_STRSTR
 
+/* */
+#undef HAVE_MEMCMP
+
+/* */
+#undef HAVE_MEMCHR
+
+/* */
+#undef HAVE_MEMCPY
+
 /* */
 #undef HAVE_MEMMOVE
 
+/* */
+#undef HAVE_MEMSET
+
 /* */
 #undef HAVE_STRERROR
 
