@@ -61,31 +61,62 @@ void menu_func(struct window *, struct event *, int);
 void mainmenu_func(struct window *, struct event *, int);
 void dialog_func(struct window *, struct event *, int);
 
-void do_menu_selected(struct terminal *term, struct menu_item *items, void *data, int selected)
+unsigned char select_hotkey(struct terminal *term, unsigned char *text, unsigned char *hotkey, unsigned char *hotkeys, int n)
 {
+	unsigned char c;
+	if (hotkey == M_BAR) return 0;
+	if (text) {
+		text = stracpy(_(text, term));
+		charset_upcase_string(&text, term->spec->charset);
+	}
+	hotkey = _(hotkey, term);
+	while (1) {
+		int i;
+		c = *hotkey++;
+		if (!c) break;
+		c = charset_upcase(c, term->spec->charset);
+		for (i = 0; i < n; i++) if (hotkeys[i] == c) goto cont;
+		if (!text || strchr(text, c)) break;
+		cont:;
+	}
+	if (text) mem_free(text);
+	return c;
+}
+void do_menu_selected(struct terminal *term, struct menu_item *items, void *data, int selected, void (*free_function)(void *), void *free_data)
+{
+	int i;
 	struct menu *menu;
-	menu = mem_alloc(sizeof(struct menu));
+	for (i = 0; items[i].text; i++) if (i == MAXINT - sizeof(struct menu)) overalloc();
+	menu = mem_alloc(sizeof(struct menu) + (!i ? 0 : i - 1));
 	menu->selected = selected;
 	menu->view = 0;
+	menu->ni = i;
 	menu->items = items;
 	menu->data = data;
+	menu->free_function = free_function;
+	menu->free_data = free_data;
+	for (i = 0; i < menu->ni; i++)
+		menu->hotkeys[i] = select_hotkey(term, items[i].text, items[i].hotkey, menu->hotkeys, i);
 	add_window(term, menu_func, menu);
 }
 
 void do_menu(struct terminal *term, struct menu_item *items, void *data)
 {
-	do_menu_selected(term, items, data, 0);
+	do_menu_selected(term, items, data, 0, NULL, NULL);
 }
 
 void select_menu(struct terminal *term, struct menu *menu)
 {
-	/*int x = menu->x + 4;
-	int y = menu->y + menu->selected - menu->view + 2;*/
-	struct menu_item *it = &menu->items[menu->selected];
-	void (*func)(struct terminal *, void *, void *) = it->func;
-	void *data1 = it->data;
-	void *data2 = menu->data;
-	if (menu->selected < 0 || menu->selected >= menu->ni || it->hotkey == M_BAR) return;
+	struct menu_item *it;
+	void (*func)(struct terminal *, void *, void *);
+	void *data1;
+	void *data2;
+	if (menu->selected < 0 || menu->selected >= menu->ni) return;
+	it = &menu->items[menu->selected];
+	func = it->func;
+	data1 = it->data;
+	data2 = menu->data;
+	if (it->hotkey == M_BAR) return;
 	if (!it->in_m) {
 		struct window *win, *win1;
 		for (win = term->windows.next; (void *)win != &term->windows && (win->handler == menu_func || win->handler == mainmenu_func); win1 = win->next, delete_window(win), win = win1) ;
@@ -99,11 +130,10 @@ void count_menu_size(struct terminal *term, struct menu *menu)
 	int sy = term->y;
 	int mx = 4;
 	int my;
-	for (my = 0; menu->items[my].text; my++) {
+	for (my = 0; my < menu->ni; my++) {
 		int s = strlen(_(menu->items[my].text, term)) + strlen(_(menu->items[my].rtext, term)) + MENU_HOTKEY_SPACE * (_(menu->items[my].rtext, term)[0] != 0) + 4;
 		if (s > mx) mx = s;
 	}
-	menu->ni = my;
 	my += 2;
 	if (mx > sx) mx = sx;
 	if (my > sy) my = sy;
@@ -147,26 +177,26 @@ void scroll_menu(struct menu *menu, int d)
 void display_menu(struct terminal *term, struct menu *menu)
 {
 	int p, s;
-	fill_area(term, menu->x+1, menu->y+1, menu->xw-2, menu->yw-2, COLOR_MENU);
+	fill_area(term, menu->x+1, menu->y+1, menu->xw-2, menu->yw-2, COLOR_MENU_TEXT | ' ');
 	draw_frame(term, menu->x, menu->y, menu->xw, menu->yw, COLOR_MENU_FRAME, 1);
 	for (p = menu->view, s = menu->y + 1; p < menu->ni && p < menu->view + menu->yw - 2; p++, s++) {
 		int x;
 		int h = 0;
 		unsigned char c;
 		unsigned char *tmptext = _(menu->items[p].text, term);
-		int co = p == menu->selected ? h = 1, COLOR_MENU_SELECTED : COLOR_MENU;
+		int co = p == menu->selected ? h = 1, COLOR_MENU_SELECTED : COLOR_MENU_TEXT;
 		if (h) {
 			set_cursor(term, menu->x+1, s, term->x - 1, term->y - 1);
 			/*set_window_ptr(menu->win, menu->x+3, s+1);*/
 			set_window_ptr(menu->win, menu->x+menu->xw, s);
-			fill_area(term, menu->x+1, s, menu->xw-2, 1, co);
+			fill_area(term, menu->x+1, s, menu->xw-2, 1, co | ' ');
 		}
 		if (menu->items[p].hotkey != M_BAR || (tmptext && tmptext[0])) {
 			int l = strlen(_(menu->items[p].rtext, term));
 			for (x = l - 1; x >= 0 && menu->xw - 4 >= l - x && (c = _(menu->items[p].rtext, term)[x]); x--)
 				set_char(term, menu->x + menu->xw - 2 - l + x, s, c | co);
 			for (x = 0; x < menu->xw - 4 && (c = tmptext[x]); x++)
-				set_char(term, menu->x + x + 2, s, !h && strchr(_(menu->items[p].hotkey, term), upcase(c)) ? h = 1, COLOR_MENU_HOTKEY | c : co | c);
+				set_char(term, menu->x + x + 2, s, !h && charset_upcase(c, term->spec->charset) == menu->hotkeys[p] ? h = 1, COLOR_MENU_HOTKEY | c : co | c);
 		} else {
 			set_char(term, menu->x, s, COLOR_MENU_FRAME | ATTR_FRAME | 0xc3);
 			fill_area(term, menu->x+1, s, menu->xw-2, 1, COLOR_MENU_FRAME | ATTR_FRAME | 0xc4);
@@ -256,12 +286,13 @@ void menu_func(struct window *win, struct event *ev, int fwd)
 					}
 					if (ev->x > ' ' && ev->x < 256) {
 						int i;
-						for (i = 0; i < menu->ni; i++)
-							if (strchr(_(menu->items[i].hotkey, win->term), upcase(ev->x))) {
+						for (i = 0; i < menu->ni; i++) {
+							if (charset_upcase(ev->x, win->term->spec->charset) == menu->hotkeys[i]) {
 								menu->selected = i;
 								scroll_menu(menu, 0);
 								s = 1;
 							}
+						}
 					}
 					break;
 			}
@@ -275,23 +306,30 @@ void menu_func(struct window *win, struct event *ev, int fwd)
 		case EV_ABORT:
 			if (menu->items->free_i) {
 				int i;
-				for (i = 0; menu->items[i].text; i++) {
+				for (i = 0; i < menu->ni; i++) {
 					if (menu->items[i].free_i & 2) mem_free(menu->items[i].text);
 					if (menu->items[i].free_i & 4) mem_free(menu->items[i].rtext);
 				}
 				mem_free(menu->items);
 			}
+			if (menu->free_function)
+				register_bottom_half(menu->free_function, menu->free_data);
 			break;
 	}
 }
 
 void do_mainmenu(struct terminal *term, struct menu_item *items, void *data, int sel)
 {
+	int i;
 	struct mainmenu *menu;
-	menu = mem_alloc(sizeof(struct mainmenu));
+	for (i = 0; items[i].text; i++) if (i == MAXINT - sizeof(struct mainmenu)) overalloc();
+	menu = mem_alloc(sizeof(struct mainmenu) + (!i ? 0 : i - 1));
 	menu->selected = sel == -1 ? 0 : sel;
+	menu->ni = i;
 	menu->items = items;
 	menu->data = data;
+	for (i = 0; i < menu->ni; i++)
+		menu->hotkeys[i] = select_hotkey(term, NULL, items[i].hotkey, menu->hotkeys, i);
 	add_window(term, mainmenu_func, menu);
 	if (sel != -1) {
 		struct event ev = {EV_KBD, KBD_ENTER, 0, 0};
@@ -305,32 +343,33 @@ void display_mainmenu(struct terminal *term, struct mainmenu *menu)
 	int i;
 	int p = 2;
 	fill_area(term, 0, 0, term->x, 1, COLOR_MAINMENU | ' ');
-	for (i = 0; menu->items[i].text; i++) {
+	for (i = 0; i < menu->ni; i++) {
 		int s = 0;
 		int j;
 		unsigned char c;
 		unsigned char *tmptext = _(menu->items[i].text, term);
 		int co = i == menu->selected ? s = 1, COLOR_MAINMENU_SELECTED : COLOR_MAINMENU;
 		if (s) {
-			fill_area(term, p, 0, 2, 1, co);
-			fill_area(term, p+strlen(tmptext)+2, 0, 2, 1, co);
+			fill_area(term, p, 0, 2, 1, co | ' ');
+			fill_area(term, p+strlen(tmptext)+2, 0, 2, 1, co | ' ');
 			menu->sp = p;
 			set_cursor(term, p, 0, term->x - 1, term->y - 1);
 			set_window_ptr(menu->win, p, 1);
 		}
 		p += 2;
 		for (j = 0; (c = tmptext[j]); j++, p++)
-			set_char(term, p, 0, (!s && strchr(_(menu->items[i].hotkey, term), upcase(c)) ? s = 1, COLOR_MAINMENU_HOTKEY : co) | c);
+			set_char(term, p, 0, (!s && charset_upcase(c, term->spec->charset) == menu->hotkeys[i] ? s = 1, COLOR_MAINMENU_HOTKEY : co) | c);
 		p += 2;
 	}
-	menu->ni = i;
 	redraw_from_window(menu->win);
 }
 
 void select_mainmenu(struct terminal *term, struct mainmenu *menu)
 {
-	struct menu_item *it = &menu->items[menu->selected];
-	if (menu->selected < 0 || menu->selected >= menu->ni || it->hotkey == M_BAR) return;
+	struct menu_item *it;
+	if (menu->selected < 0 || menu->selected >= menu->ni) return;
+	it = &menu->items[menu->selected];
+	if (it->hotkey == M_BAR) return;
 	if (!it->in_m) {
 		struct window *win, *win1;
 		for (win = term->windows.next; (void *)win != &term->windows && (win->handler == menu_func || win->handler == mainmenu_func); win1 = win->next, delete_window(win), win = win1) ;
@@ -396,11 +435,12 @@ void mainmenu_func(struct window *win, struct event *ev, int fwd)
 			if (ev->x > ' ' && ev->x < 256) {
 				int i;
 				s = 1;
-				for (i = 0; i < menu->ni; i++)
-					if (strchr(_(menu->items[i].hotkey, win->term), upcase(ev->x))) {
+				for (i = 0; i < menu->ni; i++) {
+					if (charset_upcase(ev->x, win->term->spec->charset) == menu->hotkeys[i]) {
 						menu->selected = i;
 						s = 2;
 					}
+				}
 			} else if (!s) {
 				delete_window_ev(win, (ev->x >= KBD_F1 && ev->x <= KBD_F12) || ev->y & KBD_ALT ? ev : NULL);
 				break;
@@ -475,7 +515,7 @@ void display_dlg_item(struct dialog_data *dlg, struct dialog_item_data *di, int
 			if (di->vpos + di->l <= di->cpos) di->vpos = di->cpos - di->l + 1;
 			if (di->vpos > di->cpos) di->vpos = di->cpos;
 			if (di->vpos < 0) di->vpos = 0;
-			fill_area(term, di->x, di->y, di->l, 1, COLOR_DIALOG_FIELD);
+			fill_area(term, di->x, di->y, di->l, 1, COLOR_DIALOG_FIELD | ' ');
 			if (di->item->type == D_FIELD) {
 				print_text(term, di->x, di->y, strlen(di->cdata + di->vpos) <= (size_t)di->l ? (int)strlen(di->cdata + di->vpos) : di->l, di->cdata + di->vpos, COLOR_DIALOG_FIELD_TEXT);
 			} else {
@@ -529,9 +569,9 @@ void dlg_select_item(struct dialog_data *dlg, struct dialog_item_data *di)
 void dlg_set_history(struct dialog_item_data *di)
 {
 	unsigned char *s = "";
-	int l;
+	size_t l;
 	if ((void *)di->cur_hist != &di->history) s = di->cur_hist->d;
-	if ((l = strlen(s)) > di->item->dlen) l = di->item->dlen - 1;
+	if ((l = strlen(s)) >= (size_t)di->item->dlen) l = di->item->dlen - 1;
 	memcpy(di->cdata, s, l);
 	di->cdata[l] = 0;
 	di->cpos = l;
@@ -597,8 +637,8 @@ void tab_compl(struct terminal *term, unsigned char *item, struct window *win)
 {
 	struct event ev = {EV_REDRAW, 0, 0, 0};
 	struct dialog_item_data *di = &((struct dialog_data*)win->data)->items[((struct dialog_data*)win->data)->selected];
-	int l = strlen(item);
-	if (l >= di->item->dlen) l = di->item->dlen - 1;
+	size_t l = strlen(item);
+	if (l >= (size_t)di->item->dlen) l = di->item->dlen - 1;
 	memcpy(di->cdata, item, l);
 	di->cdata[l] = 0;
 	di->cpos = l;
@@ -637,7 +677,7 @@ void do_tab_compl(struct terminal *term, struct list_head *history, struct windo
 	}
 	if (n) {
 		memset(&items[n], 0, sizeof(struct menu_item));
-		do_menu_selected(term, items, win, n - 1);
+		do_menu_selected(term, items, win, n - 1, NULL, NULL);
 	}
 }
 
@@ -683,7 +723,6 @@ void dialog_func(struct window *win, struct event *ev, int fwd)
 				if (di->item->type == D_FIELD || di->item->type == D_FIELD_PASS) {
 					if (di->item->history) {
 						struct history_item *j;
-						/*int l = di->item->dlen;*/
 						foreach(j, di->item->history->items) {
 							struct history_item *hi;
 							hi = mem_alloc(sizeof(struct history_item) + strlen(j->d) + 1);
@@ -695,6 +734,7 @@ void dialog_func(struct window *win, struct event *ev, int fwd)
 				}
 			}
 			dlg->selected = 0;
+			/*-fallthrough*/
 		case EV_RESIZE:
 		case EV_REDRAW:
 			dlg->dlg->fn(dlg);
@@ -762,8 +802,10 @@ void dialog_func(struct window *win, struct event *ev, int fwd)
 						goto dsp_f;
 					case ACT_PASTE_CLIPBOARD: {
 						/* Paste from clipboard */
-						unsigned char * clipboard = get_clipboard_text();
+						unsigned char *clipboard = get_clipboard_text();
 						if (clipboard) {
+							unsigned char *nl = clipboard;
+							while ((nl = strchr(nl, '\n'))) *nl = ' ';
 							safe_strncpy(di->cdata, clipboard, di->item->dlen);
 							di->cpos = strlen(di->cdata);
 							mem_free(clipboard);
@@ -794,7 +836,7 @@ void dialog_func(struct window *win, struct event *ev, int fwd)
 			}
 			gh:
 			if (ev->x > ' ' && ev->x < 0x100) for (i = 0; i < dlg->n; i++)
-				if (dlg->dlg->items[i].type == D_BUTTON && upcase(_(dlg->dlg->items[i].text, term)[0]) == upcase(ev->x)) {
+				if (dlg->dlg->items[i].type == D_BUTTON && charset_upcase(_(dlg->dlg->items[i].text, term)[0], term->spec->charset) == charset_upcase(ev->x, term->spec->charset)) {
 					sel:
 					if (dlg->selected != i) {
 						display_dlg_item(dlg, &dlg->items[dlg->selected], 0);
@@ -911,7 +953,7 @@ void center_dlg(struct dialog_data *dlg)
 
 void draw_dlg(struct dialog_data *dlg)
 {
-	fill_area(dlg->win->term, dlg->x, dlg->y, dlg->xw, dlg->yw, COLOR_DIALOG);
+	fill_area(dlg->win->term, dlg->x, dlg->y, dlg->xw, dlg->yw, COLOR_DIALOG | ' ');
 }
 
 void max_text_width(struct terminal *term, unsigned char *text, int *width)
@@ -1048,7 +1090,6 @@ void dlg_format_field(struct terminal *term, struct terminal *t2, struct dialog_
 	item->x = x;
 	item->y = *y;
 	item->l = w;
-	/*if ((item->l = w) > item->item->dlen - 1) item->l = item->item->dlen - 1;*/
 	if (rw && item->l > *rw) if ((*rw = item->l) > w) *rw = w;
 	(*y)++;
 }
@@ -1219,6 +1260,7 @@ int msg_box_button(struct dialog_data *dlg, struct dialog_item_data *di)
 	return 0;
 }
 
+/* coverity[+free : arg-1] */
 void msg_box(struct terminal *term, struct memory_list *ml, unsigned char *title, int align, /*unsigned char *text, void *data, int n,*/ ...)
 {
 	struct dialog *dlg;
@@ -1361,11 +1403,12 @@ void input_field_fn(struct dialog_data *dlg)
 	dlg_format_buttons(term, term, dlg->items + 1, 2, dlg->x + DIALOG_LB, &y, w, NULL, AL_CENTER);
 }
 
+/* coverity[+free : arg-1] */
 void input_field(struct terminal *term, struct memory_list *ml, unsigned char *title, unsigned char *text, unsigned char *okbutton, unsigned char *cancelbutton, void *data, struct history *history, int l, unsigned char *def, int min, int max, int (*check)(struct dialog_data *, struct dialog_item_data *), void (*fn)(void *, unsigned char *), void (*cancelfn)(void *))
 {
 	struct dialog *dlg;
 	unsigned char *field;
-	if ((unsigned)l > MAXINT - sizeof(struct dialog) + 4 * sizeof(struct dialog_item)) overalloc();
+	if ((unsigned)l > MAXINT - sizeof(struct dialog) - 4 * sizeof(struct dialog_item)) overalloc();
 	dlg = mem_alloc(sizeof(struct dialog) + 4 * sizeof(struct dialog_item) + l);
 	memset(dlg, 0, sizeof(struct dialog) + 4 * sizeof(struct dialog_item) + l);
 	*(field = (unsigned char *)dlg + sizeof(struct dialog) + 4 * sizeof(struct dialog_item)) = 0;
@@ -1468,7 +1511,7 @@ void show_dlg_item_box(struct dialog_data *dlg, struct dialog_item_data *box_ite
 	/* FIXME: Counting here SHOULD be unnecessary */
 	n = 0;
 
-	fill_area(term, box_item_data->x, box_item_data->y, box_item_data->l, box_item_data->item->gid, COLOR_DIALOG_FIELD);
+	fill_area(term, box_item_data->x, box_item_data->y, box_item_data->l, box_item_data->item->gid, COLOR_DIALOG_FIELD | ' ');
 
 	foreach (citem, box->items) {
 		int len; /* Length of the current text field. */
