@@ -31,85 +31,63 @@ void finalize_bookmarks() {
 
 /* Loads the bookmarks from file */
 void read_bookmarks() {
-	unsigned char in_buffer[MAX_STR_LEN];	
 	unsigned char *file_name;
-	unsigned char *title, *url;	/* Pointers to the start of each field (in buffer)*/
-	FILE *f;
+	unsigned char *url;	/* Pointer to the start of url field */
+	unsigned char *str, *ptr, *nl;
 	
+	if (!links_home) return;
 	file_name = stracpy(links_home);
-	if (!file_name) return;
 	add_to_strn(&file_name, "bookmarks");
-	
-	f = fopen(file_name, "r");
+
+	str = read_config_file(file_name);
 	mem_free(file_name);
-	if (f == NULL)
-		return;
-	
-	title = in_buffer;
-	
-	while (fgets(in_buffer, MAX_STR_LEN, f)) {
-		url = strchr(in_buffer, '|');
-		if (url == NULL) 
-			continue;
-		*url = '\0';
-		url++;
-		if (strchr(url, '\n')) *( strchr(url, '\n') ) = '\0';
-		add_bookmark(title, url);
+	if (!str) return;
+
+	for (ptr = str; *ptr; ptr = nl) {
+		nl = strchr(ptr, '\n');
+		if (nl) *nl++ = 0;
+		else nl = strchr(ptr, 0);
+
+		url = strchr(ptr, '|');
+		if (!url) continue;
+		*url++ = 0;
+		add_bookmark(ptr, url);
 	}
-	
-	fclose(f);
+
+	mem_free(str);
 }
 
 /* Saves the bookmarks to file */
 void write_bookmarks() {
 	struct bookmark *bm;
-	FILE *out;
-	unsigned char *tmp_file_name, *file_name;
-
-	tmp_file_name = stracpy(links_home);
-	if (!tmp_file_name) return;
-	add_to_strn(&tmp_file_name, "bookmarks.tmp");
-	
-	out = fopen(tmp_file_name, "w");
-	if (!out) {
-		mem_free(tmp_file_name);
-		return;
-	}
-
-	foreachback(bm, bookmarks) {
-		unsigned char *p = stracpy(bm->title);
-		int i;
-		for (i = strlen(p) - 1; i >= 0; i--) if (p[i] < ' '|| p[i] == '|') p[i] = ' ';
-		fputs(p,out);
-		fputc('|', out);
-		fputs(bm->url,out);
-		fputc('\n',out);
-		mem_free(p);
-	}
-
-	if (ferror(out)) {
-		fclose(out);
-		goto err;
-	}
-	if (fclose(out) == EOF) {
-		err:
-		remove(tmp_file_name);
-		mem_free(tmp_file_name);
-		return;
-	}
+	unsigned char *file_name;
+	unsigned char *s;
+	int l;
 
+	if (!links_home) return;
 	file_name = stracpy(links_home);
-	if (!file_name) {
-		mem_free(tmp_file_name);
-		return;
-	}
 	add_to_strn(&file_name, "bookmarks");
 
-#ifndef RENAME_OVER_EXISTING_FILES
-	unlink(file_name);		/* OS/2 needs this */
-#endif
-	rename(tmp_file_name, file_name);
-	mem_free(tmp_file_name);
+	s = init_str();
+	l = 0;
+	
+	foreachback(bm, bookmarks) {
+		int i;
+		unsigned char *p = stracpy(bm->title);
+		for (i = strlen(p) - 1; i >= 0; i--) if (p[i] < ' '|| p[i] == '|') p[i] = ' ';
+		add_to_str(&s, &l, p);
+		mem_free(p);
+		add_chr_to_str(&s, &l, '|');
+		p = stracpy(bm->url);
+		for (i = strlen(p) - 1; i >= 0; i--) if (p[i] < ' ') p[i] = ' ';
+		add_to_str(&s, &l, p);
+		mem_free(p);
+		add_chr_to_str(&s, &l, '\n');
+	}
+
+	write_to_config_file(file_name, s);
+
+	mem_free(s);
 	mem_free(file_name);
 
 }
@@ -274,7 +252,7 @@ int bookmark_dlg_list_update(struct list_head *bm_list) {
 		/* Deleted in bookmark_dlg_clear_list() */
 		item = mem_alloc( sizeof(struct box_item) + strlen(bm->title) + 1);
 		item->text = text = ((unsigned char *)item + sizeof(struct box_item));
-		item->data = (void *)(id = bm->id);
+		item->data = (void *)(my_uintptr_t)(id = bm->id);
 	
 		/* Note that free_i is left at zero */
 	
@@ -313,7 +291,7 @@ bookmark_id bookmark_dlg_box_id_get(struct dlg_data_item_data_box *box) {
 		until sel equals zero, and return the id at that point */
 	foreach(citem, box->items) {
 		if (sel == 0) 
-			return (bookmark_id)(citem->data);
+			return (bookmark_id)(my_uintptr_t)(citem->data);
 		sel--;
 	}
 
@@ -344,20 +322,20 @@ int bookmark_dialog_event_handler(struct dialog_data *dlg, struct event *ev) {
 			if (ev->x == KBD_RIGHT || (ev->x == KBD_TAB && !ev->y)) {
 				/* MP: dirty crap!!! this should be done in bfu.c */
 				/* Move right */
-                display_dlg_item(dlg, &dlg->items[dlg->selected], 0);
+		display_dlg_item(dlg, &dlg->items[dlg->selected], 0);
 				if (++dlg->selected >= BM_BOX_IND) 
 					dlg->selected = 0; 
-                display_dlg_item(dlg, &dlg->items[dlg->selected], 1);
+		display_dlg_item(dlg, &dlg->items[dlg->selected], 1);
 				
 				return EVENT_PROCESSED;
 			}
 			
 			if (ev->x == KBD_LEFT || (ev->x == KBD_TAB && ev->y)) {
 				/* Move left */
-                display_dlg_item(dlg, &dlg->items[dlg->selected], 0);
+		display_dlg_item(dlg, &dlg->items[dlg->selected], 0);
 				if (--dlg->selected < 0) 
 					dlg->selected = BM_BOX_IND - 1; 
-                display_dlg_item(dlg, &dlg->items[dlg->selected], 1);
+		display_dlg_item(dlg, &dlg->items[dlg->selected], 1);
 				
 				return EVENT_PROCESSED;
 			}
@@ -400,7 +378,7 @@ int bookmark_dialog_event_handler(struct dialog_data *dlg, struct event *ev) {
 		case EV_ABORT:
 		break;
 		default:
-			internal("Unknown event received: %d", ev->ev);
+			internal("Unknown event received: %d", (int)ev->ev);
 	}
 
 	return EVENT_NOT_PROCESSED;
@@ -560,7 +538,7 @@ int push_goto_button(struct dialog_data *dlg, struct dialog_item_data *goto_btn)
 
 /* Called when an edit is complete. */
 void bookmark_edit_done(struct dialog *d) {
-	bookmark_id id = (bookmark_id)d->udata2;
+	bookmark_id id = (bookmark_id)(my_uintptr_t)d->udata2;
 	struct dialog_data *parent;
 	
 	bookmark_update(id, d->items[0].data, d->items[1].data);
@@ -585,7 +563,7 @@ int push_edit_button(struct dialog_data *dlg, struct dialog_item_data *edit_btn)
 		const unsigned char *name = bookmark_get_name(id);
 		const unsigned char *url = bookmark_get_url(id);
 
-		bookmark_edit_dialog(dlg->win->term, TEXT_(T_EDIT_BOOKMARK), name, url, (struct session*)edit_btn->item->udata, dlg, bookmark_edit_done, (void *)id);
+		bookmark_edit_dialog(dlg->win->term, TEXT_(T_EDIT_BOOKMARK), name, url, (struct session*)edit_btn->item->udata, dlg, bookmark_edit_done, (void *)(my_uintptr_t)id);
 	}
 	/* FIXME There really should be some feedback to the user here */
 	return 0;
