@@ -2,13 +2,13 @@
 
 struct list_head cache = {&cache, &cache};
 
-long cache_size;
+my_uintptr_t cache_size = 0;
 
 int cache_count = 0;
 
-long cache_info(int type)
+my_uintptr_t cache_info(int type)
 {
-	int i = 0;
+	my_uintptr_t i = 0;
 	struct cache_entry *ce;
 	switch (type) {
 		case CI_BYTES:
@@ -23,7 +23,7 @@ long cache_info(int type)
 			foreach(ce, cache) i += is_entry_used(ce);
 			return i;
 		case CI_LIST:
-			return (long) &cache;
+			return (my_uintptr_t) &cache;
 		default:
 			internal("cache_info: bad request");
 	}
@@ -83,26 +83,28 @@ int add_fragment(struct cache_entry *e, off_t offset, unsigned char *data, off_t
 {
 	struct fragment *f;
 	struct fragment *nf;
-	int a = 0;
 	int trunc = 0;
 	volatile off_t ca;
 	if (!length) return 0;
 	e->incomplete = 1;
-	if (offset + length < 0 || offset + length < offset) overalloc();
-	if (offset + (off_t)C_ALIGN(length) < 0 || offset + (off_t)C_ALIGN(length) < offset) overalloc();
+	if ((off_t)(0UL + offset + length) < 0 || (off_t)(0UL + offset + length) < offset) overalloc();
+	if ((off_t)(0UL + offset + (off_t)C_ALIGN(length)) < 0 || (off_t)(0UL + offset + (off_t)C_ALIGN(length)) < offset) overalloc();
 	if (e->length < offset + length) e->length = offset + length;
 	e->count = cache_count++;
+	if (!list_empty(e->frag)) {
+		f = e->frag.prev;
+		if (f->offset + f->length == offset) goto have_f;
+	}
 	foreach(f, e->frag) {
+have_f:
 		if (f->offset > offset) break;
 		if (f->offset <= offset && f->offset+f->length >= offset) {
 			if (offset+length > f->offset+f->length) {
 				if (memcmp(f->data+offset-f->offset, data, f->offset+f->length-offset)) trunc = 1;
-				a = 1; /* !!! FIXME */
 				if (offset-f->offset+length <= f->real_length) {
 					sf((offset+length) - (f->offset+f->length));
 					f->length = offset-f->offset+length;
-				}
-				else {
+				} else {
 					sf(-(f->offset+f->length-offset));
 					f->length = offset-f->offset;
 					f = f->next;
@@ -120,7 +122,6 @@ int add_fragment(struct cache_entry *e, off_t offset, unsigned char *data, off_t
 	ca = C_ALIGN(length);
 	if (ca > MAXINT - (int)sizeof(struct fragment) || ca < 0) overalloc();
 	nf = mem_alloc(sizeof(struct fragment) + ca);
-	a = 1;
 	sf(length);
 	nf->offset = offset;
 	nf->length = length;
@@ -152,7 +153,11 @@ int add_fragment(struct cache_entry *e, off_t offset, unsigned char *data, off_t
 		foreach(f, e->frag) fprintf(stderr, "%d, %d, %d\n", f->offset, f->length, f->real_length);
 		debug("a-");
 	}*/
-	return a;
+	if (e->length > e->max_length) {
+		e->max_length = e->length;
+		return 1;
+	}
+	return 0;
 }
 
 void defrag_entry(struct cache_entry *e)
@@ -167,8 +172,11 @@ void defrag_entry(struct cache_entry *e)
 		return;
 	}
 	if (g == f->next && f->length == f->real_length) return;
-	for (l = 0, h = f; h != g; h = h->next) l += h->length;
-	if (l > MAXINT - (int)sizeof(struct fragment) || l < 0) overalloc();
+	for (l = 0, h = f; h != g; h = h->next) {
+		if ((off_t)(0UL + l + h->length) < 0 || (off_t)(0UL + l + h->length) < l) overalloc();
+		l += h->length;
+	}
+	if (l > MAXINT - (int)sizeof(struct fragment)) overalloc();
 	n = mem_alloc(sizeof(struct fragment) + l);
 	n->offset = 0;
 	n->length = l;
@@ -257,10 +265,10 @@ void delete_entry_content(struct cache_entry *e)
 	free_list(e->frag);
 	e->length = 0;
 	e->incomplete = 1;
-	if ((cache_size -= e->data_size) < 0) {
-		internal("cache_size underflow: %ld", cache_size);
-		cache_size = 0;
+	if (cache_size < (my_uintptr_t)e->data_size) {
+		internal("cache_size underflow: %lu, %lu", (unsigned long)cache_size, (unsigned long)e->data_size);
 	}
+	cache_size -= e->data_size;
 	e->data_size = 0;
 	if (e->last_modified) {
 		mem_free(e->last_modified);
@@ -270,7 +278,7 @@ void delete_entry_content(struct cache_entry *e)
 
 void delete_cache_entry(struct cache_entry *e)
 {
-	if (e->refcount) internal("deleteing locked cache entry");
+	if (e->refcount) internal("deleting locked cache entry");
 #ifdef DEBUG
 	if (is_entry_used(e)) internal("deleting loading cache entry");
 #endif
@@ -289,35 +297,37 @@ void delete_cache_entry(struct cache_entry *e)
 void garbage_collection(int u)
 {
 	struct cache_entry *e, *f;
-	long ncs = cache_size;
-	long ccs = 0;
-	if (!u && cache_size <= memory_cache_size) return;
+	my_uintptr_t ncs = cache_size;
+	my_uintptr_t ccs = 0;
+	if (!u && cache_size <= (my_uintptr_t)memory_cache_size) return;
 	foreach(e, cache) {
-		if (e->refcount || is_entry_used(e)) if ((ncs -= e->data_size) < 0) {
-			internal("cache_size underflow: %ld", ncs);
-			ncs = 0;
+		if (e->refcount || is_entry_used(e)) {
+			if (ncs < (my_uintptr_t)e->data_size) {
+				internal("cache_size underflow: %lu, %lu", (unsigned long)ncs, (unsigned long)e->data_size);
+			}
+			ncs -= e->data_size;
 		}
 		ccs += e->data_size;
 	}
-	if (ccs != cache_size) internal("cache size badly computed: %ld != %ld", cache_size, ccs), cache_size = ccs;
-	if (!u && ncs <= memory_cache_size) return;
+	if (ccs != cache_size) internal("cache size badly computed: %lu != %lu", (unsigned long)cache_size, (unsigned long)ccs), cache_size = ccs;
+	if (!u && ncs <= (my_uintptr_t)memory_cache_size) return;
 	for (e = cache.prev; (void *)e != &cache; e = e->prev) {
-		if (!u && ncs <= (longlong)memory_cache_size * MEMORY_CACHE_GC_PERCENT) goto g;
+		if (!u && (longlong)ncs <= (longlong)memory_cache_size * MEMORY_CACHE_GC_PERCENT) goto g;
 		if (e->refcount || is_entry_used(e)) {
 			e->tgc = 0;
 			continue;
 		}
 		e->tgc = 1;
-		if ((ncs -= e->data_size) < 0) {
-			internal("cache_size underflow: %ld", ncs);
-			ncs = 0;
+		if (ncs < (my_uintptr_t)e->data_size) {
+			internal("cache_size underflow: %lu, %lu", (unsigned long)ncs, (unsigned long)e->data_size);
 		}
+		ncs -= e->data_size;
 	}
-	if (/*!no &&*/ ncs) internal("cache_size(%ld) is larger than size of all objects(%ld)", cache_size, cache_size - ncs);
+	if (ncs) internal("cache_size(%lu) is larger than size of all objects(%lu)", (unsigned long)cache_size, (unsigned long)(cache_size - ncs));
 	g:
 	if ((void *)(e = e->next) == &cache) return;
 	if (!u) for (f = e; (void *)f != &cache; f = f->next) {
-		if (ncs + f->data_size <= (longlong)memory_cache_size * MEMORY_CACHE_GC_PERCENT) {
+		if (f->data_size && (longlong)ncs + f->data_size <= (longlong)memory_cache_size * MEMORY_CACHE_GC_PERCENT) {
 			ncs += f->data_size;
 			f->tgc = 0;
 		}
@@ -326,15 +336,13 @@ void garbage_collection(int u)
 		f = f->next;
 		if (f->prev->tgc) delete_cache_entry(f->prev);
 	}
-	/*if (!no && cache_size > (longlong)memory_cache_size * MEMORY_CACHE_GC_PERCENT) {
-		internal("garbage collection doesn't work, cache size %ld", cache_size);
-	}*/
 }
 
 void init_cache(void)
 {
 #ifdef HAVE_GETPAGESIZE
-	int getpg = getpagesize();
+	int getpg;
+	EINTRLOOP(getpg, getpagesize());
 	if (getpg > 0 && getpg < 0x10000 && !(getpg & (getpg - 1))) page_size = getpg;
 #endif
 }
