@@ -16,7 +16,7 @@ struct codepage_desc {
 #include "entity.inc"
 #include "upcase.inc"
 
-char strings[256][2] = {
+unsigned char strings[256][2] = {
 	"\000", "\001", "\002", "\003", "\004", "\005", "\006", "\007",
 	"\010", "\011", "\012", "\013", "\014", "\015", "\016", "\017",
 	"\020", "\021", "\022", "\023", "\024", "\025", "\026", "\033",
@@ -58,7 +58,7 @@ void free_translation_table(struct conv_table *p)
 	mem_free(p);
 }
 
-unsigned char *no_str = NULL;
+unsigned char no_str[] = "*";
 
 void new_translation_table(struct conv_table *p)
 {
@@ -70,15 +70,15 @@ void new_translation_table(struct conv_table *p)
 
 #define BIN_SEARCH(table, entry, entries, key, result)			\
 {									\
-	int _s = 0, _e = (entries) - 1;					\
-	while (_s <= _e || !((result) = -1)) {				\
-		int _m = (_s + _e) / 2;					\
-		if ((table)[_m].entry == (key)) {			\
-			(result) = _m;					\
+	int s_ = 0, e_ = (entries) - 1;					\
+	while (s_ <= e_ || !((result) = -1)) {				\
+		int m_ = ((unsigned)s_ + (unsigned)e_) / 2;		\
+		if ((table)[m_].entry == (key)) {			\
+			(result) = m_;					\
 			break;						\
 		}							\
-		if ((table)[_m].entry > (key)) _e = _m - 1;		\
-		if ((table)[_m].entry < (key)) _s = _m + 1;		\
+		if ((table)[m_].entry > (key)) e_ = m_ - 1;		\
+		if ((table)[m_].entry < (key)) s_ = m_ + 1;		\
 	}								\
 }									\
 
@@ -94,8 +94,8 @@ static inline unsigned char *u2cp(int u, int to, int fallback)
 	int j, s;
 	again:
 	if (u < 128) return strings[u];
-	if (u == 0xa0) return "\001";
-	if (u == 0xad) return "";
+	if (u == 0xa0) return strings[1];
+	if (u == 0xad) return strings[0];
 	if (u < 0xa0) {
 		u = strange_chars[u - 0x80];
 		if (!u) return NULL;
@@ -193,6 +193,7 @@ struct conv_table *get_translation_table_to_utf_8(int from)
 	static int lfr = -1;
 	if (from == -1) return NULL;
 	if (from == lfr) return utf_table;
+	lfr = from;
 	if (utf_table_init) memset(utf_table, 0, sizeof(struct conv_table) * 256), utf_table_init = 0;
 	else free_utf_table();
 	for (i = 0; i < 128; i++) utf_table[i].u.str = strings[i];
@@ -219,7 +220,6 @@ void free_conv_table()
 	if (!utf_table_init) free_utf_table();
 	if (first) memset(table, 0, sizeof(struct conv_table) * 256), first = 0;
 	new_translation_table(table);
-	if (no_str) mem_free(no_str), no_str = NULL;
 }
 
 struct conv_table *get_translation_table(int from, int to)
@@ -227,7 +227,6 @@ struct conv_table *get_translation_table(int from, int to)
 	int i;
 	static int lfr = -1;
 	static int lto = -1;
-	if (!no_str) no_str = stracpy("*");
 	if (first) memset(table, 0, sizeof(struct conv_table) * 256), first = 0;
 	if (/*from == to ||*/ from == -1 || to == -1) return NULL;
 	if (codepages[to].table == table_utf_8) return get_translation_table_to_utf_8(from);
@@ -236,7 +235,7 @@ struct conv_table *get_translation_table(int from, int to)
 	new_translation_table(table);
 	if (codepages[from].table == table_utf_8) {
 		int j;
-		for (j = 0; codepages[to].table[j].c; j++) add_utf_8(table, codepages[to].table[j].u, codepages[to].table[j].u == 0xa0 ? "\001" : codepages[to].table[j].u == 0xad ? "" : strings[codepages[to].table[j].c]);
+		for (j = 0; codepages[to].table[j].c; j++) add_utf_8(table, codepages[to].table[j].u, codepages[to].table[j].u == 0xa0 ? strings[1] : codepages[to].table[j].u == 0xad ? strings[0] : strings[codepages[to].table[j].c]);
 		for (i = 0; unicode_7b[i].x != -1; i++) if (unicode_7b[i].x >= 0x80) add_utf_8(table, unicode_7b[i].x, unicode_7b[i].s);
 	} else for (i = 128; i < 256; i++) {
 		int j;
