@@ -7,11 +7,20 @@
 #ifdef LOG_TRANSFER
 void log_data(unsigned char *data, int len)
 {
+	static int hlaseno = 0;
 	int fd;
-	if ((fd = open(LOG_TRANSFER, O_WRONLY | O_APPEND | O_CREAT, 0622)) != -1) {
+	if (!hlaseno) {
+		printf("\n"ANSI_SET_BOLD"WARNING -- LOGGING NETWORK TRANSFERS !!!"ANSI_CLEAR_BOLD ANSI_BELL"\n");
+		fflush(stdout);
+		sleep(1);
+		hlaseno = 1;
+	}
+	EINTRLOOP(fd, open(LOG_TRANSFER, O_WRONLY | O_APPEND | O_CREAT, 0600));
+	if (fd != -1) {
+		int rw;
 		set_bin(fd);
-		write(fd, data, len);
-		close(fd);
+		EINTRLOOP(rw, write(fd, data, len));
+		EINTRLOOP(rw, close(fd));
 	}
 }
 
@@ -27,8 +36,9 @@ void exception(struct connection *c)
 
 void close_socket(int *s)
 {
+	int rs;
 	if (*s == -1) return;
-	close(*s);
+	EINTRLOOP(rs, close(*s));
 	set_handlers(*s, NULL, NULL, NULL, NULL);
 	*s = -1;
 }
@@ -37,7 +47,6 @@ void connected(struct connection *);
 
 struct conn_info {
 	void (*func)(struct connection *);
-	struct sockaddr_in sa;
 	ip addr;
 	int port;
 	int *sock;
@@ -74,26 +83,32 @@ void make_connection(struct connection *c, int port, int *sock, void (*func)(str
 int get_pasv_socket(struct connection *c, int cc, int *sock, unsigned char *port)
 {
 	int s;
+	int rs;
 	struct sockaddr_in sa;
 	struct sockaddr_in sb;
 	socklen_t len = sizeof(sa);
 	memset(&sa, 0, sizeof sa);
 	memset(&sb, 0, sizeof sb);
-	if (getsockname(cc, (struct sockaddr *)&sa, &len)) {
+	EINTRLOOP(rs, getsockname(cc, (struct sockaddr *)&sa, &len));
+	if (rs) {
 		e:
 		setcstate(c, get_error_from_errno(errno));
 		retry_connection(c);
 		return -1;
 	}
-	if ((s = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1) goto e;
+	EINTRLOOP(s, socket(PF_INET, SOCK_STREAM, IPPROTO_TCP));
+	if (s == -1) goto e;
+	set_nonblock(s);
 	*sock = s;
-	fcntl(s, F_SETFL, O_NONBLOCK);
 	memcpy(&sb, &sa, sizeof(struct sockaddr_in));
-	sb.sin_port = 0;
-	if (bind(s, (struct sockaddr *)&sb, sizeof sb)) goto e;
+	sb.sin_port = htons(0);
+	EINTRLOOP(rs, bind(s, (struct sockaddr *)&sb, sizeof sb));
+	if (rs) goto e;
 	len = sizeof(sa);
-	if (getsockname(s, (struct sockaddr *)&sa, &len)) goto e;
-	if (listen(s, 1)) goto e;
+	EINTRLOOP(rs, getsockname(s, (struct sockaddr *)&sa, &len));
+	if (rs) goto e;
+	EINTRLOOP(rs, listen(s, 1));
+	if (rs) goto e;
 	memcpy(port, &sa.sin_addr.s_addr, 4);
 	memcpy(port + 4, &sa.sin_port, 2);
 	return 0;
@@ -106,7 +121,9 @@ void ssl_want_read(struct connection *c)
 
 	set_timeout(c);
 
-	if (c->no_tsl) c->ssl->options |= SSL_OP_NO_TLSv1;
+#ifdef SSL_OP_NO_TLSv1
+	if (c->no_tls) SSL_set_options(c->ssl, SSL_OP_NO_TLSv1);
+#endif
 	switch (SSL_get_error(c->ssl, SSL_connect(c->ssl))) {
 		case SSL_ERROR_NONE:
 			c->newconn = NULL;
@@ -120,7 +137,7 @@ void ssl_want_read(struct connection *c)
 			set_handlers(*b->sock, NULL, (void(*)(void *))ssl_want_read, (void(*)(void *))exception, c);
 			break;
 		default:
-			c->no_tsl++;
+			c->no_tls++;
 			setcstate(c, S_SSL_ERROR);
 			retry_connection(c);
 			break;
@@ -131,24 +148,28 @@ void ssl_want_read(struct connection *c)
 void dns_found(struct connection *c, int state)
 {
 	int s;
+	int rs;
 	struct conn_info *b = c->newconn;
+	struct sockaddr_in sa;
 	if (state) {
 		setcstate(c, S_NO_DNS);
 		abort_connection(c);
 		return;
 	}
-	if ((s = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1) {
+	EINTRLOOP(s, socket(PF_INET, SOCK_STREAM, IPPROTO_TCP));
+	if (s == -1) {
 		setcstate(c, get_error_from_errno(errno));
 		retry_connection(c);
 		return;
 	}
+	set_nonblock(s);
 	*b->sock = s;
-	fcntl(s, F_SETFL, O_NONBLOCK);
-	memset(&b->sa, 0, sizeof(struct sockaddr_in));
-	b->sa.sin_family = AF_INET;
-	b->sa.sin_addr.s_addr = b->addr;
-	b->sa.sin_port = htons(b->port);
-	if (connect(s, (struct sockaddr *)&b->sa, sizeof b->sa)) {
+	memset(&sa, 0, sizeof(struct sockaddr_in));
+	sa.sin_family = AF_INET;
+	sa.sin_addr.s_addr = b->addr;
+	sa.sin_port = htons(b->port);
+	EINTRLOOP(rs, connect(s, (struct sockaddr *)&sa, sizeof sa));
+	if (rs) {
 		if (errno != EALREADY && errno != EINPROGRESS) {
 #ifdef BEOS
 			if (errno == EWOULDBLOCK) errno = ETIMEDOUT;
@@ -169,24 +190,40 @@ void connected(struct connection *c)
 	struct conn_info *b = c->newconn;
 	int err = 0;
 	socklen_t len = sizeof(int);
-	if (getsockopt(*b->sock, SOL_SOCKET, SO_ERROR, (void *)&err, &len))
+	int rs;
+#ifdef SO_ERROR
+	errno = 0;
+	EINTRLOOP(rs, getsockopt(*b->sock, SOL_SOCKET, SO_ERROR, (void *)&err, &len));
+	if (!rs) {
+		if (err >= 10000) err -= 10000;	/* Why does EMX return so large values? */
+	} else {
 		if (!(err = errno)) {
-			err = -(S_STATE);
-			goto bla;
+			setcstate(c, S_STATE);
+			retry_connection(c);
+			return;
 		}
-	if (err >= 10000) err -= 10000;	/* Why does EMX return so large values? */
-	if (err > 0) {
-		bla:
+	}
+	if (err > 0
+#ifdef EISCONN
+	    && err != EISCONN
+#endif
+	    ) {
+#ifdef DOS
+		if (err == EALREADY) err = ETIMEDOUT;
+#endif
 		setcstate(c, get_error_from_errno(err));
 		retry_connection(c);
 		return;
 	}
+#endif
 	set_timeout(c);
 #ifdef HAVE_SSL
 	if (c->ssl) {
 		c->ssl = getSSL();
 		SSL_set_fd(c->ssl, *b->sock);
-		if (c->no_tsl) c->ssl->options |= SSL_OP_NO_TLSv1;
+#ifdef SSL_OP_NO_TLSv1
+		if (c->no_tls) SSL_set_options(c->ssl, SSL_OP_NO_TLSv1);
+#endif
 		switch (SSL_get_error(c->ssl, SSL_connect(c->ssl))) {
 			case SSL_ERROR_WANT_READ:
 				setcstate(c, S_SSL_NEG);
@@ -199,7 +236,7 @@ void connected(struct connection *c)
 			case SSL_ERROR_NONE:
 				break;
 			default:
-				c->no_tsl++;
+				c->no_tls++;
 				setcstate(c, S_SSL_ERROR);
 				retry_connection(c);
 				return;
@@ -248,10 +285,12 @@ void write_select(struct connection *c)
 		}
 	} else
 #endif
-		if ((wr = write(wb->sock, wb->data + wb->pos, wb->len - wb->pos)) <= 0) {
-#ifdef ATHEOS
+	{
+		EINTRLOOP(wr, write(wb->sock, wb->data + wb->pos, wb->len - wb->pos));
+		if (wr <= 0) {
+#if defined(ATHEOS) || defined(DOS)
 	/* Workaround for a bug in Syllable */
-			if (wr && errno == EAGAIN) {
+			if (wr && (errno == EAGAIN || errno == EWOULDBLOCK)) {
 				return;
 			}
 #endif
@@ -259,8 +298,8 @@ void write_select(struct connection *c)
 			retry_connection(c);
 			return;
 		}
+	}
 
-	/*printf("wr: %d\n", wr);*/
 	if ((wb->pos += wr) == wb->len) {
 		void (*f)(struct connection *) = wb->done;
 		c->buffer = NULL;
@@ -324,7 +363,9 @@ void read_select(struct connection *c)
 		}
 	} else
 #endif
-		if ((rd = read(rb->sock, rb->data + rb->len, READ_SIZE)) <= 0) {
+	{
+		EINTRLOOP(rd, read(rb->sock, rb->data + rb->len, READ_SIZE));
+		if (rd <= 0) {
 			if (rb->close && !rd) {
 				rb->close = 2;
 				rb->done(c, rb);
@@ -335,6 +376,7 @@ void read_select(struct connection *c)
 			retry_connection(c);
 			return;
 		}
+	}
 	log_data(rb->data + rb->len, rd);
 	rb->len += rd;
 	rb->done(c, rb);
