@@ -32,7 +32,7 @@ struct list_head c_domains = { &c_domains, &c_domains };
 struct c_server {
 	struct c_server *next;
 	struct c_server *prev;
-	int accept;
+	int accpt;
 	unsigned char server[1];
 };
 
@@ -132,7 +132,7 @@ int set_cookie(struct terminal *term, unsigned char *url, unsigned char *str)
 	}
 	cookie->id = cookie_id++;
 	foreach (cs, c_servers) if (!strcasecmp(cs->server, server)) {
-		if (cs->accept) goto ok;
+		if (cs->accpt) goto ok;
 		else {
 			free_cookie(cookie);
 			mem_free(cookie);
@@ -193,7 +193,7 @@ void delete_cookie(struct cookie *c)
 
 struct cookie *find_cookie_id(void *idp)
 {
-	int id = (int)idp;
+	int id = (int)(my_uintptr_t)idp;
 	struct cookie *c;
 	foreach(c, cookies) if (c->id == id) return c;
 	return NULL;
@@ -216,7 +216,7 @@ void cookie_default(void *idp, int a)
 	strcpy(s->server, c->server);
 	add_to_list(c_servers, s);
 	found:
-	s->accept = a;
+	s->accpt = a;
 }
 
 void accept_cookie_always(void *idp)
@@ -252,7 +252,9 @@ int is_path_prefix(unsigned char *d, unsigned char *s)
 
 int cookie_expired(struct cookie *c)	/* parse_http_date is broken */
 {
-  	return 0 && (c->expires && c->expires < time(NULL));
+	time_t t;
+	EINTRLOOPX(t, time(NULL), (time_t)-1);
+  	return 0 && (c->expires && c->expires < t);
 }
 
 void send_cookies(unsigned char **s, int *l, unsigned char *url)
