@@ -1,30 +1,53 @@
 #include "links.h"
 
+unsigned char system_name[MAX_STR_LEN];
+
 void get_system_name()
 {
-#if defined(HAVE_SYS_UTSNAME_H) && defined(HAVE_UNAME)
-	struct utsname name;
-	memset(&name, 0, sizeof name);
-	if (!uname(&name)) {
-		unsigned char *str = init_str();
-		int l = 0;
-		add_to_str(&str, &l, name.sysname);
-		add_to_str(&str, &l, " ");
-		add_to_str(&str, &l, name.release);
-		add_to_str(&str, &l, " ");
-		add_to_str(&str, &l, name.machine);
-		if (l >= MAX_STR_LEN) str[MAX_STR_LEN - 1] = 0;
-		strcpy(system_name, str);
-		mem_free(str);
+#ifdef OS2
+	if (!os_get_system_name(system_name))
 		return;
+#endif
+#if defined(HAVE_SYS_UTSNAME_H) && defined(HAVE_UNAME)
+	{
+		struct utsname name;
+		int rs;
+		memset(&name, 0, sizeof name);
+		EINTRLOOP(rs, uname(&name));
+		if (!rs) {
+#ifdef OPENVMS
+			unsigned char * volatile p;
+#endif
+			unsigned char *str = init_str();
+			int l = 0;
+			add_to_str(&str, &l, name.sysname);
+			add_to_str(&str, &l, " ");
+#ifdef OPENVMS
+			add_to_str(&str, &l, name.version);
+#else
+			add_to_str(&str, &l, name.release);
+#endif
+			add_to_str(&str, &l, " ");
+#ifdef OPENVMS
+			p = name.nodename + sizeof(name.nodename);
+			if ((unsigned char *)(&name + 1) - p >= 16 && memchr(p, 0, 16))
+				add_to_str(&str, &l, cast_uchar p);
+			else
+#endif
+			add_to_str(&str, &l, name.machine);
+			safe_strncpy(system_name, str, MAX_STR_LEN);
+			mem_free(str);
+			return;
+		}
 	}
 #endif
 #ifdef HAVE_POPEN
-	{
+	if (0) {
 		FILE *f;
 		unsigned char *p;
 		memset(system_name, 0, MAX_STR_LEN);
-		if (!(f = popen("uname -srm", "r"))) goto fail;
+		ENULLLOOP(f, popen("uname -srm", "r"));
+		if (!f) goto fail;
 		if (fread(system_name, 1, MAX_STR_LEN - 1, f) <= 0) {
 			pclose(f);
 			goto fail;
@@ -41,6 +64,328 @@ void get_system_name()
 	strcpy(system_name, SYSTEM_NAME);
 }
 
+unsigned char compiler_name[MAX_STR_LEN];
+
+static void get_compiler_name(void)
+{
+#if defined(__BORLANDC__)
+
+	int w = __BORLANDC__+0;
+	int v1 = w / 0x100;
+	int v2 = w / 0x10 % 0x10;
+	int v3 = w % 0x10;
+	if (v1 == 4 && v2 < 5) v1 = 3;
+	if (v1 == 4 && v2 == 5) v2 = 0;
+
+	if (!v3) sprintf(cast_char compiler_name, "Borland C %d.%d", v1, v2);
+	else sprintf(cast_char compiler_name, "Borland C %d.%d.%d", v1, v2, v3);
+
+#elif defined(__clang__)
+
+#if !defined(__clang_major__) || !defined(__clang_minor__)
+	sprintf(cast_char compiler_name, "LLVM/Clang");
+#else
+	int v1 = __clang_major__+0;
+	int v2 = __clang_minor__+0;
+#ifdef __clang_patchlevel__
+	int v3 = __clang_patchlevel__+0;
+#else
+	int v3 = 0;
+#endif
+	if (v3 > 0) sprintf(cast_char compiler_name, "LLVM/Clang %d.%d.%d", v1, v2, v3);
+	else sprintf(cast_char compiler_name, "LLVM/Clang %d.%d", v1, v2);
+#endif
+
+#elif defined(__COMO_VERSION__)
+
+	int w = __COMO_VERSION__+0;
+	int v1 = w / 100;
+	int v2 = w % 100;
+	if (!(v2 % 10)) sprintf(cast_char compiler_name, "Comeau C %d.%d", v1, v2 / 10);
+	else sprintf(cast_char compiler_name, "Comeau C %d.%02d", v1, v2);
+
+#elif defined(__convexc__)
+
+	sprintf(cast_char compiler_name, "Convex C");
+
+#elif defined(_CRAYC)
+
+#if !defined(_RELEASE) || !defined(_RELEASE_MINOR)
+	sprintf(cast_char compiler_name, "Cray C");
+#else
+	int v1 = _RELEASE+0;
+	int v2 = _RELEASE_MINOR+0;
+	sprintf(cast_char compiler_name, "Cray C %d.%d", v1, v2);
+#endif
+
+#elif defined(__DCC__)
+
+#ifndef __VERSION_NUMBER__
+	sprintf(cast_char compiler_name, "Diab C");
+#else
+	int w = __VERSION_NUMBER__+0;
+	int v1 = w / 1000;
+	int v2 = w / 100 % 10;
+	int v3 = w % 100;
+	sprintf(cast_char compiler_name, "Diab C %d.%d.%02d", v1, v2, v3);
+#endif
+
+#elif defined(__DMC__)
+
+	int w = __DMC__+0;
+	int v1 = w / 0x100;
+	int v2 = w / 0x10 % 0x10;
+	int v3 = w % 0x10;
+	if (!v3) sprintf(cast_char compiler_name, "Digital Mars C %d.%d", v1, v2);
+	else sprintf(cast_char compiler_name, "Digital Mars C %d.%d.%d", v1, v2, v3);
+
+#elif defined(__DECC_VER)
+
+	int w = __DECC_VER+0;
+	int v1 = w / 10000000;
+	int v2 = w / 100000 % 100;
+	int v3 = w % 10000;
+	sprintf(cast_char compiler_name, "DEC C %d.%d-%03d", v1, v2, v3);
+
+#elif defined(__ghs__)
+
+#ifndef __GHS_VERSION_NUMBER__
+	sprintf(cast_char compiler_name, "Green Hill C");
+#else
+	int w = __GHS_VERSION_NUMBER__+0;
+	int v1 = w / 100;
+	int v2 = w / 10 % 10;
+	int v3 = w % 10;
+	sprintf(cast_char compiler_name, "Green Hill C %d.%d.%d", v1, v2, v3);
+#endif
+
+#elif defined(__HIGHC__)
+
+	sprintf(cast_char compiler_name, "MetaWare High C");
+
+#elif defined(__HP_cc)
+
+	int w = __HP_cc+0;
+	int v1 = w / 10000;
+	int v2 = w / 100 % 100;
+	int v3 = w % 100;
+	if (w <= 1) sprintf(cast_char compiler_name, "HP CC");
+	else sprintf(cast_char compiler_name, "HP CC %d.%02d.%02d", v1, v2, v3);
+
+#elif defined(__xlc__)
+
+	int w = __xlc__+0;
+	int v1 = w / 0x100;
+	int v2 = w % 0x100;
+	sprintf(cast_char compiler_name, "IBM XL C %X.%X", v1, v2);
+
+#elif defined(__IBMC__) && defined(__COMPILER_VER__)
+
+	unsigned w = __COMPILER_VER__+0;
+	int v0 = w / 0x10000000;
+	int v1 = w / 0x1000000 % 0x10;
+	int v2 = w / 0x10000 % 0x100;
+	int v3 = w % 0x10000;
+	unsigned char *os = !v0 ? "S/370" : v0 == 1 ? "OS/390" : v0 == 4 ? "z/OS" : "";
+	sprintf(cast_char compiler_name, "IBM%s%s XL C %X.%0X.%X", *os ? " " : "", os, v1, v2, v3);
+
+#elif defined(__ICC)
+
+	int w = __ICC+0;
+	int v1 = w / 100;
+	int v2 = w % 100;
+	if (!(v2 % 10)) sprintf(cast_char compiler_name, "Intel C %d.%d", v1, v2 / 10);
+	else sprintf(cast_char compiler_name, "Intel C %d.%02d", v1, v2);
+
+#elif defined(__LCC__)
+
+	sprintf(cast_char compiler_name, "LCC");
+
+#elif defined(__NDPC__)
+
+	sprintf(cast_char compiler_name, "Microway NDP C");
+
+#elif defined(_MSC_VER)
+
+	int w = _MSC_VER+0;
+	int v1 = w / 100;
+	int v2 = w % 100;
+	unsigned char *visual = cast_uchar "";
+	if (v1 >= 8) {
+		v1 -= 6;
+		if (v1 == 2) v1 = 1;
+		visual = cast_uchar "Visual ";
+	}
+	if (!(v2 % 10)) sprintf(cast_char compiler_name, "Microsoft %sC %d.%d", visual, v1, v2 / 10);
+	else sprintf(cast_char compiler_name, "Microsoft %sC %d.%02d", visual, v1, v2);
+
+#elif defined(__MWERKS__)
+
+	int w = __MWERKS__+0;
+	int v1 = w / 0x1000;
+	int v2 = w / 0x100 % 0x10;
+	int v3 = w % 0x100;
+	if (w <= 1) sprintf(cast_char compiler_name, "Metrowerks CodeWarrior");
+	sprintf(cast_char compiler_name, "Metrowerks CodeWarrior %x.%x.%x", v1, v2, v3);
+
+#elif defined(__NWCC__)
+
+	sprintf(cast_char compiler_name, "NWCC");
+
+#elif defined(__OPEN64__)
+
+	unsigned char *n = cast_uchar "Open64 " __OPEN64__;
+	if (strlen(cast_const_char n) >= sizeof(cast_char compiler_name)) n = cast_uchar "Open64";
+	strcpy(cast_char compiler_name, cast_const_char n);
+
+#elif defined(__PATHSCALE__)
+
+	unsigned char *n = cast_uchar "PathScale " __PATHSCALE__;
+	if (strlen(cast_const_char n) >= sizeof(cast_char compiler_name)) n = cast_uchar "PathScale";
+	strcpy(cast_char compiler_name, cast_const_char n);
+
+#elif defined(__PCC__)
+
+	int v1 = __PCC__+0;
+#ifdef __PCC_MINOR__
+	int v2 = __PCC_MINOR__+0;
+#else
+	int v2 = 0;
+#endif
+#ifdef __PCC_MINORMINOR__
+	int v3 = __PCC_MINORMINOR__+0;
+#else
+	int v3 = 0;
+#endif
+	sprintf(cast_char compiler_name, "PCC %d.%d.%d", v1, v2, v3);
+
+#elif defined(__PGI) || defined(__PGIC__)
+
+#if !defined(__PGIC__) || !defined(__PGIC_MINOR__)
+	sprintf(cast_char compiler_name, "The Portland Group C");
+#else
+	int v1 = __PGIC__+0;
+	int v2 = __PGIC_MINOR__+0;
+#ifdef __PGIC_PATCHLEVEL__
+	int v3 = __PGIC_PATCHLEVEL__+0;
+#else
+	int v3 = 0;
+#endif
+	if (v3 > 0) sprintf(cast_char compiler_name, "The Portland Group C %d.%d.%d", v1, v2, v3);
+	else sprintf(cast_char compiler_name, "The Portland Group C %d.%d", v1, v2);
+#endif
+
+#elif defined(__SASC__)
+
+	int w = __SASC__+0;
+	int v1 = w / 100;
+	int v2 = w % 100;
+	sprintf(cast_char compiler_name, "SAS C %d.%02d", v1, v2);
+
+#elif (defined(__sgi) && defined(_COMPILER_VERSION)) || defined(_SGI_COMPILER_VERSION)
+
+#ifdef _SGI_COMPILER_VERSION
+	int w = _SGI_COMPILER_VERSION;
+#else
+	int w = _COMPILER_VERSION;
+#endif
+	int v1 = w / 100;
+	int v2 = w / 10 % 10;
+	int v3 = w % 10;
+	sprintf(cast_char compiler_name, "MIPSpro %d.%d.%d", v1, v2, v3);
+
+#elif defined(__SUNPRO_C)
+
+	int w = __SUNPRO_C+0;
+	int div = w >= 0x1000 ? 0x1000 : 0x100;
+	int v2_digits = w >= 0x1000 ? 2 : 1;
+	int v1 = w / div;
+	int v2 = w % div / 0x10;
+	int v3 = w % 0x10;
+	if (!v3) sprintf(cast_char compiler_name, "Sun C %X.%0*X", v1, v2_digits, v2);
+	else sprintf(cast_char compiler_name, "Sun C %X.%0*X.%X", v1, v2_digits, v2, v3);
+
+#elif defined(__SYSC__) && defined(__SYSC_VER__)
+
+	int w = __SYSC_VER__+0;
+	int v1 = w / 10000;
+	int v2 = w / 100 % 100;
+	int v3 = w % 100;
+	sprintf(cast_char compiler_name, "Dignus Systems C %d.%02d.%02d", v1, v2, v3);
+
+#elif defined(__TenDRA__)
+
+	sprintf(cast_char compiler_name, "TenDRA C");
+
+#elif defined(__TINYC__)
+
+	sprintf(cast_char compiler_name, "Tiny C");
+
+#elif defined(_UCC)
+
+#if !defined(_MAJOR_REV) || !defined(_MINOR_REV)
+	sprintf(cast_char compiler_name, "Ultimate C");
+#else
+	int v1 = _MAJOR_REV+0;
+	int v2 = _MAJOR_REV+0;
+	sprintf(cast_char compiler_name, "Ultimate C %d.%d", v1, v2);
+#endif
+
+#elif defined(__USLC__)
+
+	sprintf(cast_char compiler_name, "USL C");
+
+#elif defined(__VAXC)
+
+	sprintf(cast_char compiler_name, "VAX C");
+
+#elif defined(__VOSC__)
+
+	sprintf(cast_char compiler_name, "Stratus VOS C");
+
+#elif defined(__WATCOMC__)
+
+	int w = __WATCOMC__+0;
+	int v1 = w / 100;
+	int v2 = w % 100;
+	unsigned char *op = cast_uchar "";
+	if (v1 >= 12) {
+		v1 -= 11;
+		op = cast_uchar "Open";
+	}
+	if (!(v2 % 10)) sprintf(cast_char compiler_name, "%sWatcom C %d.%d", op, v1, v2 / 10);
+	else sprintf(cast_char compiler_name, "%sWatcom C %d.%02d", op, v1, v2);
+
+#elif defined(__GNUC__)
+
+	int v1 = __GNUC__+0;
+#ifdef __GNUC_MINOR__
+	int v2 = __GNUC_MINOR__+0;
+#else
+	int v2 = -1;
+#endif
+#ifdef __GNUC_PATCHLEVEL__
+	int v3 = __GNUC_PATCHLEVEL__+0;
+#else
+	int v3 = 0;
+#endif
+#if defined(__llvm__)
+	unsigned char *prefix = cast_uchar "LLVM/";
+#else
+	unsigned char *prefix = cast_uchar "";
+#endif
+	if (v1 == 2 && (v2 >= 90 && v2 <= 91)) sprintf(cast_char compiler_name, "%sEGCS 1.%d", prefix, v2 - 90);
+	else if (v3 > 0 && v2 >= 0) sprintf(cast_char compiler_name, "%sGNU C %d.%d.%d", prefix, v1, v2, v3);
+	else if (v2 >= 0) sprintf(cast_char compiler_name, "%sGNU C %d.%d", prefix, v1, v2);
+	else sprintf(cast_char compiler_name, "%sGNU C %d", prefix, v1);
+
+#else
+
+	strcpy(cast_char compiler_name, "unknown compiler");
+
+#endif
+}
+
 extern struct option links_options[];
 extern struct option html_options[];
 
@@ -49,8 +394,14 @@ struct option *all_options[] = { links_options, html_options, NULL, };
 unsigned char *p_arse_options(int argc, unsigned char *argv[], struct option **opt)
 {
 	unsigned char *e, *u = NULL;
+	int i;
+	for (i = 0; i < argc; i++) {
+		if (strlen(argv[i]) >= MAXINT) {
+			fprintf(stderr, "Too long parameter\n");
+			return NULL;
+		}
+	}
 	while (argc) {
-		int i;
 		argv++, argc--;
 		if (argv[-1][0] == '-') {
 			struct option *options;
@@ -179,63 +530,106 @@ unsigned char *read_config_file(unsigned char *name)
 	int h, r;
 	int l = 0;
 	unsigned char *s;
-	if ((h = open(name, O_RDONLY | O_NOCTTY)) == -1) return NULL;
+	int rs;
+	EINTRLOOP(h, open(name, O_RDONLY | O_NOCTTY));
+	if (h == -1) return NULL;
 	set_bin(h);
 	s = init_str();
-	while ((r = read(h, cfg_buffer, FILE_BUF)) > 0) {
+	while ((r = hard_read(h, cfg_buffer, FILE_BUF)) > 0) {
 		int i;
 		for (i = 0; i < r; i++) if (!cfg_buffer[i]) cfg_buffer[i] = ' ';
 		add_bytes_to_str(&s, &l, cfg_buffer, r);
 	}
 	if (r == -1) mem_free(s), s = NULL;
-	close(h);
+	EINTRLOOP(rs, close(h));
 	return s;
 }
 
 int write_to_config_file(unsigned char *name, unsigned char *c)
 {
 	int rr;
-	int r;
 	int h, w;
-	unsigned char *tmp_name = stracpy(name);
-	unsigned char *ds, *dt, *dot;
-	for (dt = ds = tmp_name; *dt; dt++) if (dir_sep(*dt)) ds = dt;
-	if ((dot = strchr(ds, '.'))) *dot = 0;
-	add_to_strn(&tmp_name, ".tmp");
-	if ((h = open(tmp_name, O_WRONLY | O_NOCTTY | O_CREAT | O_TRUNC, 0666)) == -1) {
+	int count = 0;
+	int tmp_namel;
+	unsigned char *tmp_name;
+	int rs, err;
+try_new_count:
+	tmp_namel = 0;
+	tmp_name = init_str();
+	add_to_str(&tmp_name, &tmp_namel, name);
+	for (w = tmp_namel - 1; w >= 0; w--) {
+		if (dir_sep(tmp_name[w]))
+			break;
+		if (tmp_name[w] == '.') {
+			if (w <= tmp_namel - 2) {
+				tmp_name[w + 2] = 0;
+				tmp_namel = w + 2;
+			}
+			break;
+		}
+	}
+	add_num_to_str(&tmp_name, &tmp_namel, count);
+	EINTRLOOP(h, open(tmp_name, O_WRONLY | O_NOCTTY | O_CREAT | O_TRUNC | O_EXCL, 0600));
+	if (h == -1) {
+		if (errno == EEXIST && count < MAXINT) {
+			count++;
+			mem_free(tmp_name);
+			goto try_new_count;
+		}
 		mem_free(tmp_name);
 		return get_error_from_errno(errno);
 	}
 	set_bin(h);
 	rr = strlen(c);
-	r = rr;
-	while (r > 0) {
-		if ((w = write(h, c + rr - r, r)) <= 0) {
-			int err = !w ? ENOSPC : errno;
-			close(h);
-			unlink(tmp_name);
-			mem_free(tmp_name);
-			return get_error_from_errno(err);
-		}
-		r -= w;
+	if (hard_write(h, c, rr) != rr) {
+		err = errno;
+		EINTRLOOP(rs, close(h));
+		goto unlink_err;
 	}
-	close(h);
-#ifndef RENAME_OVER_EXISTING_FILES
-	unlink(name);
+	EINTRLOOP(rs, close(h));
+	if (rs) {
+		err = errno;
+		goto unlink_err;
+	}
+#if defined(OPENVMS)
+	/* delete all versions of the file */
+	count = 0;
+	do {
+		EINTRLOOP(rs, unlink(name));
+	} while (!rs && ++count < 65536);
+#elif !defined(RENAME_OVER_EXISTING_FILES)
+	EINTRLOOP(rs, unlink(name));
 #endif
-	if (rename(tmp_name, name)) {
-		int err = errno;
-		unlink(tmp_name);
-		mem_free(tmp_name);
-		return get_error_from_errno(err);
+	EINTRLOOP(rs, rename(tmp_name, name));
+	if (rs) {
+		err = errno;
+		goto unlink_err;
 	}
 	mem_free(tmp_name);
 	return 0;
+
+	unlink_err:
+	EINTRLOOP(rs, unlink(cast_const_char tmp_name));
+	mem_free(tmp_name);
+	return get_error_from_errno(err);
 }
 
+#ifdef OPENVMS
+void translate_vms_to_unix(unsigned char **str)
+{
+	unsigned char *n;
+	if (!*str || strchr(cast_const_char *str, '/')) return;
+	n = cast_uchar decc$translate_vms(cast_const_char *str);
+	if (!n || (int)n == -1) return;
+	mem_free(*str);
+	*str = stracpy(n);
+}
+#endif
+
 unsigned char *get_home(int *n)
 {
 	struct stat st;
+	int rs;
 	unsigned char *home = NULL;
 	unsigned char *home_links;
 	unsigned char *config_dir = stracpy(getenv("CONFIG_DIR"));
@@ -265,9 +659,12 @@ unsigned char *get_home(int *n)
 skip_path_conv:;
 		}
 #endif
-		if (home && (stat(home, &st) == -1 || !S_ISDIR(st.st_mode))) {
-			mem_free(home);
-			home = NULL;
+		if (home) {
+			EINTRLOOP(rs, stat(home, &st));
+			if (rs || !S_ISDIR(st.st_mode)) {
+				mem_free(home);
+				home = NULL;
+			}
 		}
 	}
 #endif
@@ -279,6 +676,11 @@ skip_path_conv:;
 		mem_free(home);
 		home = NULL;
 	}
+#endif
+#ifdef OPENVMS
+	if (!home) home = stracpy(cast_uchar "/SYS$LOGIN");
+	translate_vms_to_unix(&home);
+	translate_vms_to_unix(&config_dir);
 #endif
 	if (!home) {
 		int i;
@@ -294,14 +696,14 @@ skip_path_conv:;
 		home[0] = 0;
 		br:;
 	}
-	while (home[0] && dir_sep(home[strlen(home) - 1])) home[strlen(home) - 1] = 0;
+	while (home[0] && home[1] && dir_sep(home[strlen(home) - 1])) home[strlen(home) - 1] = 0;
 	if (home[0]) add_to_strn(&home, "/");
 	home_links = stracpy(home);
-	if (config_dir) 	
-	{
+	if (config_dir) {
 		add_to_strn(&home_links, config_dir);
 		while (home_links[0] && dir_sep(home_links[strlen(home_links) - 1])) home_links[strlen(home_links) - 1] = 0;
-		if (stat(home_links, &st) != -1 && S_ISDIR(st.st_mode)) {
+		EINTRLOOP(rs, stat(home_links, &st));
+		if (!rs && S_ISDIR(st.st_mode)) {
 			add_to_strn(&home_links, "/links");
 	    	} else {
 			fprintf(stderr, "CONFIG_DIR set to %s. But directory %s doesn't exist.\n\007", config_dir, home_links);
@@ -310,11 +712,15 @@ skip_path_conv:;
 			home_links = stracpy(home);
 			add_to_strn(&home_links, ".links");		
 		}
-		mem_free(config_dir);
 	} else add_to_strn(&home_links, ".links");
-	if (stat(home_links, &st)) {
+	EINTRLOOP(rs, stat(home_links, &st));
+	if (rs) {
 #ifdef HAVE_MKDIR
-		if (!mkdir(home_links, 0777)) goto home_creat;
+		EINTRLOOP(rs, mkdir(home_links, 0700));
+		if (!rs) goto home_creat;
+#ifdef OPENVMS
+		if (errno == EEXIST) goto home_ok;
+#endif
 #endif
 		if (config_dir) goto failed;
 		goto first_failed;
@@ -323,14 +729,26 @@ skip_path_conv:;
 	/* This is a Cygwin hack! Cygwin reports stat for "links" if no
 	   "links" exists and only "links.exe" does. So try to create directory
 	   anyway. */
-	if (!mkdir(home_links, 0777)) goto home_creat;
+#ifdef HAVE_MKDIR
+	EINTRLOOP(rs, mkdir(home_links, 0700));
+	if (!rs) goto home_creat;
+#endif
 	first_failed:
 	mem_free(home_links);
 	home_links = stracpy(home);
-	add_to_strn(&home_links, "links");
-	if (stat(home_links, &st)) {
+#ifdef DOS
+	add_to_strn(&home_links, cast_uchar "links.cfg");
+#else
+	add_to_strn(&home_links, cast_uchar "links");
+#endif
+	EINTRLOOP(rs, stat(home_links, &st));
+	if (rs) {
 #ifdef HAVE_MKDIR
-		if (!mkdir(home_links, 0777)) goto home_creat;
+		EINTRLOOP(rs, mkdir(home_links, 0700));
+		if (!rs) goto home_creat;
+#ifdef OPENVMS
+		if (errno == EEXIST) goto home_ok;
+#endif
 #else
 		mem_free(home_links);
 		home_links = stracpy(home);
@@ -339,26 +757,32 @@ skip_path_conv:;
 		goto failed;
 	}
 	if (S_ISDIR(st.st_mode)) goto home_ok;
-	if (!mkdir(home_links, 0777)) goto home_creat;
+#ifdef HAVE_MKDIR
+	EINTRLOOP(rs, mkdir(home_links, 0700));
+	if (!rs) goto home_creat;
+#endif
 	failed:
 	mem_free(home_links);
 	mem_free(home);
+	if (config_dir) mem_free(config_dir);
 	return NULL;
 
 	home_ok:
 	if (n) *n = 0;
 	home_creat:
 #ifdef HAVE_CHMOD
-	chmod(home_links, 0700);
+	EINTRLOOP(rs, chmod(home_links, 0700));
 #endif
 	add_to_strn(&home_links, "/");
 	mem_free(home);
+	if (config_dir) mem_free(config_dir);
 	return home_links;
 }
 
 void init_home()
 {
 	get_system_name();
+	get_compiler_name();
 	links_home = get_home(&first_use);
 	if (!links_home) {
 		fprintf(stderr, "Unable to find or create links config directory. Please check, that you have $HOME variable set correctly and that you have write permission to your home directory.\n\007");
@@ -478,7 +902,7 @@ unsigned char *str_rd(struct option *o, unsigned char *c)
 {
 	unsigned char *tok = get_token(&c);
 	unsigned char *e = NULL;
-	if (!tok) return NULL;
+	if (!tok) tok = stracpy(cast_uchar "");
 	if (strlen(tok) + 1 > (size_t)o->max) e = "String too long";
 	else strcpy(o->ptr, tok);
 	mem_free(tok);
@@ -552,31 +976,30 @@ int getnum(unsigned char *s, int *n, int r1, int r2)
 unsigned char *type_rd(struct option *o, unsigned char *c)
 {
 	unsigned char *err = "Error reading association specification";
-	struct assoc new;
+	struct assoc neww;
 	unsigned char *w;
 	int n;
-	memset(&new, 0, sizeof(struct assoc));
-	if (!(new.label = get_token(&c))) goto err;
-	if (!(new.ct = get_token(&c))) goto err;
-	if (!(new.prog = get_token(&c))) goto err;
+	memset(&neww, 0, sizeof(struct assoc));
+	if (!(neww.label = get_token(&c))) goto err;
+	if (!(neww.ct = get_token(&c))) goto err;
+	if (!(neww.prog = get_token(&c))) goto err;
 	if (!(w = get_token(&c))) goto err;
 	if (getnum(w, &n, 0, 32)) goto err_f;
 	mem_free(w);
-	new.cons = !!(n & 1);
-	new.xwin = !!(n & 2);
-	new.ask = !!(n & 4);
-	if ((n & 8) || (n & 16)) new.block = !!(n & 16);
-	else new.block = !new.xwin || new.cons;
+	neww.cons = !!(n & 1);
+	neww.xwin = !!(n & 2);
+	neww.ask = !!(n & 4);
+	if ((n & 8) || (n & 16)) neww.block = !!(n & 16);
+	else neww.block = !neww.xwin || neww.cons;
 	if (!(w = get_token(&c))) goto err;
-	if (strlen(w) != 1 || w[0] < '0' || w[0] > '9') goto err_f;
-	new.system = w[0] - '0';
+	if (getnum(w, &neww.system, 0, 256)) goto err_f;
 	mem_free(w);
-	update_assoc(&new);
+	update_assoc(&neww);
 	err = NULL;
 	err:
-	if (new.label) mem_free(new.label);
-	if (new.ct) mem_free(new.ct);
-	if (new.prog) mem_free(new.prog);
+	if (neww.label) mem_free(neww.label);
+	if (neww.ct) mem_free(neww.ct);
+	if (neww.prog) mem_free(neww.prog);
 	return err;
 	err_f:
 	mem_free(w);
@@ -603,15 +1026,15 @@ void type_wr(struct option *o, unsigned char **s, int *l)
 unsigned char *ext_rd(struct option *o, unsigned char *c)
 {
 	unsigned char *err = "Error reading extension specification";
-	struct extension new;
-	memset(&new, 0, sizeof(struct extension));
-	if (!(new.ext = get_token(&c))) goto err;
-	if (!(new.ct = get_token(&c))) goto err;
-	update_ext(&new);
+	struct extension neww;
+	memset(&neww, 0, sizeof(struct extension));
+	if (!(neww.ext = get_token(&c))) goto err;
+	if (!(neww.ct = get_token(&c))) goto err;
+	update_ext(&neww);
 	err = NULL;
 	err:
-	if (new.ext) mem_free(new.ext);
-	if (new.ct) mem_free(new.ct);
+	if (neww.ext) mem_free(neww.ext);
+	if (neww.ct) mem_free(neww.ct);
 	return err;
 }
 
@@ -630,10 +1053,11 @@ unsigned char *prog_rd(struct option *o, unsigned char *c)
 {
 	unsigned char *err = "Error reading program specification";
 	unsigned char *prog, *w;
+	int n;
 	if (!(prog = get_token(&c))) goto err_1;
 	if (!(w = get_token(&c))) goto err_2;
-	if (strlen(w) != 1 || w[0] < '0' || w[0] > '9') goto err_3;
-	update_prog(o->ptr, prog, w[0] - '0');
+	if (getnum(w, &n, 0, 256)) goto err_3;
+	update_prog(o->ptr, prog, n);
 	err = NULL;
 	err_3:
 	mem_free(w);
@@ -777,10 +1201,12 @@ unsigned char *lookup_cmd(struct option *o, unsigned char ***argv, int *argc)
 #else
 		fprintf(stderr, "error: host not found\n");
 #endif
+		exit(RET_ERROR);
 		return "";
 	}
 	printf("%d.%d.%d.%d\n", (int)p[0], (int)p[1], (int)p[2], (int)p[3]);
 	fflush(stdout);
+	exit(RET_OK);
 	return "";
 }
 
@@ -788,6 +1214,7 @@ unsigned char *version_cmd(struct option *o, unsigned char ***argv, int *argc)
 {
 	printf("Links " VERSION_STRING "\n");
 	fflush(stdout);
+	exit(RET_OK);
 	return "";
 }
 
@@ -987,6 +1414,7 @@ Options are:\n\
 	q	 quit\n"));
 
 	fflush(stdout);
+	exit(RET_OK);
 	return "";
 }
 
@@ -997,8 +1425,6 @@ void end_config()
 
 int anonymous = 0;
 
-unsigned char system_name[MAX_STR_LEN];
-
 unsigned char *links_home = NULL;
 int first_use = 0;
 int created_home = 0;
@@ -1010,8 +1436,14 @@ int force_html = 0;
 
 int async_lookup = 1;
 int download_utime = 0;
+#ifdef DOS
+/* DOS networking is slow with too many connections */
+int max_connections = 3;
+int max_connections_to_host = 2;
+#else
 int max_connections = 10;
 int max_connections_to_host = 2;
+#endif
 int max_tries = 3;
 int receive_timeout = 120;
 int unrestartable_receive_timeout = 600;
@@ -1156,9 +1588,7 @@ void save_url_history(void)
 	hsl = 0;
 	foreachback(hi, goto_url_history.items) {
 		if (!*hi->d || strchr(hi->d, 10) || strchr(hi->d, 13)) continue;
-		if (i++ > MAX_HISTORY_ITEMS)
-			break;
-		else {
+		if (!url_not_saveable(hi->d) && i++ <= MAX_HISTORY_ITEMS) {
 			add_to_str(&hs, &hsl, hi->d);
 			add_to_str(&hs, &hsl, NEWLINE);
 		}
