@@ -4,6 +4,10 @@
 #define EXTERNAL_LOOKUP
 #endif
 
+#ifdef OPENVMS_64BIT
+#define addrinfo	__addrinfo64
+#endif
+
 struct dnsentry {
 	struct dnsentry *next;
 	struct dnsentry *prev;
@@ -56,30 +60,33 @@ static int do_external_lookup(unsigned char *name, ip *host)
 	int pi[2];
 	pid_t f;
 	unsigned char *n;
-	if (pipe(pi) == -1)
+	int rs;
+	if (c_pipe(pi) == -1)
 		return -1;
-	f = fork();
+	EINTRLOOP(f, fork());
 	if (f == -1) {
-		close(pi[0]);
-		close(pi[1]);
+		EINTRLOOP(rs, close(pi[0]));
+		EINTRLOOP(rs, close(pi[1]));
 		return -1;
 	}
 	if (!f) {
-		close(pi[0]);
-		if (dup2(pi[1], 1) == -1) _exit(1);
-		if (dup2(pi[1], 2) == -1) _exit(1);
-		close(pi[1]);
-		execlp("host", "host", name, NULL);
-		execl("/usr/sbin/host", "host", name, NULL);
+		EINTRLOOP(rs, close(pi[0]));
+		EINTRLOOP(rs, dup2(pi[1], 1));
+		if (rs == -1) _exit(1);
+		EINTRLOOP(rs, dup2(pi[1], 2));
+		if (rs == -1) _exit(1);
+		EINTRLOOP(rs, close(pi[1]));
+		EINTRLOOP(rs, execlp("host", "host", name, NULL));
+		EINTRLOOP(rs, execl("/usr/sbin/host", "host", name, NULL));
 		_exit(1);
 	}
-	close(pi[1]);
+	EINTRLOOP(rs, close(pi[1]));
 	rd = hard_read(pi[0], buffer, sizeof buffer - 1);
 	if (rd >= 0) buffer[rd] = 0;
 	if (rd > 0) {
 		while (hard_read(pi[0], sink, sizeof sink) > 0);
 	}
-	close(pi[0]);
+	EINTRLOOP(rs, close(pi[0]));
 	/* Don't wait for the process, we already have sigchld handler that
 	 * does cleanup.
 	 * waitpid(f, NULL, 0); */
@@ -108,8 +115,6 @@ int do_real_lookup(unsigned char *name, ip *host)
 {
 	unsigned char *n;
 	struct hostent *hst;
-	if (!*name) return -1;
-	for (n = name; *n; n++) if (*n != '.' && (*n < '0' || *n > '9')) goto nogethostbyaddr;
 	n = name;
 	if (get_addr_byte(&n, ((unsigned char *)host + 0), '.')) goto skip_addr;
 	if (get_addr_byte(&n, ((unsigned char *)host + 1), '.')) goto skip_addr;
@@ -117,22 +122,16 @@ int do_real_lookup(unsigned char *name, ip *host)
 	if (get_addr_byte(&n, ((unsigned char *)host + 3), 0)) goto skip_addr;
 	return 0;
 	skip_addr:
-#ifdef HAVE_GETHOSTBYADDR
-	if (!(hst = gethostbyaddr(name, strlen(name), AF_INET)))
+#ifdef HAVE_GETHOSTBYNAME
+	if (!(hst = gethostbyname(name)))
 #endif
 	{
-		nogethostbyaddr:
-#ifdef HAVE_GETHOSTBYNAME
-		if (!(hst = gethostbyname(name)))
-#endif
-		{
 #ifdef EXTERNAL_LOOKUP
-			return do_external_lookup(name, host);
+		return do_external_lookup(name, host);
 #endif
-			return -1;
-		}
+		return -1;
 	}
-	memcpy(host, hst->h_addr_list[0], sizeof(ip));
+	memcpy(host, hst->h_addr, sizeof(ip));
 	return 0;
 }
 
@@ -140,25 +139,27 @@ void lookup_fn(unsigned char *name, int h)
 {
 	ip host;
 	if (do_real_lookup(name, &host)) return;
-	write(h, &host, sizeof(ip));
+	hard_write(h, (unsigned char *)&host, sizeof(ip));
 }
 
 void end_real_lookup(struct dnsquery *q)
 {
 	int r = 1;
-	if (!q->addr || read(q->h, q->addr, sizeof(ip)) != sizeof(ip)) goto end;
+	int rs;
+	if (!q->addr || hard_read(q->h, (unsigned char *)q->addr, sizeof(ip)) != sizeof(ip)) goto end;
 	r = 0;
 
 	end:
 	set_handlers(q->h, NULL, NULL, NULL, NULL);
-	close(q->h);
+	EINTRLOOP(rs, close(q->h));
 	q->xfn(q, r);
 }
 
 void failed_real_lookup(struct dnsquery *q)
 {
+	int rs;
 	set_handlers(q->h, NULL, NULL, NULL, NULL);
-	close(q->h);
+	EINTRLOOP(rs, close(q->h));
 	q->xfn(q, -1);
 }
 
@@ -259,10 +260,7 @@ void end_dns_lookup(struct dnsquery *q, int a)
 int find_host_no_cache(unsigned char *name, ip *addr, void **qp, void (*fn)(void *, int), void *data)
 {
 	struct dnsquery *q;
-	if (!(q = (struct dnsquery *)malloc(sizeof(struct dnsquery) + strlen(name) + 1))) {
-		fn(data, -1);
-		return 0;
-	}
+	q = (struct dnsquery *)malloc(sizeof(struct dnsquery) + strlen(name));
 	q->fn = fn;
 	q->data = data;
 	q->s = (struct dnsquery **)qp;
@@ -290,9 +288,6 @@ int find_host(unsigned char *name, ip *addr, void **qp, void (*fn)(void *, int),
 void kill_dns_request(void **qp)
 {
 	struct dnsquery *q = *qp;
-	/*set_handlers(q->h, NULL, NULL, NULL, NULL);
-	close(q->h);
-	mem_free(q);*/
 	q->fn = NULL;
 	q->addr = NULL;
 	*qp = NULL;
