@@ -10,8 +10,7 @@ void do_not_optimize_here(void *p)
 }
 
 #ifdef LEAK_DEBUG
-long mem_amount = 0;
-long last_mem_amount = -1;
+my_uintptr_t mem_amount = 0;
 #ifdef LEAK_DEBUG_LIST
 struct list_head memory_list = { &memory_list, &memory_list };
 #endif
@@ -19,18 +18,19 @@ struct list_head memory_list = { &memory_list, &memory_list };
 
 static inline void force_dump()
 {
-	fprintf(stderr, "\n\033[1m%s\033[0m\n", "Forcing core dump");
+	int rs;
+	fprintf(stderr, "\n"ANSI_SET_BOLD"Forcing core dump"ANSI_CLEAR_BOLD"\n");
 	fflush(stdout);
 	fflush(stderr);
 	fatal_tty_exit();
-	raise(SIGSEGV);
+	EINTRLOOP(rs, raise(SIGSEGV));
 }
 
 void check_memory_leaks()
 {
 #ifdef LEAK_DEBUG
 	if (mem_amount) {
-		fprintf(stderr, "\n\033[1mMemory leak by %ld bytes\033[0m\n", mem_amount);
+		fprintf(stderr, "\n"ANSI_SET_BOLD"Memory leak by %lu bytes"ANSI_CLEAR_BOLD"\n", (unsigned long)mem_amount);
 #ifdef LEAK_DEBUG_LIST
 		fprintf(stderr, "\nList of blocks: ");
 		{
@@ -48,19 +48,20 @@ void check_memory_leaks()
 #endif
 }
 
-void er(int b, unsigned char *m, va_list l)
+void er(int b, char *m, va_list l)
 {
-	if (b) fprintf(stderr, "%c", (char)7);
 #ifdef HAVE_VPRINTF
 	vfprintf(stderr, m, l);
 #else
 	fprintf(stderr, "%s", m);
 #endif
+	if (b) fprintf(stderr, ANSI_BELL);
 	fprintf(stderr, "\n");
+	fflush(stderr);
 	sleep(1);
 }
 
-void error(unsigned char *m, ...)
+void error(char *m, ...)
 {
 	va_list l;
 	va_start(l, m);
@@ -73,18 +74,18 @@ unsigned char *errfile;
 
 unsigned char errbuf[4096];
 
-void int_error(unsigned char *m, ...)
+void int_error(char *m, ...)
 {
 	va_list l;
 	va_start(l, m);
-	sprintf(errbuf, "\033[1mINTERNAL ERROR\033[0m at %s:%d: ", errfile, errline);
+	sprintf(errbuf, ANSI_SET_BOLD"INTERNAL ERROR"ANSI_CLEAR_BOLD" at %s:%d: ", errfile, errline);
 	strcat(errbuf, m);
 	er(1, errbuf, l);
 	force_dump();
 	va_end(l);
 }
 
-void debug_msg(unsigned char *m, ...)
+void debug_msg(char *m, ...)
 {
 	va_list l;
 	va_start(l, m);
@@ -108,7 +109,7 @@ void *debug_mem_alloc(unsigned char *file, int line, size_t size)
 	mem_amount += size;
 	size += L_D_S;
 #endif
-	if (!(p = xmalloc(size))) {
+	if (!(p = malloc(size))) {
 		error("ERROR: out of memory (malloc returned NULL)");
 		fatal_tty_exit();
 		exit(RET_FATAL);
@@ -147,7 +148,7 @@ void debug_mem_free(unsigned char *file, int line, void *p)
 #endif
 	mem_amount -= ah->size;
 #endif
-	xfree(p);
+	free(p);
 }
 
 void *debug_mem_realloc(unsigned char *file, int line, void *p, size_t size)
@@ -157,7 +158,7 @@ void *debug_mem_realloc(unsigned char *file, int line, void *p, size_t size)
 #endif
 	if (p == DUMMY) return debug_mem_alloc(file, line, size);
 	if (!p) {
-		errfile = file, errline = line, int_error("mem_realloc(NULL, %d)", size);
+		errfile = file, errline = line, int_error("mem_realloc(NULL, %lu)", (unsigned long)size);
 		return NULL;
 	}
 	if (!size) {
@@ -165,7 +166,7 @@ void *debug_mem_realloc(unsigned char *file, int line, void *p, size_t size)
 		return DUMMY;
 	}
 	if (size > MAXINT) overalloc();
-	if (!(p = xrealloc((char *)p - L_D_S, size + L_D_S))) {
+	if (!(p = realloc((char *)p - L_D_S, size + L_D_S))) {
 		error("ERROR: out of memory (realloc returned NULL)");
 		fatal_tty_exit();
 		exit(RET_FATAL);
