@@ -12,8 +12,10 @@ struct ftp_connection_info {
 	int pending_commands;
 	int opc;
 	int pasv;
+	int eprt_epsv;
 	int dir;
 	int rest_sent;
+	int we_are_in_root;
 	int conn_st;
 	int d;
 	int dpos;
@@ -31,16 +33,16 @@ static void ftp_got_info(struct connection *, struct read_buffer *);
 static void ftp_got_user_info(struct connection *, struct read_buffer *);
 static void ftp_dummy_info(struct connection *, struct read_buffer *);
 static void ftp_pass_info(struct connection *, struct read_buffer *);
-static void ftp_send_retr_req(struct connection *, int);
-static struct ftp_connection_info *add_file_cmd_to_str(struct connection *);
+static void ftp_send_retr_req(struct connection *, int, int);
+static struct ftp_connection_info *add_file_cmd_to_str(struct connection *, int);
 static void ftp_retr_1(struct connection *);
 static void ftp_retr_file(struct connection *, struct read_buffer *);
 static void ftp_got_final_response(struct connection *, struct read_buffer *);
 static void created_data_connection(struct connection *);
 static void got_something_from_data_connection(struct connection *);
-static void ftp_end_request(struct connection *);
+static void ftp_end_request(struct connection *, int);
 static int get_ftp_response(struct connection *, struct read_buffer *, int);
-static int ftp_process_dirlist(struct cache_entry *, off_t *, int *, unsigned char *, int, int, int *);
+static int ftp_process_dirlist(struct cache_entry *, off_t *, int *, unsigned char *, int, int, int, int *);
 
 
 static int get_ftp_response(struct connection *c, struct read_buffer *rb, int part)
@@ -50,7 +52,7 @@ static int get_ftp_response(struct connection *c, struct read_buffer *rb, int pa
 	again:
 	for (l = 0; l < rb->len; l++) if (rb->data[l] == 10) {
 		unsigned char *e;
-		long k = strtoul(rb->data, (char **)(void *)&e, 10);
+		long k = strtoul(cast_const_char rb->data, (char **)(void *)&e, 10);
 		if (e != rb->data + 3 || k < 100 || k >= 1000) return -1;
 		if (*e == '-') {
 			int i;
@@ -68,26 +70,40 @@ static int get_ftp_response(struct connection *c, struct read_buffer *rb, int pa
 			kill_buffer_data(rb, l + 1);
 			goto again;
 		}
-		if (part == 2) return k;
+		if (part == 2) return (int)k;
 		kill_buffer_data(rb, l + 1);
-		return k;
+		return (int)k;
 	}
 	return 0;
 }
 
 void ftp_func(struct connection *c)
 {
-	/*setcstate(c, S_CONN);*/
-	/*set_timeout(c);*/
-	if (get_keepalive_socket(c)) {
+	int we_are_in_root;
+	unsigned char *de, *d;
+	int del, bad_url;
+	d = get_url_data(c->url);
+	de = init_str(), del = 0;
+	add_conv_str(&de, &del, d, (int)strcspn(cast_const_char d, POST_CHAR_STRING), -2);
+	bad_url = !!strchr(cast_const_char de, 10);
+	mem_free(de);
+	if (bad_url) {
+		setcstate(c, S_BAD_URL);
+		abort_connection(c);
+		return;
+	}
+
+	if (get_keepalive_socket(c, &we_are_in_root)) {
 		int p;
 		if ((p = get_port(c->url)) == -1) {
-			setcstate(c, S_INTERNAL);
+			setcstate(c, S_BAD_URL);
 			abort_connection(c);
 			return;
 		}
 		make_connection(c, p, &c->sock1, ftp_options.fast_ftp ? ftp_login : ftp_get_banner);
-	} else ftp_send_retr_req(c, S_SENT);
+	} else {
+		ftp_send_retr_req(c, S_SENT, we_are_in_root);
+	}
 }
 
 static void ftp_get_banner(struct connection *c)
@@ -115,24 +131,24 @@ static void ftp_login(struct connection *c)
 	int logl = 0;
 	set_timeout(c);
 	login = init_str();
-	add_to_str(&login, &logl, "USER ");
+	add_to_str(&login, &logl, cast_uchar "USER ");
 	if ((u = get_user_name(c->url)) && *u) add_to_str(&login, &logl, u);
-	else add_to_str(&login, &logl, "anonymous");
+	else add_to_str(&login, &logl, cast_uchar "anonymous");
 	if (u) mem_free(u);
 	if (ftp_options.fast_ftp) {
 		struct ftp_connection_info *fi;
-		add_to_str(&login, &logl, "\r\nPASS ");
+		add_to_str(&login, &logl, cast_uchar "\r\nPASS ");
 		if ((u = get_pass(c->url)) && *u) add_to_str(&login, &logl, u);
 		else add_to_str(&login, &logl, ftp_options.anon_pass);
 		if (u) mem_free(u);
-		add_to_str(&login, &logl, "\r\n");
-		if (!(fi = add_file_cmd_to_str(c))) {
+		add_to_str(&login, &logl, cast_uchar "\r\n");
+		if (!(fi = add_file_cmd_to_str(c, 0))) {
 			mem_free(login);
 			return;
 		}
 		add_to_str(&login, &logl, fi->cmdbuf);
-	} else add_to_str(&login, &logl, "\r\n");
-	write_to_socket(c, c->sock1, login, strlen(login), ftp_logged);
+	} else add_to_str(&login, &logl, cast_uchar "\r\n");
+	write_to_socket(c, c->sock1, login, logl, ftp_logged);
 	mem_free(login);
 	setcstate(c, S_SENT);
 }
@@ -166,7 +182,7 @@ static void ftp_got_user_info(struct connection *c, struct read_buffer *rb)
 	if (g >= 400) { setcstate(c, S_FTP_UNAVAIL); retry_connection(c); return; }
 	if (g >= 200 && g < 300) {
 		if (ftp_options.fast_ftp) ftp_dummy_info(c, rb);
-		else ftp_send_retr_req(c, S_GETH);
+		else ftp_send_retr_req(c, S_GETH, 0);
 	} else {
 		if (ftp_options.fast_ftp) ftp_pass_info(c, rb);
 		else {
@@ -174,12 +190,12 @@ static void ftp_got_user_info(struct connection *c, struct read_buffer *rb)
 			unsigned char *u;
 			int logl = 0;
 			login = init_str();
-			add_to_str(&login, &logl, "PASS ");
+			add_to_str(&login, &logl, cast_uchar "PASS ");
 			if ((u = get_pass(c->url)) && *u) add_to_str(&login, &logl, u);
 			else add_to_str(&login, &logl, ftp_options.anon_pass);
 			if (u) mem_free(u);
-			add_to_str(&login, &logl, "\r\n");
-			write_to_socket(c, c->sock1, login, strlen(login), ftp_sent_passwd);
+			add_to_str(&login, &logl, cast_uchar "\r\n");
+			write_to_socket(c, c->sock1, login, logl, ftp_sent_passwd);
 			mem_free(login);
 			setcstate(c, S_LOGIN);
 		}
@@ -209,16 +225,35 @@ static void ftp_pass_info(struct connection *c, struct read_buffer *rb)
 	if (g >= 530 && g < 540) { setcstate(c, S_FTP_LOGIN); abort_connection(c); return; }
 	if (g >= 400) { setcstate(c, S_FTP_UNAVAIL); abort_connection(c); return; }
 	if (ftp_options.fast_ftp) ftp_retr_file(c, rb);
-	else ftp_send_retr_req(c, S_GETH);
+	else ftp_send_retr_req(c, S_GETH, 0);
 }
 
-static struct ftp_connection_info *add_file_cmd_to_str(struct connection *c)
+static void add_port_pasv(unsigned char **s, int *l, struct ftp_connection_info *inf, unsigned char *port_string)
+{
+	if (!inf->pasv) {
+		if (inf->eprt_epsv) {
+			add_to_str(s, l, cast_uchar "EPRT ");
+			add_to_str(s, l, port_string);
+		} else {
+			add_to_str(s, l, cast_uchar "PORT ");
+			add_to_str(s, l, port_string);
+		}
+	} else {
+		if (inf->eprt_epsv) {
+			add_to_str(s, l, cast_uchar "EPSV");
+		} else {
+			add_to_str(s, l, cast_uchar "PASV");
+		}
+	}
+	add_to_str(s, l, cast_uchar "\r\n");
+}
+
+static struct ftp_connection_info *add_file_cmd_to_str(struct connection *c, int we_are_in_root)
 {
 	unsigned char *d = get_url_data(c->url);
-	unsigned char *de;
+	unsigned char *dd, *de;
 	int del;
-	unsigned char pc[6];
-	int ps;
+	unsigned char port_string[50];
 	struct ftp_connection_info *inf, *inf2;
 	unsigned char *s;
 	int l;
@@ -228,87 +263,106 @@ static struct ftp_connection_info *add_file_cmd_to_str(struct connection *c)
 		abort_connection(c);
 		return NULL;
 	}
+
 	de = init_str(), del = 0;
-	add_conv_str(&de, &del, d, strlen(d), -2);
+	add_conv_str(&de, &del, d, (int)strlen(cast_const_char d), -2);
 	d = de;
 	inf = mem_alloc(sizeof(struct ftp_connection_info));
 	memset(inf, 0, sizeof(struct ftp_connection_info));
 	l = 0;
 	s = init_str();
+	inf->we_are_in_root = we_are_in_root;
 	inf->pasv = ftp_options.passive_ftp;
+#ifdef LINKS_2
+	if (*c->socks_proxy) inf->pasv = 1;
+	if (ftp_options.eprt_epsv || is_ipv6(c->sock1)) inf->eprt_epsv = 1;
+#endif
 	c->info = inf;
-	if (!inf->pasv) if ((ps = get_pasv_socket(c, c->sock1, &c->sock2, pc))) {
-		mem_free(d);
-		return NULL;
+
+	if (!inf->pasv) {
+		int ps;
+#ifdef SUPPORT_IPV6
+		if (is_ipv6(c->sock1)) {
+			ps = get_pasv_socket_ipv6(c, c->sock1, &c->sock2, port_string);
+			if (ps) {
+				mem_free(d);
+				mem_free(s);
+				return NULL;
+			}
+		} else
+#endif
+		{
+			unsigned char pc[6];
+			ps = get_pasv_socket(c, c->sock1, &c->sock2, pc);
+			if (ps) {
+				mem_free(d);
+				mem_free(s);
+				return NULL;
+			}
+			if (inf->eprt_epsv)
+				sprintf(cast_char port_string, "|1|%d.%d.%d.%d|%d|", pc[0], pc[1], pc[2], pc[3], (pc[4] << 8) | pc[5]);
+			else
+				sprintf(cast_char port_string, "%d,%d,%d,%d,%d,%d", pc[0], pc[1], pc[2], pc[3], pc[4], pc[5]);
+		}
+		if (strlen(cast_const_char port_string) >= sizeof(port_string))
+			internal("buffer overflow in get_pasv_socket_ipv6: %d > %d", (int)strlen(cast_const_char port_string), (int)sizeof(port_string));
 	}
 #ifdef HAVE_IPTOS
 	if (ftp_options.set_tos) {
+		int rx;
 		int on = IPTOS_THROUGHPUT;
-		setsockopt(c->sock2, IPPROTO_IP, IP_TOS, (char *)&on, sizeof(int));
+		EINTRLOOP(rx, setsockopt(c->sock2, IPPROTO_IP, IP_TOS, (char *)&on, sizeof(int)));
 	}
 #endif
-	if (!(de = strchr(d, POST_CHAR))) de = d + strlen(d);
-	if (d == de || de[-1] == '/') {
+	dd = d;
+	while (*dd == '/') dd++;
+	de = cast_uchar strchr(cast_const_char dd, 0);
+	if (dd == de || de[-1] == '/') {
 		inf->dir = 1;
-		inf->pending_commands = 4;
-		add_to_str(&s, &l, "TYPE A\r\n");
-		if (!inf->pasv) {
-			add_to_str(&s, &l, "PORT ");
-			add_num_to_str(&s, &l, pc[0]);
-			add_chr_to_str(&s, &l, ',');
-			add_num_to_str(&s, &l, pc[1]);
-			add_chr_to_str(&s, &l, ',');
-			add_num_to_str(&s, &l, pc[2]);
-			add_chr_to_str(&s, &l, ',');
-			add_num_to_str(&s, &l, pc[3]);
-			add_chr_to_str(&s, &l, ',');
-			add_num_to_str(&s, &l, pc[4]);
-			add_chr_to_str(&s, &l, ',');
-			add_num_to_str(&s, &l, pc[5]);
-			add_to_str(&s, &l, "\r\n");
-		} else {
-			add_to_str(&s, &l, "PASV\r\n");
+		inf->pending_commands = 3;
+		add_to_str(&s, &l, cast_uchar "TYPE A\r\n");
+		add_port_pasv(&s, &l, inf, port_string);
+		if (!inf->we_are_in_root) {
+			add_to_str(&s, &l, cast_uchar "CWD /\r\n");
+			inf->we_are_in_root = 1;
+			inf->pending_commands++;
 		}
-		add_to_str(&s, &l, "CWD /");
-		add_bytes_to_str(&s, &l, d, de - d);
-		add_to_str(&s, &l, "\r\nLIST\r\n");
+		if (dd != de) {
+			add_to_str(&s, &l, cast_uchar "CWD ");
+			add_bytes_to_str(&s, &l, dd, de - 1 - dd);
+			add_to_str(&s, &l, cast_uchar "\r\n");
+			inf->we_are_in_root = 0;
+			inf->pending_commands++;
+		}
+		add_to_str(&s, &l, cast_uchar "LIST\r\n");
 		c->from = 0;
 	} else {
 		inf->dir = 0;
 		inf->pending_commands = 3;
-		add_to_str(&s, &l, "TYPE I\r\n");
-		if (!inf->pasv) {
-			add_to_str(&s, &l, "PORT ");
-			add_num_to_str(&s, &l, pc[0]);
-			add_chr_to_str(&s, &l, ',');
-			add_num_to_str(&s, &l, pc[1]);
-			add_chr_to_str(&s, &l, ',');
-			add_num_to_str(&s, &l, pc[2]);
-			add_chr_to_str(&s, &l, ',');
-			add_num_to_str(&s, &l, pc[3]);
-			add_chr_to_str(&s, &l, ',');
-			add_num_to_str(&s, &l, pc[4]);
-			add_chr_to_str(&s, &l, ',');
-			add_num_to_str(&s, &l, pc[5]);
-			add_to_str(&s, &l, "\r\n");
-		} else {
-			add_to_str(&s, &l, "PASV\r\n");
+		add_to_str(&s, &l, cast_uchar "TYPE I\r\n");
+		add_port_pasv(&s, &l, inf, port_string);
+		if (!inf->we_are_in_root) {
+			add_to_str(&s, &l, cast_uchar "CWD /\r\n");
+			inf->we_are_in_root = 1;
+			inf->pending_commands++;
 		}
 		if (c->from && c->no_cache < NC_IF_MOD) {
-			add_to_str(&s, &l, "REST ");
+			add_to_str(&s, &l, cast_uchar "REST ");
 			add_num_to_str(&s, &l, c->from);
-			add_to_str(&s, &l, "\r\n");
+			add_to_str(&s, &l, cast_uchar "\r\n");
 			inf->rest_sent = 1;
 			inf->pending_commands++;
-		} else c->from = 0;
-		add_to_str(&s, &l, "RETR /");
-		add_bytes_to_str(&s, &l, d, de - d);
-		add_to_str(&s, &l, "\r\n");
+		} else {
+			c->from = 0;
+		}
+		add_to_str(&s, &l, cast_uchar "RETR ");
+		add_bytes_to_str(&s, &l, dd, de - dd);
+		add_to_str(&s, &l, cast_uchar "\r\n");
 	}
 	inf->opc = inf->pending_commands;
 	if ((unsigned)l > MAXINT - sizeof(struct ftp_connection_info) - 1) overalloc();
 	inf2 = mem_realloc(inf, sizeof(struct ftp_connection_info) + l + 1);
-	strcpy((inf = inf2)->cmdbuf, s);
+	strcpy(cast_char (inf = inf2)->cmdbuf, cast_const_char s);
 	mem_free(s);
 	c->info = inf;
 	mem_free(d);
@@ -316,26 +370,29 @@ static struct ftp_connection_info *add_file_cmd_to_str(struct connection *c)
 }
 
 
-static void ftp_send_retr_req(struct connection *c, int state)
+static void ftp_send_retr_req(struct connection *c, int state, int we_are_in_root)
 {
 	struct ftp_connection_info *fi;
 	unsigned char *login;
 	int logl = 0;
 	set_timeout(c);
 	login = init_str();
-	if (!c->info && !(fi = add_file_cmd_to_str(c))) {
+	if (!c->info && !(fi = add_file_cmd_to_str(c, we_are_in_root))) {
 		mem_free(login);
 		return;
-	} else fi = c->info;
-	if (ftp_options.fast_ftp) a:add_to_str(&login, &logl, fi->cmdbuf);
-	else {
-		unsigned char *nl = strchr(fi->cmdbuf, '\n');
+	} else {
+		fi = c->info;
+	}
+	if (ftp_options.fast_ftp) {
+		a:add_to_str(&login, &logl, fi->cmdbuf);
+	} else {
+		unsigned char *nl = cast_uchar strchr(cast_const_char fi->cmdbuf, '\n');
 		if (!nl) goto a;
 		nl++;
 		add_bytes_to_str(&login, &logl, fi->cmdbuf, nl - fi->cmdbuf);
-		memmove(fi->cmdbuf, nl, strlen(nl) + 1);
+		memmove(fi->cmdbuf, nl, strlen(cast_const_char nl) + 1);
 	}
-	write_to_socket(c, c->sock1, login, strlen(login), ftp_retr_1);
+	write_to_socket(c, c->sock1, login, logl, ftp_retr_1);
 	mem_free(login);
 	setcstate(c, state);
 }
@@ -354,15 +411,16 @@ static void ftp_retr_file(struct connection *c, struct read_buffer *rb)
 	if (0) {
 		rep:
 		if (!ftp_options.fast_ftp) {
-			ftp_send_retr_req(c, S_GETH);
+			ftp_send_retr_req(c, S_GETH, inf->we_are_in_root);
 			return;
 		}
 	}
 	if (inf->pending_commands > 1) {
 		unsigned char pc[6];
 		if (inf->pasv && inf->opc - (inf->pending_commands - 1) == 2) {
-			int i = 3, j;
-			while (i < rb->len) {
+			int i, j;
+			i = 3;
+			if (!inf->eprt_epsv) while (i < rb->len) {
 				if (rb->data[i] >= '0' && rb->data[i] <= '9') {
 					for (j = 0; j < 6; j++) {
 						int n = 0;
@@ -371,7 +429,7 @@ static void ftp_retr_file(struct connection *c, struct read_buffer *rb)
 							if (n >= 256) goto no_pasv;
 							if (++i >= rb->len) goto no_pasv;
 						}
-						pc[j] = n;
+						pc[j] = (unsigned char)n;
 						if (j != 5) {
 							if (rb->data[i] != ',') goto xa;
 							if (++i >= rb->len) goto xa;
@@ -390,6 +448,26 @@ static void ftp_retr_file(struct connection *c, struct read_buffer *rb)
 				i++;
 			}
 			no_pasv:
+			i = 3;
+			while (i < rb->len - 5) {
+				if (rb->data[i] == '(' && (rb->data[i + 1] < '0' || rb->data[i + 1] > '9') && rb->data[i + 1] == rb->data[i + 2] && rb->data[i + 2] == rb->data[i + 3]) {
+					unsigned char delim = rb->data[i + 1];
+					int n = 0;
+					i += 4;
+					while (rb->data[i] >= '0' && rb->data[i] <= '9') {
+						n = n * 10 + rb->data[i] - '0';
+						if (n >= 65536) goto no_epsv;
+						if (++i >= rb->len) goto no_epsv;
+					}
+					if (rb->data[i] != delim) goto no_epsv;
+					pc[4] = n >> 8;
+					pc[5] = n & 0xff;
+					pc[0] = pc[1] = pc[2] = pc[3] = 0;
+					goto pasv_ok;
+				}
+				i++;
+			}
+			no_epsv:
 			memset(pc, 0, sizeof pc);
 			pasv_ok:;
 		}
@@ -412,8 +490,9 @@ static void ftp_retr_file(struct connection *c, struct read_buffer *rb)
 				}
 				goto rep;
 			case 3:		/* REST / CWD */
+			case 4:
 				if (g >= 400) {
-					if (!inf->dir) c->from = 0;
+					if (!inf->dir && c->from && inf->pending_commands == 1) c->from = 0;
 					else { setcstate(c, S_FTP_NO_FILE); abort_connection(c); return; }
 				}
 				goto rep;
@@ -434,18 +513,15 @@ static void ftp_retr_file(struct connection *c, struct read_buffer *rb)
 		quak:
 		for (; i < rb->len; i++) if (d[i] != ' ') break;
 		if (i + 4 > rb->len) goto nol;
-		if (casecmp(&d[i], "byte", 4)) goto nol;
+		if (casecmp(&d[i], cast_uchar "byte", 4)) goto nol;
+		/* when resuming file transfer, some servers return total size,
+		   others return the number of remaining bytes. So it is not
+		   reliable to guess file size in this case */
+		if (c->from) goto nol;
 		{
-#if defined(HAVE_STRTOLL)
-			long long est = strtoll(&d[p], NULL, 10);
-#elif defined(HAVE_STRTOQ)
-			longlong est = strtoq(&d[p], NULL, 10);
-#else
-			long est = strtol(&d[p], NULL, 10);
-			if (est == MAXLONG) est = -1;
-#endif
+			my_strtoll_t est = my_strtoll(&d[p], NULL);
 			if (est < 0 || (off_t)est < 0 || (off_t)est != est) est = 0;
-			if (est && !c->from) c->est_length = est; /* !!! FIXME: when I add appedning to downloaded file */
+			if (est) c->est_length = est + c->from;
 		}
 		nol:;
 	}
@@ -462,7 +538,7 @@ static void ftp_got_final_response(struct connection *c, struct read_buffer *rb)
 	if (g == -1) { setcstate(c, S_FTP_ERROR); abort_connection(c); return; }
 	if (!g) { read_from_socket(c, c->sock1, rb, ftp_got_final_response); if (c->state != S_TRANS) setcstate(c, S_GETH); return; }
 	if (g == 425 || g == 450 || g == 500 || g == 501 || g == 550) {
-		if (c->url[strlen(c->url) - 1] == '/') goto skip_redir;
+		if (c->url[strlen(cast_const_char c->url) - 1] == '/') goto skip_redir;
 		if (!c->cache) {
 			if (get_cache_entry(c->url, &c->cache)) {
 				setcstate(c, S_OUT_OF_MEM);
@@ -474,25 +550,24 @@ static void ftp_got_final_response(struct connection *c, struct read_buffer *rb)
 		if (c->cache->redirect) mem_free(c->cache->redirect);
 		c->cache->redirect = stracpy(c->url);
 		c->cache->redirect_get = 1;
-		add_to_strn(&c->cache->redirect, "/");
+		add_to_strn(&c->cache->redirect, cast_uchar "/");
 		c->cache->incomplete = 0;
 		/*setcstate(c, S_FTP_NO_FILE);*/
-		setcstate(c, S_OK);
+		setcstate(c, S__OK);
 		abort_connection(c);
 		return;
 	}
 	skip_redir:
 	if (g >= 400) { setcstate(c, S_FTP_FILE_ERROR); abort_connection(c); return; }
 	if (inf->conn_st == 2) {
-		setcstate(c, S_OK);
-		ftp_end_request(c);
+		ftp_end_request(c, S__OK);
 	} else {
 		inf->conn_st = 1;
 		if (c->state != S_TRANS) setcstate(c, S_GETH);
 	}
 }
 
-static int is_date(char *data)	/* can touch at most data[-4] --- "n 12 "<--if fed with this --- if you change it, fix the caller */
+static int is_date(unsigned char *data)	/* can touch at most data[-4] --- "n 12 "<--if fed with this --- if you change it, fix the caller */
 {
 	/* fix for ftp://ftp.su.se/ */
 	if (*data == ' ') data--;
@@ -505,7 +580,9 @@ static int is_date(char *data)	/* can touch at most data[-4] --- "n 12 "<--if fe
 	return 1;
 }
 
-static int ftp_process_dirlist(struct cache_entry *ce, off_t *pos, int *d, unsigned char *bf, int ln, int fin, int *tr)
+#define PARSE_MODE_VMS		-1
+
+static int ftp_process_dirlist(struct cache_entry *ce, off_t *pos, int *d, unsigned char *bf, int ln, int fin, int root, int *tr)
 {
 	unsigned char *str, *buf;
 	int sl;
@@ -513,6 +590,10 @@ static int ftp_process_dirlist(struct cache_entry *ce, off_t *pos, int *d, unsig
 	int p;
 	int len;
 	int f;
+	int a;
+	int dir;
+	int sp;
+	int ee;
 	again:
 	buf = bf + ret;
 	len = ln - ret;
@@ -528,41 +609,69 @@ static int ftp_process_dirlist(struct cache_entry *ce, off_t *pos, int *d, unsig
 	pl:
 	str = init_str();
 	sl = 0;
-	/*add_to_str(&str, &sl, "   ");*/
+	/*add_to_str(&str, &sl, cast_uchar "   ");*/
+
+	dir = 0;
+	if (!*d || *d == PARSE_MODE_VMS) for (ee = 0; p - ee > 1 && !WHITECHAR(buf[ee]); ee++) {
+		if (buf[ee] == ';' && buf[ee + 1] >= '1' && buf[ee + 1] <= '9') {
+			if (!ee) goto skip_vms;
+			if (ee >= 4 && buf[ee - 4] == '.' && buf[ee - 3] == 'D' && buf[ee - 2] == 'I' && buf[ee - 1] == 'R') {
+				if (ee == 4) goto raw;
+				dir = 1, ee -= 4;
+			} else {
+				/*if (buf[ee + 1] == '1' && (ee + 2 == p || buf[ee + 2] < '0' || buf[ee + 2] > '9')) goto no_version;
+				ee += 2;
+				while (ee < p && buf[ee] >= '0' && buf[ee] < '9') ee++;
+				no_version:;*/
+			}
+			if (*d != PARSE_MODE_VMS && !root) {
+				add_to_str(&str, &sl, cast_uchar "<a href=\"../\">..</a>\n");
+			}
+			sp = 0;
+			*d = PARSE_MODE_VMS;
+			goto put_entry;
+		}
+	}
+	skip_vms:
+
+	if (*d < 0) goto raw;
+
 	f = *d;
 	if (*d && *d < p && WHITECHAR(buf[*d - 1])) {
-		int ee, dir;
+		sp = *d;
 		ppp:
-		for (ee = *d; ee <= p - 4; ee++)
-			if (!memcmp(buf + ee, " -> ", 4)) goto syml;
+		for (ee = sp; ee <= p - 4; ee++)
+			if (!memcmp(buf + ee, cast_uchar " -> ", 4)) goto syml;
 		ee = p;
 		syml:
-		if (!f) {
-			if ((ee - *d != 1 || buf[*d] != '.') &&
-			    (ee - *d != 2 || buf[*d] != '.' || buf[*d + 1] != '.')) {
+		if (!f && !root) {
+			if ((ee - sp != 1 || buf[sp] != '.') &&
+			    (ee - sp != 2 || buf[sp] != '.' || buf[sp + 1] != '.')) {
 				int i;
-				for (i = 0; i < *d; i++) add_chr_to_str(&str, &sl, ' ');
-				add_to_str(&str, &sl, "<a href=\"../\">..</a>\n");
+				for (i = 0; i < sp; i++) add_chr_to_str(&str, &sl, ' ');
+				add_to_str(&str, &sl, cast_uchar "<a href=\"../\">..</a>\n");
 			}
 		}
 		dir = buf[0] == 'd';
 		if (!dir) {
-			unsigned char *p = memacpy(buf, *d);
-			if (strstr(p, "<DIR>")) dir = 1;
+			unsigned char *p = memacpy(buf, sp);
+			if (strstr(cast_const_char p, "<DIR>")) dir = 1;
 			mem_free(p);
 		};
-		add_conv_str(&str, &sl, buf, *d, 0);
-		add_to_str(&str, &sl, "<a href=\"./");
-		add_conv_str(&str, &sl, buf + *d, ee - *d, 1);
+		put_entry:
+		add_conv_str(&str, &sl, buf, sp, 0);
+		add_to_str(&str, &sl, cast_uchar "<a href=\"./");
+		add_conv_str(&str, &sl, buf + sp, ee - sp, 1);
 		if (dir) add_chr_to_str(&str, &sl, '/');
-		add_to_str(&str, &sl, "\">");
-		add_conv_str(&str, &sl, buf + *d, ee - *d, 0);
-		add_to_str(&str, &sl, "</a>");
+		add_to_str(&str, &sl, cast_uchar "\">");
+		add_conv_str(&str, &sl, buf + sp, ee - sp, 0);
+		add_to_str(&str, &sl, cast_uchar "</a>");
 		add_conv_str(&str, &sl, buf + ee, p - ee, 0);
 	} else {
 		int pp, ppos;
 		int bp, bn;
-		if (p > 5 && !casecmp(buf, "total", 5)) goto raw;
+		if (p > 5 && !casecmp(buf, cast_uchar "total", 5)) goto raw;
+		if (p > 10 && !memcmp(buf, cast_uchar "Directory ", 10)) goto raw;
 		for (pp = p - 1; pp >= 0; pp--) if (!WHITECHAR(buf[pp])) break;
 		if (pp < 0) goto raw;
 		if (pp < p - 1) pp++;
@@ -582,18 +691,22 @@ static int ftp_process_dirlist(struct cache_entry *ce, off_t *pos, int *d, unsig
 			    buf[pp - 4] >= '0' && buf[pp - 4] <= '9' &&
 			    buf[pp - 3] == ':' &&
 			    buf[pp - 2] >= '0' && buf[pp - 2] <= '5' &&
-			    buf[pp - 1] >= '0' && buf[pp - 1] <= '9') ppos = pp;
+			    buf[pp - 1] >= '0' && buf[pp - 1] <= '9') {
+				ppos = pp;
+				if (p - pp > 2 && buf[pp + 1] == ' ' && buf[pp + 2] != ' ')
+					ppos++;
+			}
 		}
 		if (ppos != -1) {
 			pp = ppos;
 			goto done;
 		}
 
-		for (pp = 0; pp + 5 <= p; pp++)
-			if (!casecmp(&buf[pp], "<DIR>", 5)) {
+		for (pp = 0; p - pp >= 5; pp++)
+			if (!casecmp(&buf[pp], cast_uchar "<DIR>", 5)) {
 				pp += 4;
-				while (pp + 1 < p && WHITECHAR(buf[pp + 1])) pp++;
-				if (pp + 1 < p) goto done;
+				while (p - pp > 1 && WHITECHAR(buf[pp + 1])) pp++;
+				if (p - pp > 1) goto done;
 			}
 		
 		bn = -1;
@@ -616,21 +729,23 @@ static int ftp_process_dirlist(struct cache_entry *ce, off_t *pos, int *d, unsig
 		}
 		if (bn >= 0) {
 			pp = bp + bn;
-			while (pp + 1 < p && WHITECHAR(buf[pp + 1])) pp++;
-			if (pp + 1 < p) goto done;
+			while (p - pp > 1 && WHITECHAR(buf[pp + 1])) pp++;
+			if (p - pp > 1) goto done;
 		}
 
 		for (pp = p - 1; pp >= 0; pp--) if (!WHITECHAR(buf[pp])) break;
 		if (pp < 0) goto raw;
-		for (; pp >= 0; pp--) if (WHITECHAR(buf[pp]) && (pp < 3 || memcmp(buf + pp - 3, " -> ", 4)) && (pp > p - 4 || memcmp(buf + pp, " -> ", 4))) break;
+		for (; pp >= 0; pp--) if (WHITECHAR(buf[pp]) && (pp < 3 || memcmp(buf + pp - 3, cast_uchar " -> ", 4)) && (pp > p - 4 || memcmp(buf + pp, cast_uchar " -> ", 4))) break;
 		done:
-		*d = pp + 1;
+		sp = *d = pp + 1;
 		goto ppp;
 		raw:
 		add_conv_str(&str, &sl, buf, p, 0);
 	}
 	add_chr_to_str(&str, &sl, '\n');
-	if (add_fragment(ce, *pos, str, sl)) *tr = 0;
+	a = add_fragment(ce, *pos, str, sl);
+	if (a < 0) return a;
+	if (a == 1) *tr = 0;
 	*pos += sl;
 	mem_free(str);
 	goto again;
@@ -641,8 +756,9 @@ static void created_data_connection(struct connection *c)
 	struct ftp_connection_info *inf = c->info;
 #ifdef HAVE_IPTOS
 	if (ftp_options.set_tos) {
+		int rx;
 		int on = IPTOS_THROUGHPUT;
-		setsockopt(c->sock2, IPPROTO_IP, IP_TOS, (char *)&on, sizeof(int));
+		EINTRLOOP(rx, setsockopt(c->sock2, IPPROTO_IP, IP_TOS, (char *)&on, sizeof(int)));
 	}
 #endif
 	inf->d = 1;
@@ -653,13 +769,17 @@ static void got_something_from_data_connection(struct connection *c)
 {
 	struct ftp_connection_info *inf = c->info;
 	int l;
+	int m;
+	int rs;
 	set_timeout(c);
 	if (!inf->d) {
 		int ns;
 		inf->d = 1;
 		set_handlers(c->sock2, NULL, NULL, NULL, NULL);
-		if ((ns = accept(c->sock2, NULL, NULL)) == -1) goto e;
-		close(c->sock2);
+		EINTRLOOP(ns, accept(c->sock2, NULL, NULL));
+		if (ns == -1) goto e;
+		set_nonblock(ns);
+		EINTRLOOP(rs, close(c->sock2));
 		c->sock2 = ns;
 		set_handlers(ns, (void (*)(void *))got_something_from_data_connection, NULL, NULL, c);
 		return;
@@ -676,27 +796,39 @@ static void got_something_from_data_connection(struct connection *c)
 		unsigned char *ud;
 		unsigned char *s0;
 		int s0l;
+		int err = 0;
 		static unsigned char ftp_head[] = "<html><head><title>/";
 		static unsigned char ftp_head2[] = "</title></head><body><h2>Directory /";
 		static unsigned char ftp_head3[] = "</h2><pre>";
-#define A(s) add_fragment(c->cache, c->from, s, strlen(s)), c->from += strlen(s)
+#define A(s)							\
+do {								\
+	m = add_fragment(c->cache, c->from, s, strlen(cast_const_char s));	\
+	if (m < 0 && !err) err = m;				\
+	c->from += strlen(cast_const_char s);					\
+} while (0)
 		A(ftp_head);
 		ud = stracpy(get_url_data(c->url));
-		if (strchr(ud, POST_CHAR)) *strchr(ud, POST_CHAR) = 0;
+		if (strchr(cast_const_char ud, POST_CHAR)) *strchr(cast_const_char ud, POST_CHAR) = 0;
 		s0 = init_str();
 		s0l = 0;
-		add_conv_str(&s0, &s0l, ud, strlen(ud), -1);
+		add_conv_str(&s0, &s0l, ud, (int)strlen(cast_const_char ud), -1);
 		mem_free(ud);
 		A(s0);
 		A(ftp_head2);
 		A(s0);
 		A(ftp_head3);
 		mem_free(s0);
-		if (!c->cache->head) c->cache->head = stracpy("\r\n");
-		add_to_strn(&c->cache->head, "Content-Type: text/html\r\n");
+		if (!c->cache->head) c->cache->head = stracpy(cast_uchar "\r\n");
+		add_to_strn(&c->cache->head, cast_uchar "Content-Type: text/html\r\n");
+		if (err) {
+			setcstate(c, err);
+			abort_connection(c);
+			return;
+		}
 #undef A
 	}
-	if ((l = read(c->sock2, inf->ftp_buffer + inf->buf_pos, FTP_BUF - inf->buf_pos)) == -1) {
+	EINTRLOOP(l, (int)read(c->sock2, inf->ftp_buffer + inf->buf_pos, FTP_BUF - inf->buf_pos));
+	if (l == -1) {
 		e:
 		if (inf->conn_st != 1 && !inf->dir && !c->from) {
 			set_handlers(c->sock2, NULL, NULL, NULL, NULL);
@@ -716,37 +848,53 @@ static void got_something_from_data_connection(struct connection *c)
 				return;
 			}
 			c->received += l;
-			if (add_fragment(c->cache, c->from, inf->ftp_buffer, l) == 1) c->tries = 0;
+			m = add_fragment(c->cache, c->from, inf->ftp_buffer, l);
+			if (m < 0) {
+				setcstate(c, m);
+				abort_connection(c);
+				return;
+			}
+			if (m == 1) c->tries = 0;
 			c->from += l;
 		} else {
-			int m;
 			c->received += l;
-			m = ftp_process_dirlist(c->cache, &c->from, &inf->dpos, inf->ftp_buffer, l + inf->buf_pos, 0, &c->tries);
+			m = ftp_process_dirlist(c->cache, &c->from, &inf->dpos, inf->ftp_buffer, l + inf->buf_pos, 0, inf->we_are_in_root, &c->tries);
+			if (m < 0) {
+				setcstate(c, m);
+				abort_connection(c);
+				return;
+			}
 			memmove(inf->ftp_buffer, inf->ftp_buffer + m, inf->buf_pos + l - m);
 			inf->buf_pos += l - m;
 		}
 		setcstate(c, S_TRANS);
 		return;
 	}
-	ftp_process_dirlist(c->cache, &c->from, &inf->dpos, inf->ftp_buffer, inf->buf_pos, 1, &c->tries);
+	m = ftp_process_dirlist(c->cache, &c->from, &inf->dpos, inf->ftp_buffer, inf->buf_pos, 1, inf->we_are_in_root, &c->tries);
+	if (m < 0) {
+		setcstate(c, m);
+		abort_connection(c);
+		return;
+	}
 	set_handlers(c->sock2, NULL, NULL, NULL, NULL);
 	close_socket(&c->sock2);
 	if (inf->conn_st == 1) {
-		setcstate(c, S_OK);
-		ftp_end_request(c);
+		ftp_end_request(c, S__OK);
 	} else {
 		inf->conn_st = 2;
 	}
 }
 
-static void ftp_end_request(struct connection *c)
+static void ftp_end_request(struct connection *c, int state)
 {
-	if (c->state == S_OK) {
+	struct ftp_connection_info *inf = c->info;
+	if (state == S__OK) {
 		if (c->cache) {
 			truncate_entry(c->cache, c->from, 1);
 			c->cache->incomplete = 0;
 		}
 	}
-	add_keepalive_socket(c, FTP_KEEPALIVE_TIMEOUT);
+	setcstate(c, state);
+	add_keepalive_socket(c, FTP_KEEPALIVE_TIMEOUT, inf->we_are_in_root);
 }
 
