@@ -1,7 +1,5 @@
 #include "links.h"
 
-#define format format_
-
 struct list_head html_stack = {&html_stack, &html_stack};
 
 static inline int atchr(unsigned char c)
@@ -438,9 +436,9 @@ int decode_color(unsigned char *str, struct rgb *col)
 				ch = cs->rgb;
 				goto found;
 			}
-		str--;
+	} else {
+		str++;
 	}
-	str++;
 	if (strlen(str) == 6) {
 		char *end;
 		ch = strtoul(str, &end, 16);
@@ -486,14 +484,14 @@ unsigned char *get_target(unsigned char *a)
 
 void kill_html_stack_item(struct html_element *e)
 {
-	if (e->dontkill == 2) {
-		internal("trying to kill unkillable element");
-		return;
-	}
 	if (!e || (void *)e == &html_stack) {
 		internal("trying to free bad html element");
 		return;
 	}
+	if (e->dontkill == 2) {
+		internal("trying to kill unkillable element");
+		return;
+	}
 	if (e->attr.link) mem_free(e->attr.link);
 	if (e->attr.target) mem_free(e->attr.target);
 	if (e->attr.image) mem_free(e->attr.image);
@@ -507,12 +505,6 @@ void kill_html_stack_item(struct html_element *e)
 	}*/
 }
 
-static inline void kill_elem(char *e)
-{
-	if ((size_t)html_top.namelen == strlen(e) && !casecmp(html_top.name, e, html_top.namelen))
-		kill_html_stack_item(&html_top);
-}
-
 #ifdef DEBUG
 void debug_stack()
 {
@@ -649,6 +641,11 @@ void kill_until(int ls, ...)
 	ln_break(l, line_break_f, ff);
 }
 
+static inline unsigned char *top_href_base(void)
+{
+	return ((struct html_element *)html_stack.prev)->attr.href_base;
+}
+
 int get_num(unsigned char *a, unsigned char *n)
 {
 	char *al;
@@ -696,10 +693,6 @@ int get_width(unsigned char *a, unsigned char *n, int trunc)
 	return r;
 }
 
-/*int form_num;
-struct form form = { 0, NULL, 0 };
-int g_ctrl_num;*/
-
 struct form form = { NULL, NULL, 0, 0 };
 
 unsigned char *last_form_tag;
@@ -710,23 +703,23 @@ void put_link_line(unsigned char *prefix, unsigned char *linkname, unsigned char
 {
 	html_stack_dup();
 	ln_break(1, line_break_f, ff);
-	if (format.link) mem_free(format.link), format.link = NULL;
-	if (format.target) mem_free(format.target), format.target = NULL;
-	format.form = NULL;
+	if (format_.link) mem_free(format_.link), format_.link = NULL;
+	if (format_.target) mem_free(format_.target), format_.target = NULL;
+	format_.form = NULL;
 	put_chrs(prefix, strlen(prefix), put_chars_f, ff);
-	format.link = join_urls(format.href_base, link);
-	format.target = stracpy(target);
-	memcpy(&format.fg, &format.clink, sizeof(struct rgb));
+	format_.link = join_urls(format_.href_base, link);
+	format_.target = stracpy(target);
+	memcpy(&format_.fg, &format_.clink, sizeof(struct rgb));
 	put_chrs(linkname, strlen(linkname), put_chars_f, ff);
 	ln_break(1, line_break_f, ff);
 	kill_html_stack_item(&html_top);
 }
 
 void html_span(unsigned char *a) { }
-void html_bold(unsigned char *a) { format.attr |= AT_BOLD; }
-void html_italic(unsigned char *a) { format.attr |= AT_ITALIC; }
-void html_underline(unsigned char *a) { format.attr |= AT_UNDERLINE; }
-void html_fixed(unsigned char *a) { format.attr |= AT_FIXED; }
+void html_bold(unsigned char *a) { format_.attr |= AT_BOLD; }
+void html_italic(unsigned char *a) { format_.attr |= AT_ITALIC; }
+void html_underline(unsigned char *a) { format_.attr |= AT_UNDERLINE; }
+void html_fixed(unsigned char *a) { format_.attr |= AT_FIXED; }
 
 void html_a(unsigned char *a)
 {
@@ -735,18 +728,18 @@ void html_a(unsigned char *a)
 		char *all = al;
 		while (all[0] == ' ') all++;
 		while (all[0] && all[strlen(all) - 1] == ' ') all[strlen(all) - 1] = 0;
-		if (format.link) mem_free(format.link);
-		format.link = join_urls(format.href_base, all);
+		if (format_.link) mem_free(format_.link);
+		format_.link = join_urls(format_.href_base, all);
 		mem_free(al);
 		if ((al = get_target(a))) {
-			if (format.target) mem_free(format.target);
-			format.target = al;
+			if (format_.target) mem_free(format_.target);
+			format_.target = al;
 		} else {
-			if (format.target) mem_free(format.target);
-			format.target = stracpy(format.target_base);
+			if (format_.target) mem_free(format_.target);
+			format_.target = stracpy(format_.target_base);
 		}
-		/*format.attr ^= AT_BOLD;*/
-		memcpy(&format.fg, &format.clink, sizeof(struct rgb));
+		/*format_.attr ^= AT_BOLD;*/
+		memcpy(&format_.fg, &format_.clink, sizeof(struct rgb));
 	} else kill_html_stack_item(&html_top);
 	if ((al = get_attr_val(a, "name"))) {
 		special_f(ff, SP_TAG, al);
@@ -767,14 +760,14 @@ void html_font(unsigned char *a)
 		s = strtoul(nn, &end, 10);
 		if (*nn && !*end) {
 			if (s > 7) s = 7;
-			if (!p) format.fontsize = s;
-			else format.fontsize += p * s;
-			if (format.fontsize < 1) format.fontsize = 1;
-			if (format.fontsize > 7) format.fontsize = 7;
+			if (!p) format_.fontsize = s;
+			else format_.fontsize += p * s;
+			if (format_.fontsize < 1) format_.fontsize = 1;
+			if (format_.fontsize > 7) format_.fontsize = 7;
 		}
 		mem_free(al);
 	}
-	get_color(a, "color", &format.fg);
+	get_color(a, "color", &format_.fg);
 }
 
 void html_img(unsigned char *a)
@@ -786,43 +779,43 @@ void html_img(unsigned char *a)
 		unsigned char *u;
 		usemap = 1;
 		html_stack_dup();
-		if (format.link) mem_free(format.link);
-		if (format.form) format.form = NULL;
-		u = join_urls(format.href_base, al);
-		format.link = mem_alloc(strlen(u) + 5);
-		strcpy(format.link, "MAP@");
-		strcat(format.link, u);
-		format.attr |= AT_BOLD;
+		if (format_.link) mem_free(format_.link);
+		if (format_.form) format_.form = NULL;
+		u = join_urls(*al == '#' ? top_href_base() : format_.href_base, al);
+		format_.link = mem_alloc(strlen(u) + 5);
+		strcpy(format_.link, "MAP@");
+		strcat(format_.link, u);
+		format_.attr |= AT_BOLD;
 		mem_free(u);
 		mem_free(al);
 	}
-	ismap = format.link && has_attr(a, "ismap") && !usemap;
+	ismap = format_.link && has_attr(a, "ismap") && !usemap;
 	if ((!(al = get_attr_val(a, "alt")) && !(al = get_attr_val(a, "title"))) || !*al) {
 		if (al) mem_free(al);
-		if (!d_opt->images && !format.link) return;
+		if (!d_opt->images && !format_.link) return;
 		if (usemap) al = stracpy("[USEMAP]");
 		else if (ismap) al = stracpy("[ISMAP]");
 		else al = stracpy("[IMG]");
 	}
-	if (format.image) mem_free(format.image), format.image = NULL;
+	if (format_.image) mem_free(format_.image), format_.image = NULL;
 	if (al) {
 		unsigned char *s;
 		if ((s = get_url_val(a, "src")) || (s = get_attr_val(a, "dynsrc"))) {
-			format.image = join_urls(format.href_base, s);
+			format_.image = join_urls(format_.href_base, s);
 			mem_free(s);
 		}
 		if (ismap) {
 			unsigned char *h;
 			html_stack_dup();
-			h = stracpy(format.link);
+			h = stracpy(format_.link);
 			add_to_strn(&h, "?0,0");
-			mem_free(format.link);
-			format.link = h;
+			mem_free(format_.link);
+			format_.link = h;
 		}
 		put_chrs(al, strlen(al), put_chars_f, ff);
 		if (ismap) kill_html_stack_item(&html_top);
 	}
-	if (format.image) mem_free(format.image), format.image = NULL;
+	if (format_.image) mem_free(format_.image), format_.image = NULL;
 	mem_free(al);
 	if (usemap) kill_html_stack_item(&html_top);
 	/*put_chrs(" ", 1, put_chars_f, ff);*/
@@ -830,10 +823,10 @@ void html_img(unsigned char *a)
 
 void html_body(unsigned char *a)
 {
-	get_color(a, "text", &format.fg);
-	get_color(a, "link", &format.clink);
-	get_color(a, "vlink", &format.vlink);
-	get_bgcolor(a, &format.bg);
+	get_color(a, "text", &format_.fg);
+	get_color(a, "link", &format_.clink);
+	get_color(a, "vlink", &format_.vlink);
+	get_bgcolor(a, &format_.bg);
 	get_bgcolor(a, &par_format.bgcolor);
 }
 
@@ -896,6 +889,7 @@ void html_blockquote(unsigned char *a)
 
 void html_h(int h, char *a)
 {
+	do_not_optimize_here(&h);
 	par_format.align = AL_LEFT;
 	html_linebrk(a);
 	switch (par_format.align) {
@@ -937,14 +931,14 @@ void html_hr(unsigned char *a)
 	if (q >= 0 && q < 2) r = 196;
 	html_stack_dup();
 	par_format.align = AL_CENTER;
-	if (format.link) mem_free(format.link), format.link = NULL;
-	format.form = NULL;
+	if (format_.link) mem_free(format_.link), format_.link = NULL;
+	format_.form = NULL;
 	html_linebrk(a);
 	if (par_format.align == AL_BLOCK) par_format.align = AL_CENTER;
 	par_format.leftmargin = margin;
 	par_format.rightmargin = margin;
 	if ((i = get_width(a, "width", 1)) == -1) i = par_format.width - 2 * margin - 4;
-	format.attr = AT_GRAPHICS;
+	format_.attr = AT_GRAPHICS;
 	special_f(ff, SP_NOWRAP, 1);
 	while (i-- > 0) put_chrs(&r, 1, put_chars_f, ff);
 	special_f(ff, SP_NOWRAP, 0);
@@ -958,7 +952,7 @@ void html_table(unsigned char *a)
 	par_format.rightmargin = margin;
 	par_format.align = AL_LEFT;
 	html_linebrk(a);
-	format.attr = 0;
+	format_.attr = 0;
 }
 
 void html_tr(unsigned char *a)
@@ -970,7 +964,7 @@ void html_th(unsigned char *a)
 {
 	/*html_linebrk(a);*/
 	kill_until(1, "TD", "TH", "", "TR", "TABLE", NULL);
-	format.attr |= AT_BOLD;
+	format_.attr |= AT_BOLD;
 	put_chrs(" ", 1, put_chars_f, ff);
 }
 
@@ -978,7 +972,7 @@ void html_td(unsigned char *a)
 {
 	/*html_linebrk(a);*/
 	kill_until(1, "TD", "TH", "", "TR", "TABLE", NULL);
-	format.attr &= ~AT_BOLD;
+	format_.attr &= ~AT_BOLD;
 	put_chrs(" ", 1, put_chars_f, ff);
 }
 
@@ -986,13 +980,13 @@ void html_base(unsigned char *a)
 {
 	char *al;
 	if ((al = get_url_val(a, "href"))) {
-		if (format.href_base) mem_free(format.href_base);
-		format.href_base = join_urls(((struct html_element *)html_stack.prev)->attr.href_base, al);
+		if (format_.href_base) mem_free(format_.href_base);
+		format_.href_base = join_urls(top_href_base(), al);
 		mem_free(al);
 	}
 	if ((al = get_target(a))) {
-		if (format.target_base) mem_free(format.target_base);
-		format.target_base = al;
+		if (format_.target_base) mem_free(format_.target_base);
+		format_.target_base = al;
 	}
 }
 
@@ -1132,16 +1126,16 @@ void get_html_form(unsigned char *a, struct form *form)
 		char *all = al;
 		while (all[0] == ' ') all++;
 		while (all[0] && all[strlen(all) - 1] == ' ') all[strlen(all) - 1] = 0;
-		form->action = join_urls(format.href_base, all);
+		form->action = join_urls(format_.href_base, all);
 		mem_free(al);
 	} else {
-		if ((ch = strchr(form->action = stracpy(format.href_base), POST_CHAR))) *ch = 0;
+		if ((ch = strchr(form->action = stracpy(format_.href_base), POST_CHAR))) *ch = 0;
 		if (form->method == FM_GET && (ch = strchr(form->action, '?'))) *ch = 0;
 	}
 	if ((al = get_target(a))) {
 		form->target = al;
 	} else {
-		form->target = stracpy(format.target_base);
+		form->target = stracpy(format_.target_base);
 	}
 	form->num = a - startf;
 }
@@ -1170,7 +1164,7 @@ void find_form_for_input(unsigned char *i, int when_unused)
 	se:
 	while (s < i && *s != '<') sp:s++;
 	if (s >= i) goto end_parse;
-	if (s + 2 <= eofff && (s[1] == '!' || s[1] == '?')) {
+	if (eofff - s >= 2 && (s[1] == '!' || s[1] == '?')) {
 		s = skip_comment(s, i);
 		goto se;
 	}
@@ -1235,17 +1229,32 @@ void html_button(unsigned char *a)
 	if (fc->type == FC_RESET && !fc->default_value) fc->default_value = stracpy("Reset");
 	if (!fc->default_value) fc->default_value = stracpy("");
 	special_f(ff, SP_CONTROL, fc);
-	format.form = fc;
-	format.attr |= AT_BOLD;
+	format_.form = fc;
+	format_.attr |= AT_BOLD;
 	/*put_chrs("[&nbsp;", 7, put_chars_f, ff);
 	if (fc->default_value) put_chrs(fc->default_value, strlen(fc->default_value), put_chars_f, ff);
 	put_chrs("&nbsp;]", 7, put_chars_f, ff);
 	put_chrs(" ", 1, put_chars_f, ff);*/
 }
 
+void set_max_textarea_width(int *w)
+{
+	int limit;
+	if (!table_level) {
+		limit = par_format.width - (par_format.leftmargin + par_format.rightmargin);
+	} else {
+		limit = d_opt->xw - 2;
+	}
+	if (*w > limit) {
+		*w = limit;
+		if (*w < HTML_MINIMAL_TEXTAREA_WIDTH) *w = HTML_MINIMAL_TEXTAREA_WIDTH;
+	}
+}
+
 void html_input(unsigned char *a)
 {
 	int i;
+	int size;
 	unsigned char *al;
 	struct form_control *fc;
 	find_form_for_input(a, 0);
@@ -1287,9 +1296,12 @@ void html_input(unsigned char *a)
 	if (fc->type == FC_TEXT || fc->type == FC_PASSWORD) fc->default_value = get_attr_val(a, "value");
 	else if (fc->type != FC_FILE) fc->default_value = get_exact_attr_val(a, "value");
 	if (fc->type == FC_CHECKBOX && !fc->default_value) fc->default_value = stracpy("on");
-	if ((fc->size = get_num(a, "size")) == -1) fc->size = HTML_DEFAULT_INPUT_SIZE;
-	fc->size++;
-	if (fc->size > d_opt->xw) fc->size = d_opt->xw;
+	if ((size = get_num(a, "size")) <= 0) size = HTML_DEFAULT_INPUT_SIZE;
+	size++;
+	if (size > HTML_MINIMAL_TEXTAREA_WIDTH) {
+		set_max_textarea_width(&size);
+	}
+	fc->size = size;
 	if ((fc->maxlength = get_num(a, "maxlength")) == -1) fc->maxlength = MAXINT / 4;
 	if (fc->type == FC_CHECKBOX || fc->type == FC_RADIO) fc->default_state = has_attr(a, "checked");
 	fc->ro = has_attr(a, "disabled") ? 2 : has_attr(a, "readonly") ? 1 : 0;
@@ -1304,26 +1316,26 @@ void html_input(unsigned char *a)
 	if (fc->type == FC_HIDDEN) goto hid;
 	put_chrs(" ", 1, put_chars_f, ff);
 	html_stack_dup();
-	format.form = fc;
+	format_.form = fc;
 	switch (fc->type) {
 		case FC_TEXT:
 		case FC_PASSWORD:
 		case FC_FILE:
-			format.attr |= AT_BOLD;
+			format_.attr |= AT_BOLD;
 			for (i = 0; i < fc->size; i++) put_chrs("_", 1, put_chars_f, ff);
 			break;
 		case FC_CHECKBOX:
 		case FC_RADIO:
-			format.attr |= AT_BOLD;
+			format_.attr |= AT_BOLD;
 			put_chrs("[&nbsp;]", 8, put_chars_f, ff);
 			break;
 		case FC_IMAGE:
-			if (format.image) mem_free(format.image), format.image = NULL;
+			if (format_.image) mem_free(format_.image), format_.image = NULL;
 			if ((al = get_url_val(a, "src")) || (al = get_url_val(a, "dynsrc"))) {
-				format.image = join_urls(format.href_base, al);
+				format_.image = join_urls(format_.href_base, al);
 				mem_free(al);
 			}
-			format.attr |= AT_BOLD;
+			format_.attr |= AT_BOLD;
 			put_chrs("[&nbsp;", 7, put_chars_f, ff);
 			if (fc->alt) put_chrs(fc->alt, strlen(fc->alt), put_chars_f, ff);
 			else put_chrs("Submit", 6, put_chars_f, ff);
@@ -1331,7 +1343,7 @@ void html_input(unsigned char *a)
 			break;
 		case FC_SUBMIT:
 		case FC_RESET:
-			format.attr |= AT_BOLD;
+			format_.attr |= AT_BOLD;
 			put_chrs("[&nbsp;", 7, put_chars_f, ff);
 			if (fc->default_value) put_chrs(fc->default_value, strlen(fc->default_value), put_chars_f, ff);
 			put_chrs("&nbsp;]", 7, put_chars_f, ff);
@@ -1351,9 +1363,9 @@ void html_select(unsigned char *a)
 	char *al;
 	if (!(al = get_attr_val(a, "name"))) return;
 	html_top.dontkill = 1;
-	if (format.select) mem_free(format.select);
-	format.select = al;
-	format.select_disabled = 2 * has_attr(a, "disabled");
+	if (format_.select) mem_free(format_.select);
+	format_.select = al;
+	format_.select_disabled = 2 * has_attr(a, "disabled");
 }
 
 void html_option(unsigned char *a)
@@ -1361,7 +1373,7 @@ void html_option(unsigned char *a)
 	struct form_control *fc;
 	unsigned char *val;
 	find_form_for_input(a, 0);
-	if (!format.select) return;
+	if (!format_.select) return;
 	fc = mem_alloc(sizeof(struct form_control));
 	memset(fc, 0, sizeof(struct form_control));
 	if (!(val = get_exact_attr_val(a, "value"))) {
@@ -1384,7 +1396,7 @@ void html_option(unsigned char *a)
 		r = p;
 		while (r < eoff && WHITECHAR(*r)) r++;
 		if (r >= eoff) goto x;
-		if (r - 2 <= eoff && (r[1] == '!' || r[1] == '?')) {
+		if (eoff - r >= 2 && (r[1] == '!' || r[1] == '?')) {
 			p = skip_comment(r, eoff);
 			goto rrrr;
 		}
@@ -1403,15 +1415,15 @@ void html_option(unsigned char *a)
 	fc->method = form.method;
 	fc->action = stracpy(form.action);
 	fc->type = FC_CHECKBOX;
-	fc->name = stracpy(format.select);
+	fc->name = stracpy(format_.select);
 	fc->default_value = val;
 	fc->default_state = has_attr(a, "selected");
-	fc->ro = format.select_disabled;
+	fc->ro = format_.select_disabled;
 	if (has_attr(a, "disabled")) fc->ro = 2;
 	put_chrs(" ", 1, put_chars_f, ff);
 	html_stack_dup();
-	format.form = fc;
-	format.attr |= AT_BOLD;
+	format_.form = fc;
+	format_.attr |= AT_BOLD;
 	put_chrs("[ ]", 3, put_chars_f, ff);
 	kill_html_stack_item(&html_top);
 	put_chrs(" ", 1, put_chars_f, ff);
@@ -1466,7 +1478,7 @@ void new_menu_item(unsigned char *name, int data, int fullname)
 		item->rtext = data == -1 ? ">" : "";
 		item->hotkey = fullname ? "\000\001" : "\000\000"; /* dirty */
 		item->func = data == -1 ? MENU_FUNC do_select_submenu : MENU_FUNC selected_item;
-		item->data = data == -1 ? nmenu : (void *)data;
+		item->data = data == -1 ? nmenu : (void *)(my_uintptr_t)data;
 		item->in_m = data == -1 ? 1 : 0;
 		item->free_i = 0;
 		item++;
@@ -1525,14 +1537,14 @@ void menu_labels(struct menu_item *m, unsigned char *base, unsigned char **lbls)
 			}
 		} else {
 			if ((bs = stracpy(m->hotkey[1] ? (unsigned char *)"" : base))) add_to_strn(&bs, m->text);
-			lbls[(int)m->data] = bs;
+			lbls[(int)(my_uintptr_t)m->data] = bs;
 		}
 	}
 }
 
 int menu_contains(struct menu_item *m, int f)
 {
-	if (m->func != MENU_FUNC do_select_submenu) return (int)m->data == f;
+	if (m->func != MENU_FUNC do_select_submenu) return (int)(my_uintptr_t)m->data == f;
 	for (m = m->data; m->text; m++) if (menu_contains(m, f)) return 1;
 	return 0;
 }
@@ -1546,7 +1558,7 @@ void do_select_submenu(struct terminal *term, struct menu_item *menu, struct ses
 	for (m = menu; m->text; m++, sel++) if (menu_contains(m, def)) goto f;
 	sel = 0;
 	f:
-	do_menu_selected(term, menu, ses, sel);
+	do_menu_selected(term, menu, ses, sel, NULL, NULL);
 }
 
 int do_html_select(unsigned char *attr, unsigned char *html, unsigned char *eof, unsigned char **end, void *f)
@@ -1572,10 +1584,10 @@ int do_html_select(unsigned char *attr, unsigned char *html, unsigned char *eof,
 	val = DUMMY;
 	order = 0, group = 0, preselect = -1;
 	init_menu();
-        se:
-        en = html;
-        see:
-        html = en;
+	se:
+	en = html;
+	see:
+	html = en;
 	while (html < eof && *html != '<') html++;
 	if (html >= eof) {
 		int i;
@@ -1598,7 +1610,7 @@ int do_html_select(unsigned char *attr, unsigned char *html, unsigned char *eof,
 		if (q) add_to_str(&lbl, &lbl_l, q), mem_free(q);
 		add_bytes_to_str(&vlbl, &vlbl_l, s, l);
 	}
-	if (html + 2 <= eof && (html[1] == '!' || html[1] == '?')) {
+	if (eof - html >= 2 && (html[1] == '!' || html[1] == '?')) {
 		html = skip_comment(html, eof);
 		goto se;
 	}
@@ -1694,8 +1706,8 @@ int do_html_select(unsigned char *attr, unsigned char *html, unsigned char *eof,
 	menu_labels(fc->menu, "", lbls);
 	put_chrs("[", 1, put_chars_f, f);
 	html_stack_dup();
-	format.form = fc;
-	format.attr |= AT_BOLD;
+	format_.form = fc;
+	format_.attr |= AT_BOLD;
 	mw = 0;
 	for (i = 0; i < order; i++) if (lbls[i] && strlen(lbls[i]) > (size_t)mw) mw = strlen(lbls[i]);
 	for (i = 0; i < mw; i++) put_chrs("_", 1, put_chars_f, f);
@@ -1742,11 +1754,14 @@ void do_html_textarea(unsigned char *attr, unsigned char *html, unsigned char *e
 	fc->type = FC_TEXTAREA;;
 	fc->ro = has_attr(attr, "disabled") ? 2 : has_attr(attr, "readonly") ? 1 : 0;
 	fc->default_value = memacpy(html, p - html);
-	if ((cols = get_num(attr, "cols")) <= 5) cols = HTML_DEFAULT_TEXTAREA_WIDTH;
+	if ((cols = get_num(attr, "cols")) < HTML_MINIMAL_TEXTAREA_WIDTH) cols = HTML_DEFAULT_TEXTAREA_WIDTH;
 	cols++;
+	set_max_textarea_width(&cols);
 	if ((rows = get_num(attr, "rows")) <= 0) rows = HTML_DEFAULT_TEXTAREA_HEIGHT;
-	if (cols > d_opt->xw) cols = d_opt->xw;
-	if (rows > d_opt->yw) rows = d_opt->yw;
+	if (rows > d_opt->yw) {
+		rows = d_opt->yw;
+		if (rows <= 0) rows = 1;
+	}
 	fc->cols = cols;
 	fc->rows = rows;
 	fc->wrap = 1;
@@ -1759,8 +1774,8 @@ void do_html_textarea(unsigned char *attr, unsigned char *html, unsigned char *e
 	if (rows > 1) ln_break(1, line_break_f, f);
 	else put_chrs(" ", 1, put_chars_f, f);
 	html_stack_dup();
-	format.form = fc;
-	format.attr |= AT_BOLD;
+	format_.form = fc;
+	format_.attr |= AT_BOLD;
 	for (i = 0; i < rows; i++) {
 		int j;
 		for (j = 0; j < cols; j++) put_chrs("_", 1, put_chars_f, f);
@@ -1794,7 +1809,7 @@ void html_frame(unsigned char *a)
 	if (!(u2 = get_url_val(a, "src"))) {
 		url = stracpy("");
 	} else {
-		url = join_urls(format.href_base, u2);
+		url = join_urls(format_.href_base, u2);
 		mem_free(u2);
 	}
 	if (!url) return;
@@ -1984,24 +1999,44 @@ void html_link(unsigned char *a)
 	if (!(url = get_url_val(a, "href"))) return;
 	if (!(name = get_attr_val(a, "rel")))
 		if (!(name = get_attr_val(a, "rev")))
-			if (!(name = get_attr_val(a, "ref")))
-				name = stracpy(url);
-	if (!strcasecmp(name, "stylesheet") ||
-	    !strcasecmp(name, "alternate stylesheet") ||
-	    !strcasecmp(name, "made") ||
-	    !strcasecmp(name, "icon") ||
-	    !strcasecmp(name, "shortcut icon") ||
-	    !strcasecmp(name, "apple-touch-icon") ||
-	    !strcasecmp(name, "meta") ||
-	    !strcasecmp(name, "pingback") ||
-	    !strcasecmp(name, "File-List") ||
-	    !casecmp(name, "schema", 6)) goto skip;
+			name = get_attr_val(a, "ref");
+	if (name) {
+		unsigned char *lang;
+		if ((lang = get_attr_val(a, "hreflang"))) {
+			add_to_strn(&name, " ");
+			add_to_strn(&name, lang);
+			mem_free(lang);
+		}
+	}
+	if (!name)
+		name = stracpy(url);
+	if (
+	    !casecmp(name, cast_uchar "apple-touch-icon", 16) ||
+	    !casecmp(name, cast_uchar "schema", 6) ||
+	    !strcasecmp(cast_const_char name, "Edit-Time-Data") ||
+	    !strcasecmp(cast_const_char name, "File-List") ||
+	    !strcasecmp(cast_const_char name, "alternate stylesheet") ||
+	    !strcasecmp(cast_const_char name, "generator-home") ||
+	    !strcasecmp(cast_const_char name, "https://github.com/WP-API/WP-API") ||
+	    !strcasecmp(cast_const_char name, "icon") ||
+	    !strcasecmp(cast_const_char name, "made") ||
+	    !strcasecmp(cast_const_char name, "meta") ||
+	    !strcasecmp(cast_const_char name, "pingback") ||
+	    !strcasecmp(cast_const_char name, "preconnect") ||
+	    !strcasecmp(cast_const_char name, "shortcut icon") ||
+	    !strcasecmp(cast_const_char name, "stylesheet") ||
+	    !strcasecmp(cast_const_char name, "https://api.w.org/") ||
+	    !strcasecmp(cast_const_char name, "prefetch") ||
+	    !strcasecmp(cast_const_char name, "dns-prefetch") ||
+	    !strcasecmp(cast_const_char name, "prerender") ||
+	    !strcasecmp(cast_const_char name, "preload") ||
+	    0) goto skip;
 	if ((title = get_attr_val(a, "title"))) {
 		add_to_strn(&name, ": ");
 		add_to_strn(&name, title);
 		mem_free(title);
 	}
-	put_link_line("Link: ", name, url, format.target_base);
+	put_link_line("Link: ", name, url, format_.target_base);
 	skip:
 	mem_free(name);
 	mem_free(url);
@@ -2092,11 +2127,11 @@ struct element_info elements[] = {
 
 unsigned char *skip_comment(unsigned char *html, unsigned char *eof)
 {
-	int comm = html + 4 <= eof && html[2] == '-' && html[3] == '-';
+	int comm = eof - html >= 4 && html[2] == '-' && html[3] == '-';
 	html += comm ? 4 : 2;
 	while (html < eof) {
 		if (!comm && html[0] == '>') return html + 1;
-		if (comm && html + 2 <= eof && html[0] == '-' && html[1] == '-') {
+		if (comm && eof - html >= 2 && html[0] == '-' && html[1] == '-') {
 			html += 2;
 			while (html < eof && *html == '-') html++;
 			while (html < eof && WHITECHAR(*html)) html++;
@@ -2132,7 +2167,7 @@ int qd(unsigned char *html, unsigned char *eof, int *len)
 	if (html[0] != '&' || d_opt->plain) return html[0];
 	if (html + 1 >= eof) return -1;
 	if (html[1] != '#') return -1;
-	for (l = 2; l < 10 && html + l < eof; l++) if (html[l] == ';') {
+	for (l = 2; l < 10 && eof - html > l; l++) if (html[l] == ';') {
 		int n = get_entity_number(html + 2, l - 2);
 		if (n >= 0) {
 			*len = l + 1;
@@ -2177,7 +2212,7 @@ void parse_html(unsigned char *html, unsigned char *eof, int (*put_chars)(void *
 			}
 			putsp = 0;*/
 			while (h < eof && WHITECHAR(*h)) h++;
-			if (h + 1 < eof && h[0] == '<' && h[1] == '/') {
+			if (eof - h > 1 && h[0] == '<' && h[1] == '/') {
 				if (!parse_element(h, eof, &name, &namelen, &attr, &end)) {
 					put_chrs(lt, html - lt, put_chars, f);
 					lt = html = h;
@@ -2215,7 +2250,7 @@ void parse_html(unsigned char *html, unsigned char *eof, int (*put_chars)(void *
 				put_chrs(lt, html - lt, put_chars, f);
 				next_break:
 				html += l;
-				if (q == 13 && html < eof - 1 && qd(html, eof, &l) == 10) html += l;
+				if (q == 13 && eof - html > 1 && qd(html, eof, &l) == 10) html += l;
 				ln_break(1, line_break, f);
 				if (html >= eof) goto set_lt;
 				q = qd(html, eof, &l);
@@ -2234,7 +2269,7 @@ void parse_html(unsigned char *html, unsigned char *eof, int (*put_chars)(void *
 			html++;
 			goto set_lt;
 		}
-		if (html + 2 <= eof && html[0] == '<' && (html[1] == '!' || html[1] == '?') && !d_opt->plain) {
+		if (eof - html >= 2 && html[0] == '<' && (html[1] == '!' || html[1] == '?') && !d_opt->plain) {
 			/*if (putsp == 1) goto put_sp;
 			putsp = 0;*/
 			put_chrs(lt, html - lt, put_chars, f);
@@ -2383,7 +2418,7 @@ int get_image_map(unsigned char *head, unsigned char *s, unsigned char *eof, uns
 		mem_free(*menu);
 		return -1;
 	}
-	if (s + 2 <= eof && (s[1] == '!' || s[1] == '?')) {
+	if (eof - s >= 2 && (s[1] == '!' || s[1] == '?')) {
 		s = skip_comment(s, eof);
 		goto se;
 	}
@@ -2408,7 +2443,7 @@ int get_image_map(unsigned char *head, unsigned char *s, unsigned char *eof, uns
 		mem_free(*menu);
 		return -1;
 	}
-	if (s + 2 <= eof && (s[1] == '!' || s[1] == '?')) {
+	if (eof - s >= 2 && (s[1] == '!' || s[1] == '?')) {
 		s = skip_comment(s, eof);
 		goto se2;
 	}
@@ -2419,6 +2454,7 @@ int get_image_map(unsigned char *head, unsigned char *s, unsigned char *eof, uns
 		lblen = 0;
 		se3:
 		ss = s;
+		se4:
 		while (ss < eof && *ss != '<') ss++;
 		if (ss >= eof) {
 			mem_free(label);
@@ -2428,11 +2464,14 @@ int get_image_map(unsigned char *head, unsigned char *s, unsigned char *eof, uns
 		}
 		add_bytes_to_str(&label, &lblen, s, ss - s);
 		s = ss;
-		if (s + 2 <= eof && (s[1] == '!' || s[1] == '?')) {
+		if (eof - s >= 2 && (s[1] == '!' || s[1] == '?')) {
 			s = skip_comment(s, eof);
 			goto se3;
 		}
-		if (parse_element(s, eof, NULL, NULL, NULL, &ss)) goto se3;
+		if (parse_element(s, eof, NULL, NULL, NULL, &ss)) {
+			ss = s + 1;
+			goto se4;
+		}
 		if (!((namelen == 1 && !casecmp(name, "A", 1)) ||
 		      (namelen == 2 && !casecmp(name, "/A", 2)) ||
 		      (namelen == 3 && !casecmp(name, "MAP", 3)) ||
@@ -2503,7 +2542,7 @@ void scan_http_equiv(unsigned char *s, unsigned char *eof, unsigned char **head,
 	se:
 	while (s < eof && *s != '<') sp:s++;
 	if (s >= eof) return;
-	if (s + 2 <= eof && (s[1] == '!' || s[1] == '?')) {
+	if (eof - s >= 2 && (s[1] == '!' || s[1] == '?')) {
 		s = skip_comment(s, eof);
 		goto se;
 	}
@@ -2517,7 +2556,7 @@ void scan_http_equiv(unsigned char *s, unsigned char *eof, unsigned char **head,
 		while (s < eof && *s != '<') xsp:s++;
 		add_bytes_to_str(title, &tlen, s1, s - s1);
 		if (s >= eof) goto se;
-		if (s + 2 <= eof && (s[1] == '!' || s[1] == '?')) {
+		if (eof - s >= 2 && (s[1] == '!' || s[1] == '?')) {
 			s = skip_comment(s, eof);
 			goto xse;
 		}
@@ -2532,6 +2571,7 @@ void scan_http_equiv(unsigned char *s, unsigned char *eof, unsigned char **head,
 	if ((he = get_attr_val(attr, "charset"))) {
 		add_to_str(head, hdl, "Charset: ");
 		add_to_str(head, hdl, he);
+		add_to_str(head, hdl, "\r\n");
 		mem_free(he);
 	}
 	if (!(he = get_attr_val(attr, "http-equiv"))) goto se;
