@@ -1,7 +1,5 @@
 #include "links.h"
 
-#define format format_
-
 static inline int color_distance(struct rgb *c1, struct rgb *c2)
 {
 	return
@@ -169,7 +167,7 @@ static inline void xpand_line(struct part *p, int y, int x)
 			p->data->data[y].d = mem_realloc(p->data->data[y].d, XALIGN(x+1)*sizeof(chr));
 		}
 		for (i = p->data->data[y].l; i <= x; i++)
-			p->data->data[y].d[i] = (p->data->data[y].c << 11) | ' ';
+			p->data->data[y].d[i] = (p->data->data[y].c << 11) | 0x700 | ' ';
 		p->data->data[y].c = p->bgcolor;
 		p->data->data[y].l = i;
 	}
@@ -329,23 +327,11 @@ void line_break(struct part *);
 int split_line(struct part *p)
 {
 	int i;
-	/*if (!p->data) goto r;*/
-	/*printf("split: %d,%d   , %d,%d,%d\n",p->cx,p->cy,par_format.rightmargin,par_format.leftmargin,p->cx);*/
 	for (i = rm(par_format); i >= par_format.leftmargin; i--)
 		if (i < p->spl && p->spaces[i]) goto split;
-	/*for (i = p->cx - 1; i > rm(par_format) && i > par_format.leftmargin; i--)*/
 	for (i = par_format.leftmargin; i < p->cx ; i++)
 		if (i < p->spl && p->spaces[i]) goto split;
-	/*for (i = rm(par_format); i >= par_format.leftmargin; i--)
-		if ((POS(i, p->cy) & 0xff) == ' ') goto split;
-	for (i = p->cx - 1; i > rm(par_format) && i > par_format.leftmargin; i--)
-		if ((POS(i, p->cy) & 0xff) == ' ') goto split;*/
 	if (p->cx + par_format.rightmargin > p->x) p->x = p->cx + par_format.rightmargin;
-	/*if (p->y < p->cy + 1) p->y = p->cy + 1;
-	p->cy++; p->cx = -1;
-	memset(p->spaces, 0, p->spl);
-	if (p->data) xpand_lines(p, p->cy + 1);*/
-	/*line_break(p);*/
 	return 0;
 	split:
 	if (i + par_format.rightmargin > p->x) p->x = i + par_format.rightmargin;
@@ -360,7 +346,6 @@ int split_line(struct part *p)
 	memset(p->spaces + p->spl - i - 1, 0, i + 1);
 	memmove(p->spaces + par_format.leftmargin, p->spaces, p->spl - par_format.leftmargin);
 	p->cy++; p->cx -= i - par_format.leftmargin + 1;
-	/*return 1 + (p->cx == par_format.leftmargin);*/
 	if (p->cx == par_format.leftmargin) p->cx = -1;
 	if (p->y < p->cy + (p->cx != -1)) p->y = p->cy + (p->cx != -1);
 	return 1 + (p->cx == -1);
@@ -424,7 +409,7 @@ int put_chars_conv(struct part *p, unsigned char *c, int l)
 	int bp = 0;
 	int pp = 0;
 	int total = 0;
-	if (format.attr & AT_GRAPHICS) {
+	if (format_.attr & AT_GRAPHICS) {
 		put_chars(p, c, l);
 		return l;
 	}
@@ -499,16 +484,16 @@ void put_chars(struct part *p, unsigned char *c, int l)
 		last_tag_for_newline = (void *)&p->data->tags;
 	}
 	if (p->cx < par_format.leftmargin) p->cx = par_format.leftmargin;
-	if (last_link || last_image || last_form || format.link || format.image || format.form) goto process_link;
+	if (last_link || last_image || last_form || format_.link || format_.image || format_.form) goto process_link;
 	no_l:
 	/*printf("%d %d\n",p->cx, p->cy);*/
-	if (memcmp(&ta_cache, &format, sizeof(struct text_attrib_beginning))) goto format_change;
+	if (memcmp(&ta_cache, &format_, sizeof(struct text_attrib_beginning))) goto format_change;
 	bg = bg_cache, fg = fg_cache;
 	end_format_change:
 	if (p->cx == par_format.leftmargin && *c == ' ' && par_format.align != AL_NO) c++, l--;
 	if (p->y < p->cy + 1) p->y = p->cy + 1;
 	if (nowrap && p->cx + l > rm(par_format)) return;
-	set_hline(p, p->cx, p->cy, l, c, (((fg&0x08)<<3)|(bg<<3)|(fg&0x07))<<8, 1);
+	set_hline(p, p->cx, p->cy, l, c, (((fg&0x08)<<3)|(bg<<3)|(fg&0x07))<<8, par_format.align != AL_NO);
 	p->cx += l;
 	nobreak = 0;
 	if (par_format.align != AL_NO)
@@ -527,8 +512,8 @@ void put_chars(struct part *p, unsigned char *c, int l)
 	return;
 	process_link:
 	if ((last_link /*|| last_target*/ || last_image || last_form) &&
-	    !xstrcmp(format.link, last_link) && !xstrcmp(format.target, last_target) &&
-	    !xstrcmp(format.image, last_image) && format.form == last_form) {
+	    !xstrcmp(format_.link, last_link) && !xstrcmp(format_.target, last_target) &&
+	    !xstrcmp(format_.image, last_image) && format_.form == last_form) {
 		if (!p->data) goto x;
 		link = &p->data->links[p->data->nlinks - 1];
 		if (!p->data->nlinks) {
@@ -543,27 +528,27 @@ void put_chars(struct part *p, unsigned char *c, int l)
 		if (last_image) mem_free(last_image);
 		last_link = last_target = last_image = NULL;
 		last_form = NULL;
-		if (!(format.link || format.image || format.form)) goto no_l;
+		if (!(format_.link || format_.image || format_.form)) goto no_l;
 		if (d_opt->num_links) {
 			unsigned char s[64];
-			unsigned char *fl = format.link, *ft = format.target, *fi = format.image;
-			struct form_control *ff = format.form;
-			format.link = format.target = format.image = NULL;
-			format.form = NULL;
+			unsigned char *fl = format_.link, *ft = format_.target, *fi = format_.image;
+			struct form_control *ff = format_.form;
+			format_.link = format_.target = format_.image = NULL;
+			format_.form = NULL;
 			s[0] = '[';
 			snzprint(s + 1, 62, p->link_num);
 			strcat(s, "]");
 			put_chars(p, s, strlen(s));
 			if (ff && ff->type == FC_TEXTAREA) line_break(p);
 			if (p->cx == -1) p->cx = par_format.leftmargin;
-			format.link = fl, format.target = ft, format.image = fi;
-			format.form = ff;
+			format_.link = fl, format_.target = ft, format_.image = fi;
+			format_.form = ff;
 		}
 		p->link_num++;
-		last_link = stracpy(format.link);
-		last_target = stracpy(format.target);
-		last_image = stracpy(format.image);
-		last_form = format.form;
+		last_link = stracpy(format_.link);
+		last_target = stracpy(format_.target);
+		last_image = stracpy(format_.image);
+		last_form = format_.form;
 		if (!p->data) goto no_l;
 		if (!(link = new_link(p->data))) goto no_l;
 		link->num = p->link_num - 1;
@@ -579,12 +564,12 @@ void put_chars(struct part *p, unsigned char *c, int l)
 		}
 		link->where_img = stracpy(last_image);
 		if (link->type != L_FIELD && link->type != L_AREA) {
-			bg = find_nearest_color(&format.clink, 8);
-			fg = find_nearest_color(&format.bg, 8);
+			bg = find_nearest_color(&format_.clink, 8);
+			fg = find_nearest_color(&format_.bg, 8);
 			fg = fg_color(fg, bg);
 		} else {
-			fg = find_nearest_color(&format.fg, 8);
-			bg = find_nearest_color(&format.bg, 8);
+			fg = find_nearest_color(&format_.fg, 8);
+			bg = find_nearest_color(&format_.bg, 8);
 			fg = fg_color(fg, bg);
 		}
 		link->sel_color = ((fg & 8) << 3) | (fg & 7) | (bg << 3);
@@ -600,15 +585,15 @@ void put_chars(struct part *p, unsigned char *c, int l)
 	goto no_l;
 
 		format_change:
-		bg = find_nearest_color(&format.bg, 8);
-		fg = find_nearest_color(&format.fg, 16);
+		bg = find_nearest_color(&format_.bg, 8);
+		fg = find_nearest_color(&format_.fg, 16);
 		fg = fg_color(fg, bg);
-		if (format.attr & AT_ITALIC) fg = fg ^ 0x01;
-		if (format.attr & AT_UNDERLINE) fg = (fg ^ 0x04) | 0x08;
-		if (format.attr & AT_BOLD) fg = fg | 0x08;
+		if (format_.attr & AT_ITALIC) fg = fg ^ 0x01;
+		if (format_.attr & AT_UNDERLINE) fg = (fg ^ 0x04) | 0x08;
+		if (format_.attr & AT_BOLD) fg = fg | 0x08;
 		fg = fg_color(fg, bg);
-		if (format.attr & AT_GRAPHICS) bg = bg | 0x10;
-		memcpy(&ta_cache, &format, sizeof(struct text_attrib_beginning));
+		if (format_.attr & AT_GRAPHICS) bg = bg | 0x10;
+		memcpy(&ta_cache, &format_, sizeof(struct text_attrib_beginning));
 		fg_cache = fg; bg_cache = bg;
 		goto end_format_change;
 }
@@ -637,6 +622,7 @@ void line_break(struct part *p)
 	}
 	e:
 	p->cy++; p->cx = -1; p->xa = 0; /*if (p->y < p->cy) p->y = p->cy;*/
+	if (p->spl > d_opt->xw) p->spl = d_opt->xw;
 	memset(p->spaces, 0, p->spl);
 }
 
@@ -737,7 +723,7 @@ void *html_special(struct part *p, int c, ...)
 			return convert_table;
 		case SP_USED:
 			va_end(l);
-			return (void *)!!p->data;
+			return (void *)(my_uintptr_t)!!p->data;
 		case SP_FRAMESET:
 			fsp = va_arg(l, struct frameset_param *);
 			va_end(l);
@@ -924,16 +910,16 @@ void push_base_format(unsigned char *url, struct document_options *opt)
 	e = mem_alloc(sizeof(struct html_element));
 	memset(e, 0, sizeof(struct html_element));
 	add_to_list(html_stack, e);
-	format.attr = 0;
-	format.fontsize = 3;
-	format.link = format.target = format.image = format.select = NULL;
-	format.form = NULL;
-	memcpy(&format.fg, &opt->default_fg, sizeof(struct rgb));
-	memcpy(&format.bg, &opt->default_bg, sizeof(struct rgb));
-	memcpy(&format.clink, &opt->default_link, sizeof(struct rgb));
-	memcpy(&format.vlink, &opt->default_vlink, sizeof(struct rgb));
-	format.href_base = stracpy(url);
-	format.target_base = stracpy(opt->framename);
+	format_.attr = 0;
+	format_.fontsize = 3;
+	format_.link = format_.target = format_.image = format_.select = NULL;
+	format_.form = NULL;
+	memcpy(&format_.fg, &opt->default_fg, sizeof(struct rgb));
+	memcpy(&format_.bg, &opt->default_bg, sizeof(struct rgb));
+	memcpy(&format_.clink, &opt->default_link, sizeof(struct rgb));
+	memcpy(&format_.vlink, &opt->default_vlink, sizeof(struct rgb));
+	format_.href_base = stracpy(url);
+	format_.target_base = stracpy(opt->framename);
 	par_format.align = opt->plain ? AL_NO : AL_LEFT;
 	par_format.leftmargin = opt->plain ? 0 : opt->margin;
 	par_format.rightmargin = opt->plain ? 0 : opt->margin;
@@ -1171,9 +1157,9 @@ void cached_format_html(struct view_state *vs, struct f_data_c *screen, struct d
 	copy_opt(&ce->opt, opt);
 	add_to_list(format_cache, ce);
 	screen->f_data = ce;
-	ce->time_to_get = -get_time();
+	ce->time_to_get = -(uttime)get_time();
 	format_html(cee, ce);
-	ce->time_to_get += get_time();
+	ce->time_to_get += (uttime)get_time();
 	sx:
 	screen->xw = ce->opt.xw;
 	screen->yw = ce->opt.yw;
@@ -1181,7 +1167,7 @@ void cached_format_html(struct view_state *vs, struct f_data_c *screen, struct d
 	screen->yp = ce->opt.yp;
 }
 
-long formatted_info(int type)
+unsigned long formatted_info(int type)
 {
 	int i = 0;
 	struct f_data *ce;
@@ -1348,10 +1334,29 @@ void html_interpret(struct session *ses)
 	}
 }
 
+struct link *get_link_at_location(struct f_data *f, int x, int y)
+{
+	struct link *l1, *l2, *l;
+	if (y < 0 || y >= f->y) return NULL;
+	l1 = f->lines1[y];
+	l2 = f->lines2[y];
+	if (!l1 || !l2) return NULL;
+	for (l = l1; l <= l2; l++) {
+		int i;
+		for (i = 0; i < l->n; i++) if (l->pos[i].x == x && l->pos[i].y == y) return l;
+	}
+	return NULL;
+}
+
 void add_srch_chr(struct f_data *f, unsigned char c, int x, int y, int nn)
 {
 	int n = f->nsearch;
 	if (c == ' ' && (!n || f->search[n - 1].c == ' ')) return;
+	if (c == '_') {
+		struct link *l = get_link_at_location(f, x, y);
+		if (l && (l->type == L_SELECT || l->type == L_FIELD || l->type == L_AREA))
+			return;
+	}
 	f->search[n].c = c;
 	f->search[n].x = x;
 	f->search[n].y = y;
@@ -1402,6 +1407,15 @@ int get_srch(struct f_data *f)
 	struct node *n;
 	int cnt = 0;
 	int cc = !f->search;
+#define add_srch(c_, x_, y_, n_)		\
+do {						\
+	if (!cc) {				\
+		add_srch_chr(f, c_, x_, y_, n_);\
+	} else {				\
+		if (cnt == MAXINT) overalloc();	\
+		cnt++;				\
+	}					\
+} while (0)
 	foreachback(n, f->nodes) {
 		int x, y;
 		int xm = n->x + n->xw, ym = n->y + n->yw;
@@ -1415,29 +1429,28 @@ int get_srch(struct f_data *f)
 				if (c == ' ' && ns) continue;
 				c = charset_upcase(c, f->opt.cp);
 				if (ns) {
-					if (!cc) add_srch_chr(f, c, x, y, 1);
-					else cnt++;
+					add_srch(c, x, y, 1);
 					ns = 0;
 					continue;
 				}
-				if (c != ' ')  if (!cc) add_srch_chr(f, c, x, y, 1);
-					       else cnt++;
-				else {
+				if (c != ' ') {
+					add_srch(c, x, y, 1);
+				} else {
 					int xx;
 					for (xx = x + 1; xx < xm && xx < f->data[y].l; xx++) if (!is_spc(f->data[y].d[xx])) goto ja_uz_z_toho_programovani_asi_zcvoknu;
 					xx = x;
 					ja_uz_z_toho_programovani_asi_zcvoknu:
 				/* uz jsem zcvoknul, trpim poruchou osobnosti */
-					if (!cc) add_srch_chr(f, ' ', x, y, xx - x);
-					else cnt++;
-					if (xx == x) break;
+					add_srch(' ', x, y, xx - x);
+					if (xx == x) goto uz_jsem_zcvoknul__jsem_psychopat__trpim_poruchou_osobnosti;
 					x = xx - 1;
 				}
 			}
-			if (!cc) add_srch_chr(f, ' ', x, y, 0);
-			else cnt++;
+			uz_jsem_zcvoknul__jsem_psychopat__trpim_poruchou_osobnosti:
+			add_srch(' ', x, y, 0);
 		}
 	}
+#undef add_srch
 	return cnt;
 	
 }
@@ -1452,6 +1465,7 @@ void get_search_data(struct f_data *f)
 	f->search = mem_alloc(n * sizeof(struct search));
 	get_srch(f);
 	while (f->nsearch && f->search[f->nsearch - 1].c == ' ') f->nsearch--;
+	f->search = mem_realloc(f->search, f->nsearch * sizeof(struct search));
 	/*debug("!");*/
 	sort_srch(f);
 }
