@@ -1,7 +1,5 @@
 #include "links.h"
 
-#define format format_
-
 #define AL_TR		-1
 
 #define VAL_TR		-1
@@ -354,7 +352,7 @@ struct table *parse_table(unsigned char *html, unsigned char *eof, unsigned char
 		if (lbhp) (*bad_html)[*bhp-1].e = html;
 		goto scan_done;
 	}
-	if (html + 2 <= eof && (html[1] == '!' || html[1] == '?')) {
+	if (eof - html >= 2 && (html[1] == '!' || html[1] == '?')) {
 		html = skip_comment(html, eof);
 		goto se;
 	}
@@ -950,8 +948,8 @@ void display_complicated_table(struct table *t, int x, int y, int *yy)
 				}
 				html_stack_dup();
 				html_top.dontkill = 1;
-				if (cell->b) format.attr |= AT_BOLD;
-				memcpy(&format.bg, &cell->bgcolor, sizeof(struct rgb));
+				if (cell->b) format_.attr |= AT_BOLD;
+				memcpy(&format_.bg, &cell->bgcolor, sizeof(struct rgb));
 				memcpy(&par_format.bgcolor, &cell->bgcolor, sizeof(struct rgb));
 				p = format_html_part(cell->start, cell->end, cell->align, t->cellpd, xw, f, t->p->xp + xp, t->p->yp + yp + (cell->valign != VAL_MIDDLE && cell->valign != VAL_BOTTOM ? 0 : (yw - cell->height) / (cell->valign == VAL_MIDDLE ? 2 : 1)), NULL, cell->link_num);
 				cell->xpos = xp;
