@@ -106,8 +106,9 @@ again:
 	return n;
 }
 
-static int get_http_code(unsigned char *head, int *code, int *version)
+int get_http_code(unsigned char *head, int *code, int *version)
 {
+	if (!head) return -1;
 	while (head[0] == ' ') head++;
 	if (upcase(head[0]) != 'H' || upcase(head[1]) != 'T' || upcase(head[2]) != 'T' ||
 	    upcase(head[3]) != 'P') return -1;
@@ -154,21 +155,25 @@ int check_http_server_bugs(unsigned char *url, struct http_connection_info *info
 	return 0;	
 }
 
-void http_end_request(struct connection *c, int notrunc)
+void http_end_request(struct connection *c, int notrunc, int nokeepalive, int state)
 {
-	if (c->state == S_OK) {
+	if (state == S__OK) {
 		if (c->cache) {
 			if (!notrunc) truncate_entry(c->cache, c->from, 1);
 			c->cache->incomplete = 0;
 		}
 	}
+	setcstate(c, state);
 	if (c->info && !((struct http_connection_info *)c->info)->close 
 #ifdef HAVE_SSL
-	&& (!c->ssl) /* We won't keep alive ssl connections */
+	 && !c->ssl /* We won't keep alive ssl connections */
 #endif
-	&& (!http_bugs.bug_post_no_keepalive || !strchr(c->url, POST_CHAR))) {
-		add_keepalive_socket(c, HTTP_KEEPALIVE_TIMEOUT);
-	} else abort_connection(c);
+	 && !nokeepalive
+	 && (!http_bugs.bug_post_no_keepalive || !strchr(c->url, POST_CHAR))) {
+		add_keepalive_socket(c, HTTP_KEEPALIVE_TIMEOUT, 0);
+	} else {
+		abort_connection(c);
+	}
 }
 
 void http_send_header(struct connection *);
@@ -177,10 +182,10 @@ void http_func(struct connection *c)
 {
 	/*setcstate(c, S_CONN);*/
 	/*set_timeout(c);*/
-	if (get_keepalive_socket(c)) {
+	if (get_keepalive_socket(c, NULL)) {
 		int p;
 		if ((p = get_port(c->url)) == -1) {
-			setcstate(c, S_INTERNAL);
+			setcstate(c, S_BAD_URL);
 			abort_connection(c);
 			return;
 		}
@@ -227,16 +232,13 @@ void http_send_header(struct connection *c)
 	}
 	if (upcase(c->url[0]) != 'P') add_to_str(&hdr, &l, "/");
 	if (!(u = get_url_data(c->url))) {
+		http_bad_url:
 		mem_free(hdr);
-		setcstate(c, S_BAD_URL);
-		http_end_request(c, 0);
+		http_end_request(c, 0, 1, S_BAD_URL);
 		return;
 	}
 	if (post && post < u) {
-		mem_free(hdr);
-		setcstate(c, S_BAD_URL);
-		http_end_request(c, 0);
-		return;
+		goto http_bad_url;
 	}
 	if (!post) uu = stracpy(u);
 	else uu = memacpy(u, post - u - 1);
@@ -268,6 +270,8 @@ void http_send_header(struct connection *c)
 	add_to_str(&hdr, &l, "User-Agent: Links (" VERSION_STRING "; ");
 	add_to_str(&hdr, &l, system_name);
 	add_to_str(&hdr, &l, "; ");
+	add_to_str(&hdr, &l, compiler_name);
+	add_to_str(&hdr, &l, "; ");
 	if (!list_empty(terminals)) {
 		add_to_str(&hdr, &l, "text");
 	} else {
@@ -281,7 +285,7 @@ void http_send_header(struct connection *c)
 		int aclen = 0;
 		ac = init_str();
 		for (i = 0; (cs = get_cp_mime_name(i)); i++) {
-			if (aclen) add_to_str(&ac, &aclen, ", ");
+			if (aclen) add_to_str(&ac, &aclen, ",");
 			else add_to_str(&ac, &aclen, "Accept-Charset: ");
 			add_to_str(&ac, &aclen, cs);
 		}
@@ -299,6 +303,7 @@ void http_send_header(struct connection *c)
 		if (!strstr(hdr + la, "en,") && !strstr(hdr + la, "en;")) add_to_str(&hdr, &l, "en;q=0.2, ");
 		add_to_str(&hdr, &l, "*;q=0.1\r\n");
 	}
+	/*add_to_str(&hdr, &l, "DNT: 1\r\n");*/
 	if (!http10) {
 		if (upcase(c->url[0]) != 'P') add_to_str(&hdr, &l, "Connection: ");
 		else add_to_str(&hdr, &l, "Proxy-Connection: ");
@@ -369,8 +374,7 @@ void read_http_data(struct connection *c, struct read_buffer *rb)
 	struct http_connection_info *info = c->info;
 	set_timeout(c);
 	if (rb->close == 2) {
-		setcstate(c, S_OK);
-		http_end_request(c, 0);
+		http_end_request(c, 0, 0, S__OK);
 		return;
 	}
 	if (info->length != -2) {
@@ -387,8 +391,7 @@ void read_http_data(struct connection *c, struct read_buffer *rb)
 		c->from += l;
 		kill_buffer_data(rb, l);
 		if (!info->length && !rb->close) {
-			setcstate(c, S_OK);
-			http_end_request(c, 0);
+			http_end_request(c, 0, 0, S__OK);
 			return;
 		}
 	} else {
@@ -403,8 +406,7 @@ void read_http_data(struct connection *c, struct read_buffer *rb)
 				}
 				kill_buffer_data(rb, l);
 				if (l <= 2) {
-					setcstate(c, S_OK);
-					http_end_request(c, 0);
+					http_end_request(c, 0, 0, S__OK);
 					return;
 				}
 				goto next_chunk;
@@ -548,16 +550,19 @@ void http_got_header(struct connection *c, struct read_buffer *rb)
 		abort_connection(c);
 		return;
 	}
-	if (h == 304) {
-		mem_free(head);
-		setcstate(c, S_OK);
-		http_end_request(c, 1);
-		return;
-	}
 	if (h == 204) {
 		mem_free(head);
-		setcstate(c, S_HTTP_204);
-		http_end_request(c, 0);
+		http_end_request(c, 0, 0, S_HTTP_204);
+		return;
+	}
+	if (h == 304) {
+		mem_free(head);
+		http_end_request(c, 1, 0, S__OK);
+		return;
+	}
+	if (h == 416 && c->from) {
+		mem_free(head);
+		http_end_request(c, 0, 1, S__OK);
 		return;
 	}
 	if (!c->cache) {
@@ -579,7 +584,7 @@ void http_got_header(struct connection *c, struct read_buffer *rb)
 		e->ssl_info = init_str();
 		add_num_to_str(&e->ssl_info, &l, SSL_get_cipher_bits(c->ssl, NULL));
 		add_to_str(&e->ssl_info, &l, "-bit ");
-		add_to_str(&e->ssl_info, &l, SSL_get_cipher_version(c->ssl));
+		add_to_str(&e->ssl_info, &l, (unsigned char *)SSL_get_version(c->ssl));
 		add_to_str(&e->ssl_info, &l, " ");
 		add_to_str(&e->ssl_info, &l, (unsigned  char *)SSL_get_cipher_name(c->ssl));
 	}
@@ -606,14 +611,7 @@ void http_got_header(struct connection *c, struct read_buffer *rb)
 		if (strlen(d) > 6) {
 			d[5] = 0;
 			if (!(strcasecmp(d, "bytes")) && d[6] >= '0' && d[6] <= '9') {
-#if defined(HAVE_STRTOLL)
-				long long f = strtoll(d + 6, NULL, 10);
-#elif defined(HAVE_STRTOQ)
-				longlong f = strtoq(d + 6, NULL, 10);
-#else
-				long f = strtol(d + 6, NULL, 10);
-				if (f == MAXLONG) f = -1;
-#endif
+				my_strtoll_t f = my_strtoll(d + 6, NULL);
 				if (f >= 0 && (off_t)f >= 0 && (off_t)f == f) c->from = f;
 			}
 		}
@@ -630,14 +628,7 @@ void http_got_header(struct connection *c, struct read_buffer *rb)
 	}
 	if ((d = parse_http_header(e->head, "Content-Length", NULL))) {
 		unsigned char *ep;
-#if defined(HAVE_STRTOLL)
-		long long l = strtoll(d, (char **)(void *)&ep, 10);
-#elif defined(HAVE_STRTOQ)
-		longlong l = strtoq(d, (char **)(void *)&ep, 10);
-#else
-		long l = strtol(d, (char **)(void *)&ep, 10);
-		if (l == MAXLONG) l = -1;
-#endif
+		my_strtoll_t l = my_strtoll(d, &ep);
 		if (!*ep && l >= 0 && (off_t)l >= 0 && (off_t)l == l) {
 			if (!info->close || version >= 11) info->length = l;
 			if (c->from + l >= 0) c->est_length = c->from + l;
