@@ -15,7 +15,6 @@ struct itrm {
 	int sock_out;
 	int ctl_in;
 	int blocked;
-	struct termios t;
 	int flags;
 	unsigned char kqueue[IN_BUF_SIZE];
 	int qlen;
@@ -42,12 +41,19 @@ void free_all_itrms()
 	if (ditrm) free_trm(ditrm);
 }
 
+static void itrm_error(struct itrm *itrm)
+{
+	free_trm(itrm);
+	terminate_loop = 1;
+}
+
 void write_ev_queue(struct itrm *itrm)
 {
 	int l;
 	if (!itrm->eqlen) internal("event queue empty");
-	if ((l = write(itrm->sock_out, itrm->ev_queue, itrm->eqlen > 128 ? 128 : itrm->eqlen)) == -1) {
-		free_trm(itrm);
+	EINTRLOOP(l, write(itrm->sock_out, itrm->ev_queue, itrm->eqlen > 128 ? 128 : itrm->eqlen));
+	if (l == -1) {
+		itrm_error(itrm);
 		return;
 	}
 	memmove(itrm->ev_queue, itrm->ev_queue + l, itrm->eqlen -= l);
@@ -66,17 +72,19 @@ void queue_event(struct itrm *itrm, unsigned char *data, int len)
 {
 	int w = 0;
 	if (!len) return;
-	if (!itrm->eqlen && can_write(itrm->sock_out) && (w = write(itrm->sock_out, data, len)) <= 0) {
-		/*free_trm(itrm);*/
-		register_bottom_half((void (*)(void *))free_trm, itrm);
-		return;
+	if (!itrm->eqlen && can_write(itrm->sock_out)) {
+		EINTRLOOP(w, write(itrm->sock_out, data, len));
+		if (w <= 0) {
+			register_bottom_half((void (*)(void *))itrm_error, itrm);
+			return;
+		}
 	}
 	if (w < len) {
 		if ((unsigned)itrm->eqlen + (unsigned)(len - w) > MAXINT) overalloc();
 		itrm->ev_queue = mem_realloc(itrm->ev_queue, itrm->eqlen + len - w);
 		memcpy(itrm->ev_queue + itrm->eqlen, data + w, len - w);
 		itrm->eqlen += len - w;
-		set_handlers(itrm->sock_out, get_handler(itrm->sock_out, H_READ), (void (*)(void *))write_ev_queue, (void (*)(void *))free_trm, itrm);
+		set_handlers(itrm->sock_out, get_handler(itrm->sock_out, H_READ), (void (*)(void *))write_ev_queue, (void (*)(void *))itrm_error, itrm);
 	}
 }
 
@@ -86,20 +94,13 @@ void kbd_ctrl_c()
 	if (ditrm) queue_event(ditrm, (unsigned char *)&ev, sizeof(struct event));
 }
 
-/*
-unsigned char *init_seq = "\033[?1000h\033[?47h\0337";
-unsigned char *term_seq = "\033[2J\033[?1000l\033[?47l\0338\b \b";
-*/
-
 unsigned char init_seq[] = "\033)0\0337";
-unsigned char init_seq_x_mouse[] = "\033[?1000h";
+unsigned char init_seq_x_mouse[] = "\033[?1000h\033[?1002h\033[?1005h\033[?1006l\033[?1015l";
 unsigned char init_seq_tw_mouse[] = "\033[?9h";
 unsigned char term_seq[] = "\033[2J\0338\r \b";
-unsigned char term_seq_x_mouse[] = "\033[?1000l\015       \015";
+unsigned char term_seq_x_mouse[] = "\033[?1000l\r       \r\033[?1002l\r       \r\033[?1005l\r       \r";
 unsigned char term_seq_tw_mouse[] = "\033[?9l";
 
-/*unsigned char *term_seq = "\033[2J\033[?1000l\0338\b \b";*/
-
 void send_init_sequence(int h, int flags)
 {
 	want_draw();
@@ -134,9 +135,48 @@ void resize_terminal()
 	queue_event(ditrm, (unsigned char *)&ev, sizeof(struct event));
 }
 
+#if defined(OS_SETRAW)
+
+#elif defined(HAVE_TERMIOS_H)
+
+void os_cfmakeraw(struct termios *t)
+{
+#ifdef HAVE_CFMAKERAW
+	cfmakeraw(t);
+#ifdef VMIN
+	t->c_cc[VMIN] = 1;	/* cfmakeraw is broken on AIX */
+#endif
+#else
+	t->c_iflag &= ~(IGNBRK|BRKINT|PARMRK|ISTRIP|INLCR|IGNCR|ICRNL|IXON);
+	t->c_oflag &= ~OPOST;
+	t->c_lflag &= ~(ECHO|ECHONL|ICANON|ISIG|IEXTEN);
+	t->c_cflag &= ~(CSIZE|PARENB);
+	t->c_cflag |= CS8;
+	t->c_cc[VMIN] = 1;
+	t->c_cc[VTIME] = 0;
+#endif
+#if defined(NO_CTRL_Z) && defined(VSUSP)
+	t->c_cc[VSUSP] = 0;
+#endif
+}
+
+
 int ttcgetattr(int fd, struct termios *t)
 {
 	int r;
+	block_signals(
+#ifdef SIGTTOU
+		SIGTTOU
+#else
+		0
+#endif
+		,
+#ifdef SIGTTIN
+		SIGTTIN
+#else
+		0
+#endif
+		);
 #ifdef SIGTTOU
 	interruptible_signal(SIGTTOU, 1);
 #endif
@@ -150,12 +190,26 @@ int ttcgetattr(int fd, struct termios *t)
 #ifdef SIGTTIN
 	interruptible_signal(SIGTTIN, 0);
 #endif
+	unblock_signals();
 	return r;
 }
 
 int ttcsetattr(int fd, int a, struct termios *t)
 {
 	int r;
+	block_signals(
+#ifdef SIGTTOU
+		SIGTTOU
+#else
+		0
+#endif
+		,
+#ifdef SIGTTIN
+		SIGTTIN
+#else
+		0
+#endif
+		);
 #ifdef SIGTTOU
 	interruptible_signal(SIGTTOU, 1);
 #endif
@@ -169,14 +223,18 @@ int ttcsetattr(int fd, int a, struct termios *t)
 #ifdef SIGTTIN
 	interruptible_signal(SIGTTIN, 0);
 #endif
+	unblock_signals();
 	return r;
 }
 
-int setraw(int fd, struct termios *p)
+static struct termios saved_termios;
+
+int setraw(int ctl, int save)
 {
 	struct termios t;
+	if (ctl < 0) return 0;
 	memset(&t, 0, sizeof(struct termios));
-	if (ttcgetattr(fd, &t)) {
+	if (ttcgetattr(ctl, &t)) {
 		/*fprintf(stderr, "getattr result %s\n", strerror(errno));*/
 		/* If the terminal was destroyed (the user logged off),
 		 * we fake success here so that we can destroy the terminal
@@ -188,20 +246,57 @@ int setraw(int fd, struct termios *p)
 		if (errno == EIO || errno == ENXIO) return 0;
 		return -1;
 	}
-	if (p) memcpy(p, &t, sizeof(struct termios));
+	if (save) memcpy(&saved_termios, &t, sizeof(struct termios));
 	os_cfmakeraw(&t);
 	t.c_lflag |= ISIG;
 #ifdef TOSTOP
 	t.c_lflag |= TOSTOP;
 #endif
 	t.c_oflag |= OPOST;
-	if (ttcsetattr(fd, TCSANOW, &t)) {
+#ifdef DOS
+	t.c_lflag &= ~(ISIG | ECHO);
+#endif
+	if (ttcsetattr(ctl, TCSANOW, &t)) {
 		/*fprintf(stderr, "setattr result %s\n", strerror(errno));*/
 		return -1;
 	}
 	return 0;
 }
 
+static void setcooked(int ctl)
+{
+	if (ctl < 0) return;
+	ttcsetattr(ctl, TCSANOW, &saved_termios);
+}
+
+#elif defined(HAVE_SGTTY_H)
+
+static struct sgttyb saved_sgtty;
+
+static int setraw(int ctl, int save)
+{
+	struct sgttyb sgtty;
+	if (ctl < 0) return 0;
+	if (gtty(ctl, &sgtty)) {
+		return -1;
+	}
+	if (save) memcpy(&saved_sgtty, &sgtty, sizeof(struct sgttyb));
+	sgtty.sg_flags |= CBREAK;
+	sgtty.sg_flags &= ~ECHO;
+	if (stty(ctl, &sgtty)) {
+		return -1;
+	}
+	return 0;
+}
+
+static void setcooked(int ctl)
+{
+	if (ctl < 0) return;
+	stty(ctl, &saved_sgtty);
+}
+
+#endif
+
 void handle_trm(int std_in, int std_out, int sock_in, int sock_out, int ctl_in, void *init_string, int init_len)
 {
 	int x, y;
@@ -225,14 +320,14 @@ void handle_trm(int std_in, int std_out, int sock_in, int sock_out, int ctl_in,
 	itrm->tm = -1;
 	itrm->ev_queue = DUMMY;
 	itrm->eqlen = 0;
-	if (ctl_in >= 0) setraw(ctl_in, &itrm->t);
-	set_handlers(std_in, (void (*)(void *))in_kbd, NULL, (void (*)(void *))free_trm, itrm);
-	if (sock_in != std_out) set_handlers(sock_in, (void (*)(void *))in_sock, NULL, (void (*)(void *))free_trm, itrm);
+	setraw(itrm->ctl_in, 1);
+	set_handlers(std_in, (void (*)(void *))in_kbd, NULL, (void (*)(void *))itrm_error, itrm);
+	if (sock_in != std_out) set_handlers(sock_in, (void (*)(void *))in_sock, NULL, (void (*)(void *))itrm_error, itrm);
 	ev.x = x;
 	ev.y = y;
 	handle_terminal_resize(ctl_in, resize_terminal);
 	queue_event(itrm, (unsigned char *)&ev, sizeof(struct event));
-	xwin = is_xterm() * ENV_XWIN + can_twterm() * ENV_TWIN + (!!getenv("STY")) * ENV_SCREEN + get_system_env();
+	xwin = is_xterm() * ENV_XWIN + is_twterm() * ENV_TWIN + is_screen() * ENV_SCREEN + get_system_env();
 	itrm->flags = 0;
 	if (!(ts = getenv("TERM"))) ts = "";
 	if ((xwin & ENV_TWIN) && !strcmp(ts,"linux")) itrm->flags |= USE_TWIN_MOUSE;
@@ -278,11 +373,11 @@ int unblock_itrm(int fd)
 {
 	struct itrm *itrm = ditrm;
 	if (!itrm) return -1;
-	if (itrm->ctl_in >= 0 && setraw(itrm->ctl_in, NULL)) return -1;
+	if (setraw(itrm->ctl_in, 0)) return -1;
 	if (itrm->blocked != fd + 1) return -2;
 	itrm->blocked = 0;
 	send_init_sequence(itrm->std_out, itrm->flags);
-	set_handlers(itrm->std_in, (void (*)(void *))in_kbd, NULL, (void (*)(void *))free_trm, itrm);
+	set_handlers(itrm->std_in, (void (*)(void *))in_kbd, NULL, (void (*)(void *))itrm_error, itrm);
 	handle_terminal_resize(itrm->ctl_in, resize_terminal);
 	unblock_stdin();
 	resize_terminal();
@@ -298,8 +393,8 @@ void block_itrm(int fd)
 	block_stdin();
 	unhandle_terminal_resize(itrm->ctl_in);
 	send_term_sequence(itrm->std_out,itrm->flags);
-	ttcsetattr(itrm->ctl_in, TCSANOW, &itrm->t);
-	set_handlers(itrm->std_in, NULL, NULL, (void (*)(void *))free_trm, itrm);
+	setcooked(itrm->ctl_in);
+	set_handlers(itrm->std_in, NULL, NULL, (void (*)(void *))itrm_error, itrm);
 }
 
 void free_trm(struct itrm *itrm)
@@ -309,12 +404,13 @@ void free_trm(struct itrm *itrm)
 	if (itrm->orig_title) mem_free(itrm->orig_title), itrm->orig_title = NULL;
 	unhandle_terminal_resize(itrm->ctl_in);
 	send_term_sequence(itrm->std_out,itrm->flags);
-	ttcsetattr(itrm->ctl_in, TCSANOW, &itrm->t);
+	setcooked(itrm->ctl_in);
 	if (itrm->mouse_h) unhandle_mouse(itrm->mouse_h);
 	set_handlers(itrm->std_in, NULL, NULL, NULL, NULL);
 	set_handlers(itrm->sock_in, NULL, NULL, NULL, NULL);
 	set_handlers(itrm->std_out, NULL, NULL, NULL, NULL);
 	set_handlers(itrm->sock_out, NULL, NULL, NULL, NULL);
+	unregister_bottom_half((void (*)(void *))itrm_error, itrm);
 	if (itrm->tm != -1) kill_timer(itrm->tm);
 	mem_free(itrm->ev_queue);
 	mem_free(itrm);
@@ -323,7 +419,7 @@ void free_trm(struct itrm *itrm)
 
 void fatal_tty_exit(void)
 {
-	if (ditrm) ttcsetattr(ditrm->ctl_in, TCSANOW, &ditrm->t);
+	if (ditrm) setcooked(ditrm->ctl_in);
 }
 
 void resize_terminal_x(unsigned char *text)
@@ -352,7 +448,18 @@ void dispatch_special(unsigned char *text)
 
 unsigned char buf[OUT_BUF_SIZE];
 
-#define RD(xx) { unsigned char cc; if (p < c) cc = buf[p++]; else if ((hard_read(itrm->sock_in, &cc, 1)) <= 0) goto fr; xx = cc; }
+#define RD(xx)							\
+do {								\
+	unsigned char cc;					\
+	if (p < c) cc = buf[p++];				\
+	else if ((hard_read(itrm->sock_in, &cc, 1)) <= 0) {	\
+		mem_free(path);					\
+		mem_free(delete);				\
+		goto fr;					\
+	}							\
+	xx = cc;						\
+} while (0)
+
 
 void in_sock(struct itrm *itrm)
 {
@@ -361,10 +468,11 @@ void in_sock(struct itrm *itrm)
 	unsigned char ch;
 	int fg;
 	int c, i, p;
-	if ((c = read(itrm->sock_in, buf, OUT_BUF_SIZE)) <= 0) {
+	int rs;
+	EINTRLOOP(c, read(itrm->sock_in, buf, OUT_BUF_SIZE));
+	if (c <= 0) {
 		fr:
-		free_trm(itrm);
-		terminate = 1;
+		itrm_error(itrm);
 		return;
 	}
 	qwerty:
@@ -385,11 +493,10 @@ void in_sock(struct itrm *itrm)
 	memmove(buf, buf + i, OUT_BUF_SIZE - i);
 	c -= i;
 	p = 0;
-	/*fg = RD;*/
-	RD(fg);
 	path = init_str();
 	delete = init_str();
 	pl = dl = 0;
+	RD(fg);
 	while (1) {
 		RD(ch);
 		if (!ch) break;
@@ -406,7 +513,8 @@ void in_sock(struct itrm *itrm)
 		int blockh;
 		unsigned char *param;
 		if (is_blocked() && fg) {
-			if (*delete) unlink(delete);
+			if (*delete)
+				EINTRLOOP(rs, unlink(delete));
 			goto to_je_ale_hnus;
 		}
 		param = mem_alloc(strlen(path) + strlen(delete) + 3);
@@ -421,9 +529,9 @@ void in_sock(struct itrm *itrm)
 		}
 		mem_free(param);
 		if (fg == 1) {
-			set_handlers(blockh, (void (*)(void *))unblock_itrm_x, NULL, (void (*)(void *))unblock_itrm_x, (void *)blockh);
+			set_handlers(blockh, (void (*)(void *))unblock_itrm_x, NULL, (void (*)(void *))unblock_itrm_x, (void *)(my_uintptr_t)blockh);
 		} else {
-			set_handlers(blockh, close_handle, NULL, close_handle, (void *)blockh);
+			set_handlers(blockh, close_handle, NULL, close_handle, (void *)(my_uintptr_t)blockh);
 		}
 	}
 	to_je_ale_hnus:
@@ -545,6 +653,15 @@ int is_interix(void)
 #endif
 }
 
+int is_uwin(void)
+{
+#ifdef _UWIN
+	return 1;
+#else
+	return 0;
+#endif
+}
+
 int is_ibm(void)
 {
 	unsigned char *term = getenv("TERM");
@@ -594,6 +711,7 @@ int process_queue(struct itrm *itrm)
 						ev.x = KBD_END;
 						break;
 					}
+					/*-fallthrough*/
 				case 'G': ev.x = KBD_PAGE_DOWN; break;
 				case 'P':
 					if (is_ibm()) {
@@ -616,7 +734,12 @@ int process_queue(struct itrm *itrm)
 					ev.x = KBD_F5; break;
 				case 'W': ev.x = KBD_F8; break;
 				case 'X': ev.x = KBD_F9; break;
-				case 'Y': ev.x = KBD_F11; break;
+				case 'Y':
+					if (is_uwin())
+						ev.x = itrm->kqueue[1] == '[' ? KBD_END : KBD_F10;
+					else
+						ev.x = KBD_F11;
+					break;
 
 				case 'q': switch (v) {
 					case 139: ev.x = KBD_INS; break;
@@ -663,7 +786,7 @@ int process_queue(struct itrm *itrm)
 				case 'R':
 					resize_terminal();
 					break ;
-				case 'M': if (itrm->qlen - el < 3) goto ret;
+				case 'M':
 					if (v == 5) {
 						if (xterm_button == -1) xterm_button = 0; /* */
 						if (itrm->qlen - el < 5) goto ret;
@@ -681,19 +804,59 @@ int process_queue(struct itrm *itrm)
 						xterm_button = itrm->kqueue[el] - ' ';
 						el += 5;
 					} else {
-						ev.x = itrm->kqueue[el+1] - ' ' - 1;
-						ev.y = itrm->kqueue[el+2] - ' ' - 1;
-						ev.b = B_DOWN;
-						if (itrm->kqueue[el] & 4) ev.b = B_DRAG;
-						if ((ev.b |= (itrm->kqueue[el] & BM_BUTT) | B_DOWN) == 3) {
-							ev.b = B_UP;
-							if (xterm_button != -1) ev.b |= xterm_button;
+						int x, y, button;
+						if (itrm->qlen - el < 3) goto ret;
+						if ((itrm->kqueue[el] & ~0x1f) == 0x20) button = B_DOWN;
+						else if ((itrm->kqueue[el] & ~0x1f) == 0x40) button = B_DRAG;
+						else {
+							button = -1;
+							goto skip_button;
+						}
+						button |= itrm->kqueue[el] & 3;
+						if ((itrm->kqueue[el] & 3) == 3) {
+							if (xterm_button == -1)
+								xterm_button = B_LEFT;
+							button = B_UP;
+							button |= xterm_button;
 						}  
-						/*if ((itrm->kqueue[el] & 4) && ev.b != B_UP) ev.b |= B_DRAG;*/
-						xterm_button = -1;
-						if ((ev.b & BM_ACT) == B_DOWN) xterm_button = ev.b & BM_BUTT;
-						el += 3;
+						xterm_button = button & BM_BUTT;
+						skip_button:
+						el++;
+						if (itrm->kqueue[el] < ' ') {
+							x = -1;
+							el++;
+						} else if (itrm->kqueue[el] >= 0xc2 && itrm->kqueue[el] < 0xe0 && itrm->qlen - el >= 3 && itrm->kqueue[el + 1] >= 0x80 && itrm->kqueue[el + 1] < 0xc0) {
+							x = (((itrm->kqueue[el] & 0x1f) << 6) | (itrm->kqueue[el + 1] & 0x3f)) - ' ' - 1;
+							el += 2;
+						} else {
+							x = itrm->kqueue[el] - ' ' - 1;
+							el++;
+						}
+						if (itrm->kqueue[el] < ' ') {
+							y = -1;
+							el++;
+						} else if (itrm->kqueue[el] >= 0xc2 && itrm->kqueue[el] < 0xe0 && itrm->qlen - el >= 2 && itrm->kqueue[el + 1] >= 0x80 && itrm->kqueue[el + 1] < 0xc0) {
+							y = (((itrm->kqueue[el] & 0x1f) << 6) | (itrm->kqueue[el + 1] & 0x3f)) - ' ' - 1;
+							el += 2;
+						} else {
+							y = itrm->kqueue[el] - ' ' - 1;
+							el++;
+						}
+						if (x < 0 || y < 0 || button < 0)
+							break;
+						ev.b = button;
+						ev.x = x;
+						ev.y = y;
 					}
+					/*{
+						int a;
+						for (a = 0; a < el; a++) {
+							fprintf(stderr, " %02x", itrm->kqueue[a]);
+							if (itrm->kqueue[a] >= ' ') fprintf(stderr, "(%c)", itrm->kqueue[a]);
+						}
+						fprintf(stderr, "\n");
+						fprintf(stderr, "%ld %ld %ld\n", ev.x, ev.y, ev.b);
+					}*/
 					ev.ev = EV_MOUSE;
 					break;
 			}
@@ -760,6 +923,7 @@ do_alt:
 	if (ev.x == 4) ev.x = KBD_DEL;
 	if (ev.x == 5) ev.x = KBD_END;
 	if (ev.x == 6) ev.x = KBD_PAGE_DOWN;*/
+	if (ev.x == 3) ev.x = KBD_CTRL_C;
 	if (ev.x == 8) ev.x = KBD_BS;
 	if (ev.x == 9) ev.x = KBD_TAB;
 	if (ev.x == 10) ev.x = KBD_ENTER/*, ev.y = KBD_CTRL*/;
@@ -785,7 +949,7 @@ do_alt:
 		memmove(itrm->kqueue, itrm->kqueue + el, itrm->qlen -= el);
 	}
 	end:
-	if (itrm->qlen < IN_BUF_SIZE && !itrm->blocked) set_handlers(itrm->std_in, (void (*)(void *))in_kbd, NULL, (void (*)(void *))free_trm, itrm);
+	if (itrm->qlen < IN_BUF_SIZE && !itrm->blocked) set_handlers(itrm->std_in, (void (*)(void *))in_kbd, NULL, (void (*)(void *))itrm_error, itrm);
 	return el;
 	ret:
 	itrm->tm = install_timer(ESC_TIMEOUT, (void (*)(void *))kbd_timeout, itrm);
@@ -798,19 +962,25 @@ void in_kbd(struct itrm *itrm)
 	if (!can_read(itrm->std_in)) return;
 	if (itrm->tm != -1) kill_timer(itrm->tm), itrm->tm = -1;
 	if (itrm->qlen >= IN_BUF_SIZE) {
-		set_handlers(itrm->std_in, NULL, NULL, (void (*)(void *))free_trm, itrm);
+		set_handlers(itrm->std_in, NULL, NULL, (void (*)(void *))itrm_error, itrm);
 		while (process_queue(itrm));
 		return;
 	}
-	if ((r = read(itrm->std_in, itrm->kqueue + itrm->qlen, IN_BUF_SIZE - itrm->qlen)) <= 0) {
+	EINTRLOOP(r, read(itrm->std_in, itrm->kqueue + itrm->qlen, IN_BUF_SIZE - itrm->qlen));
+	if (r <= 0) {
 		struct event ev = { EV_ABORT, 0, 0, 0 };
-		set_handlers(itrm->std_in, NULL, NULL, (void (*)(void *))free_trm, itrm);
+		set_handlers(itrm->std_in, NULL, NULL, (void (*)(void *))itrm_error, itrm);
 		queue_event(itrm, (unsigned char *)&ev, sizeof(struct event));
 		return;
 	}
+	more_data:
 	if ((itrm->qlen += r) > IN_BUF_SIZE) {
 		error("ERROR: too many bytes read");
 		itrm->qlen = IN_BUF_SIZE;
 	}
+	if (itrm->qlen < IN_BUF_SIZE && can_read(itrm->std_in)) {
+		EINTRLOOP(r, read(itrm->std_in, itrm->kqueue + itrm->qlen, IN_BUF_SIZE - itrm->qlen));
+		if (r > 0) goto more_data;
+	}
 	while (process_queue(itrm));
 }
