@@ -23,13 +23,12 @@ static void add_keybinding(int km, int act, long x, long y)
 			goto add;
 		}
 	
-	if ((kb = mem_alloc(sizeof(struct keybinding)))) {
-		add:
-		kb->act = act;
-		kb->x = x;
-		kb->y = y;
-		add_to_list(keymaps[km], kb);
-	}
+	kb = mem_alloc(sizeof(struct keybinding));
+	add:
+	kb->act = act;
+	kb->x = x;
+	kb->y = y;
+	add_to_list(keymaps[km], kb);
 }
 
 static void delete_keybinding(int km, long x, long y)
@@ -69,14 +68,14 @@ int kbd_action(int kmap, struct event *ev)
  * Config file helpers.
  */
 
-struct strtonum {
+struct str2num {
 	unsigned char *str;
 	long num;
 };
 
-static long strtonum(struct strtonum *table, char *s)
+static long str2num(struct str2num *table, char *s)
 {
-	struct strtonum *p;
+	struct str2num *p;
 	for (p = table; p->str; p++) 
 		if (!strcmp(p->str, s)) return p->num;
 	return -1;
@@ -84,18 +83,18 @@ static long strtonum(struct strtonum *table, char *s)
 
 static int parse_keymap(unsigned char *s)
 {
-	struct strtonum table[] = {
+	struct str2num table[] = {
 		{ "main", KM_MAIN },
 		{ "edit", KM_EDIT },
 		{ "menu", KM_MENU },
 		{ NULL, 0 }
 	};
-	return strtonum(table, s);
+	return str2num(table, s);
 }
 
 static long parse_key(unsigned char *s)
 {
-	struct strtonum table[] = {
+	struct str2num table[] = {
 		{ "Enter", KBD_ENTER },
 		{ "Backspace", KBD_BS },
 		{ "Tab", KBD_TAB },
@@ -124,7 +123,7 @@ static long parse_key(unsigned char *s)
 		{ "F12", KBD_F12 },
 		{ NULL, 0 }
 	};
-	return (strlen(s) == 1) ? *s : strtonum(table, s);
+	return (strlen(s) == 1) ? *s : str2num(table, s);
 }
 
 static int parse_keystroke(unsigned char *s, long *x, long *y)
@@ -250,7 +249,6 @@ unsigned char *unbind_rd(struct option *o, unsigned char *line)
 	else 
 		delete_keybinding(kmap, x, y);
 
-	x:
 	if (ckey) mem_free(ckey);
 	if (ckmap) mem_free(ckmap);
 	return err;
@@ -277,7 +275,7 @@ static struct default_kb default_main_keymap[] = {
 	{ ACT_DOWN, KBD_DOWN, 0 },
 	{ ACT_UP, KBD_UP, 0 },
 	{ ACT_COPY_CLIPBOARD, KBD_INS, KBD_CTRL },
-	{ ACT_COPY_CLIPBOARD, 'C', KBD_CTRL },
+	{ ACT_COPY_CLIPBOARD, 'B', KBD_CTRL },
 	{ ACT_SCROLL_UP, KBD_INS, 0 },
 	{ ACT_SCROLL_UP, 'P', KBD_CTRL },
 	{ ACT_SCROLL_DOWN, KBD_DEL, 0 },
@@ -302,20 +300,20 @@ static struct default_kb default_main_keymap[] = {
 	{ ACT_RELOAD, 'R', KBD_CTRL },
 	{ ACT_GOTO_URL, 'g', 0 },
 	{ ACT_GOTO_URL_CURRENT, 'G', 0 },
-	{ ACT_ADD_BOOKMARK, 'a' },
-	{ ACT_ADD_BOOKMARK, 'A' },
-	{ ACT_BOOKMARK_MANAGER, 's' },
-	{ ACT_BOOKMARK_MANAGER, 'S' },
-	{ ACT_QUIT, 'q' },
-	{ ACT_REALLYQUIT, 'Q' },
-	{ ACT_DOCUMENT_INFO, '=' },
-	{ ACT_HEADER_INFO, '|' },
-	{ ACT_TOGGLE_HTML_PLAIN, '\\' },
-	{ ACT_TOGGLE_DISPLAY_IMAGES, '*' },
-	{ ACT_NEXT_FRAME, KBD_TAB },
-	{ ACT_MENU, KBD_ESC },
-	{ ACT_MENU, KBD_F9 },
-	{ ACT_FILE_MENU, KBD_F10 },
+	{ ACT_ADD_BOOKMARK, 'a', 0 },
+	{ ACT_ADD_BOOKMARK, 'A', 0 },
+	{ ACT_BOOKMARK_MANAGER, 's', 0 },
+	{ ACT_BOOKMARK_MANAGER, 'S', 0 },
+	{ ACT_QUIT, 'q', 0 },
+	{ ACT_REALLYQUIT, 'Q', 0 },
+	{ ACT_DOCUMENT_INFO, '=', 0 },
+	{ ACT_HEADER_INFO, '|', 0 },
+	{ ACT_TOGGLE_HTML_PLAIN, '\\', 0 },
+	{ ACT_TOGGLE_DISPLAY_IMAGES, '*', 0 },
+	{ ACT_NEXT_FRAME, KBD_TAB, 0 },
+	{ ACT_MENU, KBD_ESC, 0 },
+	{ ACT_MENU, KBD_F9, 0 },
+	{ ACT_FILE_MENU, KBD_F10, 0 },
 	{ 0, 0, 0 }
 };
 
@@ -329,7 +327,7 @@ static struct default_kb default_edit_keymap[] = {
 	{ ACT_END, KBD_END, 0 },
 	{ ACT_END, 'E', KBD_CTRL },
 	{ ACT_COPY_CLIPBOARD, KBD_INS, KBD_CTRL },
-	{ ACT_COPY_CLIPBOARD, 'C', KBD_CTRL },
+	{ ACT_COPY_CLIPBOARD, 'B', KBD_CTRL },
 	{ ACT_CUT_CLIPBOARD, 'X', KBD_CTRL },
 	{ ACT_PASTE_CLIPBOARD, 'V', KBD_CTRL },
 	{ ACT_ENTER, KBD_ENTER, 0 },
