@@ -1,18 +1,4 @@
-#ifndef __EXTENSIONS__
-#define __EXTENSIONS__
-#endif
-
-#ifndef _LARGEFILE_SOURCE
-#define _LARGEFILE_SOURCE	1
-#endif
-
-#ifndef _ALL_SOURCE
-#define _ALL_SOURCE             1
-#endif
-
-#ifndef _FILE_OFFSET_BITS
-#define _FILE_OFFSET_BITS	64
-#endif
+#include "com-defs.h"
 
 #ifdef HAVE_CONFIG_H
 #include "config.h"
@@ -20,20 +6,43 @@
 #ifdef HAVE_CONFIG2_H
 #include "config2.h"
 #endif
+#ifdef HAVE_CONFIG_VMS_H
+#include "config-vms.h"
+#ifdef __VAX
+#undef HAVE_LONG_LONG
+#undef HAVE_STRTOLL
+#undef HAVE_STRTOQ
+#undef HAVE_SOCKETPAIR
+#endif
+#endif
+
+#ifndef HAVE_VOLATILE
+#define volatile
+#endif
 
 #include "os_dep.h"
 #include <stdio.h>
+#ifdef HAVE_STDLIB_H
 #include <stdlib.h>
+#endif
 #include <stdarg.h>
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
+#ifdef HAVE_STRING_H
 #include <string.h>
+#endif
+#ifdef HAVE_STRINGS_H
+#include <strings.h>
+#endif
 #include <errno.h>
 #ifdef HAVE_LIMITS_H
 #include <limits.h>
 #endif
 #include <sys/types.h>
+#ifdef HAVE_INTTYPES_H
+#include <inttypes.h>
+#endif
 
 #ifndef __USE_XOPEN
 #define U_X
@@ -69,13 +78,22 @@
 #ifdef HAVE_FCNTL_H
 #include <fcntl.h>
 #endif
+#ifdef HAVE_SYS_FILE_H
+#include <sys/file.h>
+#endif
 #ifdef HAVE_DIRENT_H
 #include <dirent.h>
+#elif defined(HAVE_SYS_NDIR_H)
+#include <sys/ndir.h>
+#elif defined(HAVE_SYS_DIR_H)
+#include <sys/dir.h>
+#ifndef dirent
+#define dirent direct
+#endif
+#elif defined(HAVE_NDIR_H)
+#include <ndir.h>
 #endif
 #include <signal.h>
-/*#ifdef HAVE_SIGACTION_H
-#include <sigaction.h>
-#endif*/
 #ifdef HAVE_SYS_WAIT_H
 #include <sys/wait.h>
 #endif
@@ -88,12 +106,24 @@
 #ifdef HAVE_SYS_CYGWIN_H
 #include <sys/cygwin.h>
 #endif
+#ifdef HAVE_UWIN_H
+#include <uwin.h>
+#endif
 #ifdef HAVE_INTERIX_INTERIX_H
 #include <interix/interix.h>
 #endif
 #ifdef HAVE_IO_H
 #include <io.h>
 #endif
+#ifdef HAVE_PROCESS_H
+#include <process.h>
+#endif
+#ifdef HAVE_CYGWIN_PROCESS_H
+#include <cygwin/process.h>
+#endif
+#ifdef HAVE_UNIXLIB_H
+#include <unixlib.h>
+#endif
 #ifdef HAVE_SYS_UTSNAME_H
 #include <sys/utsname.h>
 #endif
@@ -103,6 +133,12 @@
 #ifdef HAVE_GRP_H
 #include <grp.h>
 #endif
+#ifdef HAVE_MALLOC_H
+#include <malloc.h>
+#endif
+#ifdef HAVE_ALLOCA_H
+#include <alloca.h>
+#endif
 /*
 #ifdef HAVE_WAIT_H
 #include <wait.h>
@@ -125,14 +161,22 @@
 #ifdef HAVE_ARPA_INET_H
 #include <arpa/inet.h>
 #endif
+#ifdef HAVE_UTIME_H
 #include <utime.h>
+#endif
 
+#ifdef HAVE_TERMIOS_H
 #include <termios.h>
+#elif defined(HAVE_SGTTY_H)
+#include <sgtty.h>
+#endif
 
 #ifdef HAVE_LONG_LONG
 #define longlong long long
+#define ulonglong unsigned long long
 #else
 #define longlong double
+#define ulonglong double
 #endif
 
 #if defined(__INTERIX) && defined(HAVE_STRTOQ)
@@ -143,22 +187,10 @@ __cdecl
 strtoq(const char *, char **, int);
 #endif
 
-#if defined(__hpux) && defined(__LP64__)
-#undef HAVE_SOCKLEN_T
-#endif
-
-#ifndef HAVE_SOCKLEN_T
-#define socklen_t int
-#endif
-
-#ifndef PF_INET
-#define PF_INET AF_INET
-#endif
-#ifndef PF_UNIX
-#define PF_UNIX AF_UNIX
-#endif
-
 #ifdef HAVE_SSL
+#if defined(OPENVMS) && defined(__VAX)
+#define OPENSSL_NO_SHA512
+#endif
 #include <openssl/ssl.h>
 #include <openssl/rand.h>
 #endif
@@ -167,9 +199,9 @@ strtoq(const char *, char **, int);
 
 #include "setup.h"
 
-#define LINKS_COPYRIGHT "(C) 1999 - 2011 Mikulas Patocka"
-#define LINKS_COPYRIGHT_8859_1 "(C) 1999 - 2011 Mikul\341s Patocka"
-#define LINKS_COPYRIGHT_8859_2 "(C) 1999 - 2011 Mikul\341\271 Pato\350ka"
+#define LINKS_COPYRIGHT "(C) 1999 - 2017 Mikulas Patocka"
+#define LINKS_COPYRIGHT_8859_1 "(C) 1999 - 2017 Mikul\341s Patocka"
+#define LINKS_COPYRIGHT_8859_2 "(C) 1999 - 2017 Mikul\341\271 Pato\350ka"
 
 #ifdef HAVE_POINTER_COMPARISON_BUG
 #define DUMMY ((void *)1L)
@@ -177,24 +209,23 @@ strtoq(const char *, char **, int);
 #define DUMMY ((void *)-1L)
 #endif
 
+#ifdef OPENVMS
+#define RET_OK		1	/* SS$_NORMAL */
+#define RET_ERROR	3586	/* SS$_LINEABRT */
+#define RET_SIGNAL	2096	/* SS$_CANCEL */
+#define RET_SYNTAX	20	/* SS$_BADPARAM */
+#define RET_FATAL	44	/* SS$_ABORT */
+#else
 #define RET_OK		0
 #define RET_ERROR	1
 #define RET_SIGNAL	2
 #define RET_SYNTAX	3
 #define RET_FATAL	4
-
-#ifndef HAVE_MEMMOVE
-void *memmove(void *, const void *, size_t);
 #endif
+
 #ifndef HAVE_RAISE
 int raise(int);
 #endif
-#ifndef HAVE_STRTOUL
-unsigned long strtoul(const char *, char **, int);
-#endif
-#ifndef HAVE_STRERROR
-char *strerror(int);
-#endif
 #ifndef HAVE_GETTIMEOFDAY
 struct timeval {
 	long tv_sec;
@@ -206,6 +237,18 @@ struct timezone {
 };
 int gettimeofday(struct timeval *tv, struct timezone *tz);
 #endif
+#ifndef HAVE_TEMPNAM
+char *tempnam(const char *dir, const char *pfx);
+#endif
+#ifndef HAVE_STRDUP
+char *strdup(const char *s);
+#endif
+#ifndef HAVE_STRTOL
+long strtol(const char *, char **, int);
+#endif
+#ifndef HAVE_STRTOUL
+unsigned long strtoul(const char *, char **, int);
+#endif
 #ifndef HAVE_STRLEN
 size_t strlen(const char *s);
 #endif
@@ -224,59 +267,146 @@ int strcmp(const char *s1, const char *s2);
 #ifndef HAVE_STRNCMP
 int strncmp(const char *s1, const char *s2, size_t n);
 #endif
+#ifndef HAVE_STRCASECMP
+int strcasecmp(const char *s1, const char *s2);
+#endif
+#ifndef HAVE_STRNCASECMP
+int strncasecmp(const char *s1, const char *s2, size_t n);
+#endif
 #ifndef HAVE_STRCSPN
 size_t strcspn(const char *s, const char *reject);
 #endif
+#ifndef HAVE_STRSPN
+size_t strspn(const char *s, const char *accept);
+#endif
 #ifndef HAVE_STRSTR
 char *strstr(const char *haystack, const char *needle);
 #endif
-#ifndef HAVE_TEMPNAM
-char *tempnam(const char *dir, const char *pfx);
+#ifndef HAVE_MEMCHR
+void *memchr(const void *s, int c, size_t length);
+#endif
+#ifndef HAVE_MEMCMP
+int memcmp(const void *, const void *, size_t);
+#endif
+#ifndef HAVE_MEMCPY
+void *memcpy(void *, const void *, size_t);
+#endif
+#ifndef HAVE_MEMMOVE
+void *memmove(void *, const void *, size_t);
+#endif
+#ifndef HAVE_MEMSET
+void *memset(void *, int, size_t);
+#endif
+#ifndef HAVE_STRERROR
+char *strerror(int);
+#endif
+
+#define EINTRLOOPX(ret_, call_, x_)			\
+do {							\
+	(ret_) = (call_);				\
+} while ((ret_) == (x_) && errno == EINTR)
+
+#define EINTRLOOP(ret_, call_)	EINTRLOOPX(ret_, call_, -1)
+
+#define ENULLLOOP(ret_, call_)				\
+do {							\
+	errno = 0;					\
+	(ret_) = (call_);				\
+} while (!(ret_) && errno == EINTR)
+
+#if defined(HAVE_PTHREADS) && defined(HAVE_PTHREAD_SIGMASK)
+static inline int do_sigprocmask(int how, const sigset_t *set, sigset_t *oset)
+{
+	int r;
+	r = pthread_sigmask(how, set, oset);
+	if (r) {
+		errno = r;
+		return -1;
+	}
+	return 0;
+}
+#elif defined(HAVE_SIGPROCMASK)
+#define do_sigprocmask	sigprocmask
+#else
+#ifdef sigset_t
+#undef sigset_t
+#endif
+#define sigset_t	int
+#ifndef SIG_BLOCK
+#define SIG_BLOCK	0
+#endif
+#ifndef SIG_SETMASK
+#define SIG_SETMASK	2
+#endif
+static inline int do_sigprocmask(int how, const sigset_t *set, sigset_t *oset) +{
+	sigset_t old = 0;
+#if defined(HAVE_SIGBLOCK) && defined(HAVE_SIGSETMASK)
+	switch (how) {
+		case SIG_BLOCK:
+			old = sigblock(*set);
+			break;
+		case SIG_SETMASK:
+			old = sigsetmask(*set);
+			break;
+	}
+#endif
+	if (oset) *oset = old;
+	return 0;
+}
+#ifdef sigdelset
+#undef sigdelset
+#endif
+#define sigdelset(x, s)		(*(x) &= ~(1 << (s)), 0)
+#ifndef HAVE_SIGDELSET
+#define HAVE_SIGDELSET		1
+#endif
+#ifdef HAVE_SIGFILLSET
+#undef HAVE_SIGFILLSET
+#endif
+#endif
+#ifndef HAVE_SIGFILLSET
+int sigfillset(sigset_t *set);
 #endif
 
 #define option option_dirty_workaround_for_name_clash_with_include_on_cygwin
 #define table table_dirty_workaround_for_name_clash_with_libraries_on_macos
 #define scroll scroll_dirty_workaround_for_name_clash_with_libraries_on_macos
+#define list list_dirty_workaround_for_name_clash_in_stl_with_class_list
 
 /* error.c */
 
+#if defined(DOS)
+#define ANSI_BELL		"\007"
+#define ANSI_SET_BOLD		""
+#define ANSI_CLEAR_BOLD		""
+#else
+#define ANSI_BELL		"\007"
+#define ANSI_SET_BOLD		"\033[1m"
+#define ANSI_CLEAR_BOLD		"\033[0m"
+#endif
+
 void do_not_optimize_here(void *p);
 void check_memory_leaks();
-void error(unsigned char *, ...);
-void debug_msg(unsigned char *, ...);
-void int_error(unsigned char *, ...);
+void error(char *, ...) PRINTF_FORMAT(1, 2);
+void debug_msg(char *, ...) PRINTF_FORMAT(1, 2);
+void int_error(char *, ...) PRINTF_FORMAT(1, 2);
 extern int errline;
 extern unsigned char *errfile;
 
 #define internal while (1) errfile = __FILE__, errline = __LINE__, int_error
 #define debug errfile = __FILE__, errline = __LINE__, debug_msg
 
-#ifdef SPECIAL_MALLOC
-
-void *sp_malloc(size_t);
-void sp_free(void *);
-void *sp_realloc(void *, size_t);
-
-#define xmalloc sp_malloc
-#define xfree sp_free
-#define xrealloc sp_realloc
-
-#else
-
-#define xmalloc malloc
-#define xfree free
-#define xrealloc realloc
-
-#endif
-
 /* inline */
 
+#define cast_const_char	(const char *)
+#define cast_char 	(char *)
+#define cast_uchar 	(unsigned char *)
+
 void fatal_tty_exit(void);
 
 #ifdef LEAK_DEBUG
 
-extern long mem_amount;
-extern long last_mem_amount;
+extern my_uintptr_t mem_amount;
 
 #ifndef LEAK_DEBUG_LIST
 struct alloc_header {
@@ -347,7 +477,7 @@ static inline void *mem_realloc(void *p, size_t size)
 {
 	if (p == DUMMY) return mem_alloc(size);
 	if (!p) {
-		internal("mem_realloc(NULL, %d)", size);
+		internal("mem_realloc(NULL, %lu)", (unsigned long)size);
 		return NULL;
 	}
 	if (!size) {
@@ -371,6 +501,9 @@ static inline void set_mem_comment(void *p, unsigned char *c, int l) {}
 
 #endif
 
+#define mem_alloc_mayfail	mem_alloc
+#define mem_realloc_mayfail	mem_realloc
+
 static inline unsigned char upcase(unsigned char a)
 {
 	if (a >= 'a' && a <= 'z') a -= 0x20;
@@ -379,10 +512,10 @@ static inline unsigned char upcase(unsigned char a)
 
 static inline int xstrcmp(unsigned char *s1, unsigned char *s2)
 {
-        if (!s1 && !s2) return 0;
-        if (!s1) return -1;
-        if (!s2) return 1;
-        return strcmp(s1, s2);
+	if (!s1 && !s2) return 0;
+	if (!s1) return -1;
+	if (!s2) return 1;
+	return strcmp(s1, s2);
 }
 
 static inline int cmpbeg(unsigned char *str, unsigned char *b)
@@ -396,6 +529,7 @@ static inline int cmpbeg(unsigned char *str, unsigned char *b)
 static inline unsigned char *memacpy(const unsigned char *src, size_t len)
 {
 	unsigned char *m;
+	if (!(len + 1)) overalloc();
 	m = mem_alloc(len + 1);
 	memcpy(m, src, len);
 	m[len] = 0;
@@ -429,19 +563,38 @@ static inline unsigned char *debug_stracpy(unsigned char *f, int l, unsigned cha
 
 #endif
 
-static inline int snprint(unsigned char *s, int n, off_t num)
+
+static inline int snprint(unsigned char *s, int n, my_uintptr_t num)
 {
-	off_t q = 1;
+	my_uintptr_t q = 1;
 	while (q <= num / 10) q *= 10;
-	while (n-- > 1 && q) *(s++) = num / q + '0', num %= q, q /= 10;
+	while (n-- > 1 && q) {
+		*(s++) = num / q + '0';
+		num %= q;
+#ifdef __DECC_VER
+		do_not_optimize_here(&q);	/* compiler bug */
+#endif
+		q /= 10;
+	}
 	*s = 0;
 	return !!q;
 }
 
 static inline int snzprint(unsigned char *s, int n, off_t num)
 {
+	off_t q = 1;
 	if (n > 1 && num < 0) *(s++) = '-', num = -num, n--;
-	return snprint(s, n, num);
+	while (q <= num / 10) q *= 10;
+	while (n-- > 1 && q) {
+		*(s++) = num / q + '0';
+		num %= q;
+#ifdef __DECC_VER
+		do_not_optimize_here(&q);	/* compiler bug */
+#endif
+		q /= 10;
+	}
+	*s = 0;
+	return !!q;
 }
 
 static inline void add_to_strn(unsigned char **s, unsigned char *a)
@@ -469,7 +622,7 @@ static inline void add_to_str(unsigned char **s, int *l, unsigned char *a)
 {
 	unsigned char *p;
 	size_t ll = strlen(a);
-	if (ll > MAXINT || *l + ll > MAXINT) overalloc();
+	if (ll >= MAXINT || *l + ll >= MAXINT) overalloc();
 	p = *s;
 	if (((size_t)*l & ~(ALLOC_GR - 1)) != ((*l + ll) & ~(ALLOC_GR - 1))) {
 		p = *s = mem_realloc(*s, (*l + ll + ALLOC_GR) & ~(ALLOC_GR - 1));
@@ -480,7 +633,7 @@ static inline void add_to_str(unsigned char **s, int *l, unsigned char *a)
 static inline void add_bytes_to_str(unsigned char **s, int *l, unsigned char *a, int ll)
 {
 	unsigned char *p;
-	if (*l + ll < 0) overalloc();
+	if (ll < 0 || (size_t)*l + ll >= MAXINT) overalloc();
 	p = *s;
 	if ((*l & ~(ALLOC_GR - 1)) != ((*l + ll) & ~(ALLOC_GR - 1))) {
 		p = *s = mem_realloc(*s, (*l + ll + ALLOC_GR) & ~(ALLOC_GR - 1));
@@ -491,7 +644,7 @@ static inline void add_bytes_to_str(unsigned char **s, int *l, unsigned char *a,
 static inline void add_chr_to_str(unsigned char **s, int *l, unsigned char a)
 {
 	unsigned char *p;
-	if (*l + 1 < 0) overalloc();
+	if ((size_t)*l + 1 >= MAXINT) overalloc();
 	p = *s;
 	if ((*l & (ALLOC_GR - 1)) == ALLOC_GR - 1) {
 		p = *s = mem_realloc(*s, (*l + 1 + ALLOC_GR) & ~(ALLOC_GR - 1));
@@ -499,10 +652,16 @@ static inline void add_chr_to_str(unsigned char **s, int *l, unsigned char a)
 	*(p + *l) = a; *(p + ++(*l)) = 0;
 }
 
+static inline void add_unsigned_long_num_to_str(unsigned char **s, int *l, my_uintptr_t n)
+{
+	unsigned char a[64];
+	snprint(a, 64, n);
+	add_to_str(s, l, a);
+}
+
 static inline void add_num_to_str(unsigned char **s, int *l, off_t n)
 {
 	unsigned char a[64];
-	/*sprintf(a, "%d", n);*/
 	snzprint(a, 64, n);
 	add_to_str(s, l, a);
 }
@@ -537,6 +696,29 @@ static inline long strtolx(unsigned char *c, unsigned char **end)
 	return l;
 }
 
+#if defined(HAVE_STRTOLL) || defined(HAVE_STRTOQ) || defined(HAVE_STRTOIMAX)
+#define my_strtoll_t	longlong
+#else
+#define my_strtoll_t	long
+#endif
+
+static inline my_strtoll_t my_strtoll(unsigned char *string, unsigned char **end)
+{
+	my_strtoll_t f;
+	errno = 0;
+#if defined(HAVE_STRTOLL)
+	f = strtoll(cast_const_char string, (char **)(void *)end, 10);
+#elif defined(HAVE_STRTOQ)
+	f = strtoq(cast_const_char string, (char **)(void *)end, 10);
+#elif defined(HAVE_STRTOIMAX)
+	f = strtoimax(cast_const_char string, (char **)(void *)end, 10);
+#else
+	f = strtol(cast_const_char string, (char **)(void *)end, 10);
+#endif
+	if (f < 0 || errno) return -1;
+	return f;
+}
+
 /* Copies at most dst_size chars into dst. Ensures null termination of dst. */
 static inline unsigned char *safe_strncpy(unsigned char *dst, const unsigned char *src, size_t dst_size)
 {
@@ -585,7 +767,7 @@ struct xlist_head {
 static inline int isA(unsigned char c)
 {
 	return (c >= 'A' && c <= 'Z') || (c >= 'a' && c <= 'z') || (c >= '0' && c <= '9') ||
-	        c == '_' || c == '-';
+		c == '_' || c == '-';
 }
 
 static inline int casecmp(unsigned char *c1, unsigned char *c2, size_t len)
@@ -599,18 +781,53 @@ static inline int can_write(int fd)
 {
 	fd_set fds;
 	struct timeval tv = {0, 0};
+	int rs;
 	FD_ZERO(&fds);
+	if (fd < 0)
+		internal("can_write: handle %d", fd);
+	if (fd >= (int)FD_SETSIZE) {
+		error("too big handle %d", fd);
+		fatal_tty_exit();
+		exit(RET_FATAL);
+	}
 	FD_SET(fd, &fds);
-	return select(fd + 1, NULL, &fds, NULL, &tv);
+	EINTRLOOP(rs, select(fd + 1, NULL, &fds, NULL, &tv));
+	if (rs < 0) {
+		error("ERROR: select for write failed: %s", strerror(errno));
+		fatal_tty_exit();
+		exit(RET_FATAL);
+	}
+	return rs;
+}
+
+static inline int can_read_timeout(int fd, int sec)
+{
+	fd_set fds;
+	struct timeval tv;
+	int rs;
+	tv.tv_sec = sec;
+	tv.tv_usec = 0;
+	FD_ZERO(&fds);
+	if (fd < 0)
+		internal("can_read: handle %d", fd);
+	if (fd >= (int)FD_SETSIZE) {
+		error("too big handle %d", fd);
+		fatal_tty_exit();
+		exit(RET_FATAL);
+	}
+	FD_SET(fd, &fds);
+	EINTRLOOP(rs, select(fd + 1, &fds, NULL, NULL, &tv));
+	if (rs < 0) {
+		error("ERROR: select for read failed: %s", strerror(errno));
+		fatal_tty_exit();
+		exit(RET_FATAL);
+	}
+	return rs;
 }
 
 static inline int can_read(int fd)
 {
-	fd_set fds;
-	struct timeval tv = {0, 0};
-	FD_ZERO(&fds);
-	FD_SET(fd, &fds);
-	return select(fd + 1, &fds, NULL, NULL, &tv);
+	return can_read_timeout(fd, 0);
 }
 
 #define CI_BYTES	1
@@ -625,6 +842,10 @@ static inline int can_read(int fd)
 
 /* os_dep.c */
 
+#if defined(HAVE_GPM_H) && defined(HAVE_LIBGPM)
+#define USE_GPM
+#endif
+
 struct terminal;
 
 struct open_in_new {
@@ -635,72 +856,107 @@ struct open_in_new {
 
 void close_fork_tty();
 int get_system_env();
+int is_twterm();
+int is_screen();
 int is_xterm();
-int can_twterm();
 int get_terminal_size(int, int *, int *);
 void handle_terminal_resize(int, void (*)());
 void unhandle_terminal_resize(int);
 void set_bin(int);
 int c_pipe(int *);
+void set_nonblock(int);
 int get_input_handle();
 int get_output_handle();
 int get_ctl_handle();
+#ifdef OS_SETRAW
+int setraw(int ctl, int save);
+void setcooked(int ctl);
+#endif
 void want_draw();
 void done_draw();
 void terminate_osdep();
 void *handle_mouse(int, void (*)(void *, unsigned char *, int), void *);
 void unhandle_mouse(void *);
-int check_file_name(unsigned char *);
+void add_gpm_version(unsigned char **s, int *l);
+#ifdef OPENVMS
+extern int vms_thread_high_priority;
+#endif
 int start_thread(void (*)(void *, int), void *, int);
-char *get_clipboard_text();
-void set_clipboard_text(char *);
+unsigned char *get_clipboard_text();
+void set_clipboard_text(unsigned char *);
 void set_window_title(unsigned char *);
 unsigned char *get_window_title();
 int is_safe_in_shell(unsigned char);
 unsigned char *escape_path(unsigned char *);
 void check_shell_security(unsigned char **);
+void check_filename(unsigned char **);
 int check_shell_url(unsigned char *);
+void do_signal(int sig, void (*handler)(int));
+void ignore_signals(void);
 void block_stdin();
 void unblock_stdin();
 void init_os(void);
 void init_os_terminal(void);
 void get_path_to_exe(void);
+int os_get_system_name(unsigned char *buffer);
 unsigned char *os_conv_to_external_path(unsigned char *, unsigned char *);
 unsigned char *os_fixup_external_program(unsigned char *);
-int exe(char *);
+int exe(unsigned char *);
 int resize_window(int, int);
-int can_resize_window(int);
+int can_resize_window(struct terminal *);
 int can_open_os_shell(int);
 struct open_in_new *get_open_in_new(int);
 void set_highpri();
 #ifdef HAVE_OPEN_PREALLOC
-int can_prealloc(char *);
-int open_prealloc(char *, int, int, off_t);
+int can_prealloc(unsigned char *);
+int open_prealloc(unsigned char *, int, int, off_t);
 void prealloc_truncate(int, off_t);
 #else
 static inline void prealloc_truncate(int x, off_t y) {}
 #endif
-void os_cfmakeraw(struct termios *t);
 
 /* select.c */
 
 #ifndef FD_SETSIZE
-#define FD_SETSIZE 1024
+#define FD_SETSIZE	(sizeof(fd_set) * 8)
 #endif
 
+#define NBIT(p)		(sizeof((p)->fds_bits[0]) * 8)
+
+#ifndef FD_SET
+#define FD_SET(n, p)	((p)->fds_bits[(n)/NBIT(p)] |= (1 << ((n) % NBIT(p))))
+#endif
+#ifndef FD_CLR
+#define FD_CLR(n, p)	((p)->fds_bits[(n)/NBIT(p)] &= ~(1 << ((n) % NBIT(p))))
+#endif
+#ifndef FD_ISSET
+#define FD_ISSET(n, p)	((p)->fds_bits[(n)/NBIT(p)] & (1 << ((n) % NBIT(p))))
+#endif
+#ifndef FD_ZERO
+#define FD_ZERO(p)	memset((void *)(p), 0, sizeof(*(p)))
+#endif
+
+#ifdef HAVE_LONG_LONG
+typedef long long ttime;
+typedef unsigned long long uttime;
+typedef unsigned long long tcount;
+#else
 typedef long ttime;
 typedef unsigned long uttime;
-typedef unsigned tcount;
+typedef unsigned long tcount;
+#endif
 
-extern int terminate;
+extern int terminate_loop;
 
-long select_info(int);
+unsigned long select_info(int);
 void select_loop(void (*)());
-int register_bottom_half(void (*)(void *), void *);
+void register_bottom_half(void (*)(void *), void *);
+void unregister_bottom_half(void (*)(void *), void *);
 void check_bottom_halves();
 int install_timer(ttime, void (*)(void *), void *);
 void kill_timer(int);
 ttime get_time();
+void portable_sleep(unsigned msec);
 
 #define H_READ	0
 #define H_WRITE	1
@@ -709,8 +965,11 @@ ttime get_time();
 
 void *get_handler(int, int);
 void set_handlers(int, void (*)(void *), void (*)(void *), void (*)(void *), void *);
+void clear_events(int, int);
 void install_signal_handler(int, void (*)(void *), void *, int);
 void interruptible_signal(int sig, int in);
+void block_signals(int except1, int except2);
+void unblock_signals(void);
 void set_sigcld();
 
 /* dns.c */
@@ -733,6 +992,7 @@ struct cache_entry {
 	unsigned char *redirect;
 	int redirect_get;
 	off_t length;
+	off_t max_length;
 	int incomplete;
 	int tgc;
 	unsigned char *last_modified;
@@ -757,7 +1017,7 @@ struct fragment {
 extern int page_size;
 
 void init_cache();
-long cache_info(int);
+my_uintptr_t cache_info(int);
 int find_in_cache(unsigned char *, struct cache_entry **);
 int get_cache_entry(unsigned char *, struct cache_entry **);
 int add_fragment(struct cache_entry *, off_t, unsigned char *, off_t);
@@ -786,7 +1046,7 @@ struct remaining_info {
 	ttime elapsed;
 	ttime last_time;
 	ttime dis_b;
-	int data_in_secs[CURRENT_SPD_SEC];
+	off_t data_in_secs[CURRENT_SPD_SEC];
 	int timer;
 };
 
@@ -806,6 +1066,7 @@ struct connection {
 	void *dnsquery;
 	pid_t pid;
 	int tries;
+	tcount netcfg_stamp;
 	struct list_head statuss;
 	void *info;
 	void *buffer;
@@ -820,10 +1081,12 @@ struct connection {
 	int detached;
 #ifdef HAVE_SSL
 	SSL *ssl;
-	int no_tsl;
+	int no_tls;
 #endif
 };
 
+extern tcount netcfg_stamp;
+
 extern struct list_head queue;
 
 struct k_conn {
@@ -835,6 +1098,7 @@ struct k_conn {
 	int conn;
 	ttime timeout;
 	ttime add_time;
+	int protocol_data;
 };
 
 extern struct list_head keepalive_connections;
@@ -863,7 +1127,7 @@ static inline int getpri(struct connection *c)
 #define S_PROC		7
 #define S_TRANS		8
 
-#define S_OK			(-2000000000)
+#define S__OK			(-2000000000)
 #define S_INTERRUPTED		(-2000000001)
 #define S_EXCEPT		(-2000000002)
 #define S_INTERNAL		(-2000000003)
@@ -922,11 +1186,11 @@ struct status {
 };
 
 void check_queue(void *dummy);
-long connect_info(int);
+unsigned long connect_info(int);
 void send_connection_info(struct connection *c);
 void setcstate(struct connection *c, int);
-int get_keepalive_socket(struct connection *c);
-void add_keepalive_socket(struct connection *c, ttime);
+int get_keepalive_socket(struct connection *c, int *protocol_data);
+void add_keepalive_socket(struct connection *c, ttime timeout, int protocol_data);
 void run_connection(struct connection *c);
 int is_connection_restartable(struct connection *c);
 void retry_connection(struct connection *c);
@@ -937,6 +1201,7 @@ void change_connection(struct status *, struct status *, int);
 void detach_connection(struct status *, off_t);
 void abort_all_connections();
 void abort_background_connections();
+void abort_all_keepalive_connections();
 int is_entry_used(struct cache_entry *);
 void connection_timeout(struct connection *);
 void set_timeout(struct connection *);
@@ -954,7 +1219,8 @@ void free_blacklist();
 
 struct session;
 
-#define POST_CHAR 1
+#define POST_CHAR		1
+#define POST_CHAR_STRING	"\001"
 
 static inline int end_of_dir(unsigned char c)
 {
@@ -975,6 +1241,7 @@ unsigned char *join_urls(unsigned char *, unsigned char *);
 unsigned char *translate_url(unsigned char *, unsigned char *);
 unsigned char *extract_position(unsigned char *);
 void get_filename_from_url(unsigned char *, unsigned char **, int *);
+int url_not_saveable(unsigned char *);
 void add_conv_str(unsigned char **s, int *l, unsigned char *b, int ll, int encode_special);
 
 /* connect.c */
@@ -1007,6 +1274,7 @@ void cleanup_cookies();
 
 unsigned char *parse_http_header(unsigned char *, unsigned char *, unsigned char **);
 unsigned char *parse_header_param(unsigned char *, unsigned char *);
+int get_http_code(unsigned char *head, int *code, int *version);
 void http_func(struct connection *);
 void proxy_func(struct connection *);
 
@@ -1110,10 +1378,10 @@ extern unsigned char term_seq_tw_mouse[];
 typedef unsigned short chr;
 
 struct event {
-	long ev;
-	long x;
-	long y;
-	long b;
+	int ev;
+	int x;
+	int y;
+	my_intptr_t b;
 };
 
 #define EV_INIT		0
@@ -1134,7 +1402,7 @@ struct window {
 
 #define MAX_TERM_LEN	32	/* this must be multiple of 8! (alignment problems) */
 
-#define MAX_CWD_LEN	256	/* this must be multiple of 8! (alignment problems) */	
+#define MAX_CWD_LEN	4096	/* this must be multiple of 8! (alignment problems) */	
 
 #define ENV_XWIN	1
 #define ENV_SCREEN	2
@@ -1258,7 +1526,7 @@ void set_language(int);
 int n_languages();
 unsigned char *language_name(int);
 
-#define _(_x_, _y_) get_text_translation(_x_, _y_)
+#define _(x_, y_) get_text_translation(x_, y_)
 #define TEXT_(x) (dummyarray + x) /* TEXT causes name clash on windows */
 
 /* af_unix.c */
@@ -1284,7 +1552,7 @@ void shrink_memory(int);
 struct assoc {
 	struct assoc *next;
 	struct assoc *prev;
-	tcount cnt;
+	unsigned cnt;
 	unsigned char *label;
 	unsigned char *ct;
 	unsigned char *prog;
@@ -1298,7 +1566,7 @@ struct assoc {
 struct extension {
 	struct extension *next;
 	struct extension *prev;
-	tcount cnt;
+	unsigned cnt;
 	unsigned char *ext;
 	unsigned char *ct;
 };
@@ -1463,7 +1731,7 @@ struct document_options {
 	int xw, yw; /* size of window */
 	int xp, yp; /* pos of window */
 	int col, cp, assume_cp, hard_assume;
-	int tables, frames, images, margin;  /* if you add anything, fix it in compare_opt */
+	int tables, frames, images, margin;	/* if you add anything, fix it in compare_opt */
 	int plain;
 	int num_links, table_order;
 	struct rgb default_fg;
@@ -1471,7 +1739,7 @@ struct document_options {
 	struct rgb default_link;
 	struct rgb default_vlink;
 	unsigned char *framename;
-	int real_cp;    /* codepage of document. Does not really belong here. Must not be compared. Used only in get_attr_val */
+	int real_cp;	/* codepage of document. Does not really belong here. Must not be compared. Used only in get_attr_val */
 };
 
 static inline void ds2do(struct document_setup *ds, struct document_options *doo)
@@ -1748,6 +2016,9 @@ struct menu {
 	void *data;
 	struct window *win;
 	struct menu_item *items;
+	void (*free_function)(void *);
+	void *free_data;
+	unsigned char hotkeys[1];
 };
 
 struct mainmenu {
@@ -1757,6 +2028,7 @@ struct mainmenu {
 	void *data;
 	struct window *win;
 	struct menu_item *items;
+	unsigned char hotkeys[1];
 };
 
 struct history_item {
@@ -1833,7 +2105,7 @@ struct dialog_data {
 struct menu_item *new_menu(int);
 void add_to_menu(struct menu_item **, unsigned char *, unsigned char *, unsigned char *, void (*)(struct terminal *, void *, void *), void *, int);
 void do_menu(struct terminal *, struct menu_item *, void *);
-void do_menu_selected(struct terminal *, struct menu_item *, void *, int);
+void do_menu_selected(struct terminal *, struct menu_item *, void *, int, void (*)(void *), void *);
 void do_mainmenu(struct terminal *, struct menu_item *, void *, int);
 void do_dialog(struct terminal *, struct dialog *, struct memory_list *);
 int check_number(struct dialog_data *, struct dialog_item_data *);
@@ -2125,7 +2397,7 @@ void free_table_cache();
 
 struct conv_table *get_convert_table(unsigned char *, int, int, int *, int *, int);
 extern int format_cache_entries;
-long formatted_info(int);
+unsigned long formatted_info(int);
 void shrink_format_cache(int);
 void count_format_cache();
 void delete_unused_format_cache_entries();
@@ -2133,6 +2405,7 @@ void format_cache_reactivate(struct f_data *);
 struct part *format_html_part(unsigned char *, unsigned char *, int, int, int, struct f_data *, int, int, unsigned char *, int);
 void cached_format_html(struct view_state *, struct f_data_c *, struct document_options *);
 void html_interpret(struct session *);
+struct link *get_link_at_location(struct f_data *f, int x, int y);
 void get_search_data(struct f_data *);
 
 /* html_tbl.c */
@@ -2170,6 +2443,7 @@ void save_url_history();
 extern int anonymous;
 
 extern unsigned char system_name[];
+extern unsigned char compiler_name[];
 
 extern unsigned char *links_home;
 extern int first_use;
