@@ -7,7 +7,7 @@ void unhandle_basic_signals(struct terminal *);
 void sig_terminate(struct terminal *t)
 {
 	unhandle_basic_signals(t);
-	terminate = 1;
+	terminate_loop = 1;
 	retval = RET_SIGNAL;
 }
 
@@ -15,7 +15,7 @@ void sig_intr(struct terminal *t)
 {
 	if (!t) {
 		unhandle_basic_signals(t);
-		terminate = 1;
+		terminate_loop = 1;
 	} else {
 		unhandle_basic_signals(t);
 		exit_prog(t, NULL, NULL);
@@ -38,21 +38,29 @@ void sig_tstp(struct terminal *t)
 {
 #if defined(SIGSTOP) && !defined(NO_CTRL_Z)
 #if defined(SIGCONT) && defined(SIGTTOU) && defined(HAVE_GETPID)
-	pid_t pid = getpid();
-	pid_t newpid;
+	pid_t pid, newpid;
+	EINTRLOOP(pid, getpid());
 #endif
 	block_itrm(1);
-#if defined(SIGCONT) && defined(SIGTTOU)
-	if (!(newpid = fork())) {
+#if defined(SIGCONT) && defined(SIGTTOU) && defined(HAVE_GETPID)
+	EINTRLOOP(newpid, fork());
+	if (!newpid) {
 		while (1) {
+			int rr;
 			sleep(1);
-			kill(pid, SIGCONT);
+			EINTRLOOP(rr, kill(pid, SIGCONT));
 		}
 	}
 #endif
-	raise(SIGSTOP);
-#if defined(SIGCONT) && defined(SIGTTOU)
-	if (newpid != -1) kill(newpid, SIGKILL);
+	{
+		int rr;
+		EINTRLOOP(rr, raise(SIGSTOP));
+	}
+#if defined(SIGCONT) && defined(SIGTTOU) && defined(HAVE_GETPID)
+	if (newpid != -1) {
+		int rr;
+		EINTRLOOP(rr, kill(newpid, SIGKILL));
+	}
 #endif
 #endif
 	if (fg_poll_timer != -1) kill_timer(fg_poll_timer);
@@ -70,7 +78,7 @@ void poll_fg(void *t)
 	if (r == -2) {
 		/* This will unblock externally spawned viewer, if it exists */
 #ifdef SIGCONT
-		kill(0, SIGCONT);
+		EINTRLOOP(r, kill(0, SIGCONT));
 #endif
 	}
 }
@@ -85,6 +93,9 @@ void handle_basic_signals(struct terminal *term)
 	install_signal_handler(SIGHUP, (void (*)(void *))sig_intr, term, 0);
 	install_signal_handler(SIGINT, (void (*)(void *))sig_ctrl_c, term, 0);
 	/*install_signal_handler(SIGTERM, (void (*)(void *))sig_terminate, term, 0);*/
+#ifdef WIN32
+	install_signal_handler(SIGQUIT, (void (*)(void *))sig_terminate, term, 0);
+#endif
 #ifdef SIGTSTP
 	install_signal_handler(SIGTSTP, (void (*)(void *))sig_tstp, term, 0);
 #endif
@@ -144,16 +155,17 @@ int terminal_pipe[2];
 int attach_terminal(int in, int out, int ctl, void *info, int len)
 {
 	struct terminal *term;
-	fcntl(terminal_pipe[0], F_SETFL, O_NONBLOCK);
-	fcntl(terminal_pipe[1], F_SETFL, O_NONBLOCK);
+	int rs;
+	set_nonblock(terminal_pipe[0]);
+	set_nonblock(terminal_pipe[1]);
 	handle_trm(in, out, out, terminal_pipe[1], ctl, info, len);
 	mem_free(info);
 	if ((term = init_term(terminal_pipe[0], out, win_func))) {
 		handle_basic_signals(term);	/* OK, this is race condition, but it must be so; GPM installs it's own buggy TSTP handler */
 		return terminal_pipe[1];
 	}
-	close(terminal_pipe[0]);
-	close(terminal_pipe[1]);
+	EINTRLOOP(rs, close(terminal_pipe[0]));
+	EINTRLOOP(rs, close(terminal_pipe[1]));
 	return -1;
 }
 
@@ -183,8 +195,11 @@ void end_dump(struct status *stat, void *p)
 			struct fragment *frag;
 			nextfrag:
 			foreach(frag, ce->frag) if (frag->offset <= dump_pos && frag->offset + frag->length > dump_pos) {
-				int l = frag->length - (dump_pos - frag->offset);
-				int w = hard_write(oh, frag->data + dump_pos - frag->offset, l);
+				off_t l;
+				int w;
+				l = frag->length - (dump_pos - frag->offset);
+				if (l > MAXINT) l = MAXINT;
+				w = hard_write(oh, frag->data + dump_pos - frag->offset, l);
 				if (w != l) {
 					detach_connection(stat, dump_pos);
 					if (w < 0) fprintf(stderr, "Error writing to stdout: %s.\n", strerror(errno));
@@ -222,20 +237,21 @@ void end_dump(struct status *stat, void *p)
 		o.framename = "";
 		init_vs(vs, stat->ce->url);
 		cached_format_html(vs, &fd, &o);
-		dump_to_file(fd.f_data, oh);
+		if (dump_to_file(fd.f_data, oh))
+			fprintf(stderr, "Error writing to stdout.\n");
 		detach_formatted(&fd);
 		destroy_vs(vs);
 		mem_free(vs);
 
 	}
-	if (stat->state != S_OK) {
+	if (stat->state != S__OK) {
 		unsigned char *m = get_err_msg(stat->state);
 		fprintf(stderr, "%s\n", get_english_translation(m));
 		retval = RET_ERROR;
 		goto terminate;
 	}
 	terminate:
-	terminate = 1;
+	terminate_loop = 1;
 }
 
 int g_argc;
@@ -251,6 +267,8 @@ void init()
 	void *info;
 	int len;
 	unsigned char *u;
+	int rs;
+
 	init_trans();
 	set_sigcld();
 	init_home();
@@ -271,8 +289,8 @@ void init()
 		init_os_terminal();
 	}
 	if (!no_connect && (uh = bind_to_af_unix()) != -1) {
-		close(terminal_pipe[0]);
-		close(terminal_pipe[1]);
+		EINTRLOOP(rs, close(terminal_pipe[0]));
+		EINTRLOOP(rs, close(terminal_pipe[1]));
 		if (!(info = create_session_info(base_session, u, &len))) {
 			retval = RET_FATAL;
 			goto ttt;
@@ -291,18 +309,18 @@ void init()
 	u = parse_options(g_argc - 1, g_argv + 1);
 	if (!u) {
 		ttt:
-		terminate = 1;
+		terminate_loop = 1;
 		return;
 	}
 	if (!dmp) {
 		if (!((info = create_session_info(base_session, u, &len)) && attach_terminal(get_input_handle(), get_output_handle(), get_ctl_handle(), info, len) != -1)) {
 			retval = RET_FATAL;
-			terminate = 1;
+			terminate_loop = 1;
 		}
 	} else {
 		unsigned char *uu, *wd;
-		close(terminal_pipe[0]);
-		close(terminal_pipe[1]);
+		EINTRLOOP(rs, close(terminal_pipe[0]));
+		EINTRLOOP(rs, close(terminal_pipe[1]));
 		if (!*u) {
 			fprintf(stderr, "URL expected after %s\n.", dmp == D_DUMP ? "-dump" : "-source");
 			retval = RET_SYNTAX;
