@@ -1,17 +1,113 @@
 #include "links.h"
 
-/*static inline struct session *get_term_session(struct terminal *term)
+static unsigned char *version_texts[] = {
+	TEXT_(T_LINKS_VERSION),
+	TEXT_(T_OPERATING_SYSTEM_TYPE),
+	TEXT_(T_OPERATING_SYSTEM_VERSION),
+	TEXT_(T_COMPILER),
+	TEXT_(T_WORD_SIZE),
+	TEXT_(T_ENCRYPTION),
+#if defined(__linux__) || defined(__SPAD__) || defined(USE_GPM)
+	TEXT_(T_GPM_MOUSE_DRIVER),
+#endif
+#ifdef OS2
+	TEXT_(T_XTERM_FOR_OS2),
+#endif
+	NULL,
+};
+
+void add_and_pad(unsigned char **s, int *l, struct terminal *term, unsigned char *str, int maxlen)
 {
-	if ((void *)term->windows.prev == &term->windows) {
-		internal("terminal has no windows");
-		return NULL;
+	unsigned char *x = _(str, term);
+	int len = strlen(x);
+	add_to_str(s, l, x);
+	add_to_str(s, l, ":  ");
+	while (len++ < maxlen) add_chr_to_str(s, l, ' ');
+}
+
+void menu_version(struct terminal *term)
+{
+	int i;
+	int maxlen = 0;
+	unsigned char *s;
+	int l;
+	unsigned char **text_ptr;
+	for (i = 0; version_texts[i]; i++) {
+		unsigned char *t = _(version_texts[i], term);
+		int tl = strlen(t);
+		if (tl > maxlen)
+			maxlen = tl;
 	}
-	return ((struct window *)term->windows.prev)->data;
-}*/
+
+	s = init_str();
+	l = 0;
+	text_ptr = version_texts;
+
+	add_and_pad(&s, &l, term, *text_ptr++, maxlen);
+	add_to_str(&s, &l, VERSION_STRING);
+	add_to_str(&s, &l, "\n");
+
+	add_and_pad(&s, &l, term, *text_ptr++, maxlen);
+	add_to_str(&s, &l, SYSTEM_NAME);
+	add_to_str(&s, &l, "\n");
+
+	add_and_pad(&s, &l, term, *text_ptr++, maxlen);
+	add_to_str(&s, &l, system_name);
+	add_to_str(&s, &l, "\n");
+
+	add_and_pad(&s, &l, term, *text_ptr++, maxlen);
+	add_to_str(&s, &l, compiler_name);
+	add_to_str(&s, &l, "\n");
+
+	add_and_pad(&s, &l, term, *text_ptr++, maxlen);
+	add_to_str(&s, &l, _(TEXT_(T_MEMORY), term));
+	add_to_str(&s, &l, " ");
+	add_num_to_str(&s, &l, sizeof(void *) * 8);
+	add_to_str(&s, &l, "-bit, ");
+	add_to_str(&s, &l, _(TEXT_(T_FILE_SIZE), term));
+	add_to_str(&s, &l, " ");
+	add_num_to_str(&s, &l, sizeof(off_t) * 8 /*- ((off_t)-1 < 0)*/);
+	add_to_str(&s, &l, "-bit");
+	add_to_str(&s, &l, "\n");
+
+	add_and_pad(&s, &l, term, *text_ptr++, maxlen);
+#ifdef HAVE_SSL
+	add_to_str(&s, &l, (unsigned char *)SSLeay_version(SSLEAY_VERSION));
+#else
+	add_to_str(&s, &l, _(TEXT_(T_NO), term));
+#endif
+	add_to_str(&s, &l, "\n");
+
+#if defined(__linux__) || defined(__SPAD__) || defined(USE_GPM)
+	add_and_pad(&s, &l, term, *text_ptr++, maxlen);
+#ifdef USE_GPM
+	add_gpm_version(&s, &l);
+#else
+	add_to_str(&s, &l, _(TEXT_(T_NO), term));
+#endif
+	add_to_str(&s, &l, "\n");
+#endif
+
+#ifdef OS2
+	add_and_pad(&s, &l, term, *text_ptr++, maxlen);
+#ifdef X2
+	add_to_str(&s, &l, _(TEXT_(T_YES), term));
+#else
+	add_to_str(&s, &l, _(TEXT_(T_NO), term));
+#endif
+	add_to_str(&s, &l, "\n");
+#endif
+
+	s[l - 1] = 0;
+	if (*text_ptr)
+		internal("menu_version: text mismatched");
+	
+	msg_box(term, getml(s, NULL), TEXT_(T_VERSION_INFORMATION), AL_LEFT, s, NULL, 1, TEXT_(T_OK), NULL, B_ENTER | B_ESC);
+}
 
 void menu_about(struct terminal *term, void *d, struct session *ses)
 {
-	msg_box(term, NULL, TEXT_(T_ABOUT), AL_CENTER, TEXT_(T_LINKS__LYNX_LIKE), NULL, 1, TEXT_(T_OK), NULL, B_ENTER | B_ESC);
+	msg_box(term, NULL, TEXT_(T_ABOUT), AL_CENTER, TEXT_(T_LINKS__LYNX_LIKE), term, 2, TEXT_(T_OK), NULL, B_ENTER | B_ESC, TEXT_(T_VERSION), menu_version, 0);
 }
 
 void menu_keys(struct terminal *term, void *d, struct session *ses)
@@ -117,6 +213,7 @@ void cache_inf(struct terminal *term, void *d, struct session *ses)
 	unsigned char *a1, *a2, *a3, *a4, *a5, *a6, *a7, *a8, *a9, *a10, *a11, *a12, *a13, *a14, *a15, *a16;
 	int l = 0;
 	struct refresh *r;
+
 	r = mem_alloc(sizeof(struct refresh));
 	r->term = term;
 	r->win = NULL;
@@ -125,24 +222,24 @@ void cache_inf(struct terminal *term, void *d, struct session *ses)
 	r->data = d;
 	r->timer = -1;
 	l = 0;
-	l = 0, a1 = init_str(); add_to_str(&a1, &l, ": "); add_num_to_str(&a1, &l, select_info(CI_FILES));add_to_str(&a1, &l, " ");
-	l = 0, a2 = init_str(); add_to_str(&a2, &l, ", "); add_num_to_str(&a2, &l, select_info(CI_TIMERS));add_to_str(&a2, &l, " ");
+	l = 0, a1 = init_str(); add_to_str(&a1, &l, ": "); add_unsigned_long_num_to_str(&a1, &l, select_info(CI_FILES));add_to_str(&a1, &l, " ");
+	l = 0, a2 = init_str(); add_to_str(&a2, &l, ", "); add_unsigned_long_num_to_str(&a2, &l, select_info(CI_TIMERS));add_to_str(&a2, &l, " ");
 	l = 0, a3 = init_str(); add_to_str(&a3, &l, ".\n");
 
-	l = 0, a4 = init_str(); add_to_str(&a4, &l, ": "); add_num_to_str(&a4, &l, connect_info(CI_FILES));add_to_str(&a4, &l, " ");
-	l = 0, a5 = init_str(); add_to_str(&a5, &l, ", "); add_num_to_str(&a5, &l, connect_info(CI_CONNECTING));add_to_str(&a5, &l, " ");
-	l = 0, a6 = init_str(); add_to_str(&a6, &l, ", "); add_num_to_str(&a6, &l, connect_info(CI_TRANSFER));add_to_str(&a6, &l, " ");
-	l = 0, a7 = init_str(); add_to_str(&a7, &l, ", "); add_num_to_str(&a7, &l, connect_info(CI_KEEP));add_to_str(&a7, &l, " ");
+	l = 0, a4 = init_str(); add_to_str(&a4, &l, ": "); add_unsigned_long_num_to_str(&a4, &l, connect_info(CI_FILES));add_to_str(&a4, &l, " ");
+	l = 0, a5 = init_str(); add_to_str(&a5, &l, ", "); add_unsigned_long_num_to_str(&a5, &l, connect_info(CI_CONNECTING));add_to_str(&a5, &l, " ");
+	l = 0, a6 = init_str(); add_to_str(&a6, &l, ", "); add_unsigned_long_num_to_str(&a6, &l, connect_info(CI_TRANSFER));add_to_str(&a6, &l, " ");
+	l = 0, a7 = init_str(); add_to_str(&a7, &l, ", "); add_unsigned_long_num_to_str(&a7, &l, connect_info(CI_KEEP));add_to_str(&a7, &l, " ");
 	l = 0, a8 = init_str(); add_to_str(&a8, &l, ".\n");
 
-	l = 0, a9 = init_str(); add_to_str(&a9, &l, ": "); add_num_to_str(&a9, &l, cache_info(CI_BYTES));add_to_str(&a9, &l, " ");
-	l = 0, a10 = init_str(); add_to_str(&a10, &l, ", "); add_num_to_str(&a10, &l, cache_info(CI_FILES));add_to_str(&a10, &l, " ");
-	l = 0, a11 = init_str(); add_to_str(&a11, &l, ", "); add_num_to_str(&a11, &l, cache_info(CI_LOCKED));add_to_str(&a11, &l, " ");
-	l = 0, a12 = init_str(); add_to_str(&a12, &l, ", "); add_num_to_str(&a12, &l, cache_info(CI_LOADING));add_to_str(&a12, &l, " ");
+	l = 0, a9 = init_str(); add_to_str(&a9, &l, ": "); add_unsigned_long_num_to_str(&a9, &l, cache_info(CI_BYTES));add_to_str(&a9, &l, " ");
+	l = 0, a10 = init_str(); add_to_str(&a10, &l, ", "); add_unsigned_long_num_to_str(&a10, &l, cache_info(CI_FILES));add_to_str(&a10, &l, " ");
+	l = 0, a11 = init_str(); add_to_str(&a11, &l, ", "); add_unsigned_long_num_to_str(&a11, &l, cache_info(CI_LOCKED));add_to_str(&a11, &l, " ");
+	l = 0, a12 = init_str(); add_to_str(&a12, &l, ", "); add_unsigned_long_num_to_str(&a12, &l, cache_info(CI_LOADING));add_to_str(&a12, &l, " ");
 	l = 0, a13 = init_str(); add_to_str(&a13, &l, ".\n");
 
-	l = 0, a14 = init_str(); add_to_str(&a14, &l, ": "); add_num_to_str(&a14, &l, formatted_info(CI_FILES));add_to_str(&a14, &l, " ");
-	l = 0, a15 = init_str(); add_to_str(&a15, &l, ", "); add_num_to_str(&a15, &l, formatted_info(CI_LOCKED));add_to_str(&a15, &l, " ");
+	l = 0, a14 = init_str(); add_to_str(&a14, &l, ": "); add_unsigned_long_num_to_str(&a14, &l, formatted_info(CI_FILES));add_to_str(&a14, &l, " ");
+	l = 0, a15 = init_str(); add_to_str(&a15, &l, ", "); add_unsigned_long_num_to_str(&a15, &l, formatted_info(CI_LOCKED));add_to_str(&a15, &l, " ");
 	l = 0, a16 = init_str(); add_to_str(&a16, &l, ".");
 
 	msg_box(term, getml(a1, a2, a3, a4, a5, a6, a7, a8, a9, a10, a11, a12, a13, a14, a15, a16, NULL), TEXT_(T_RESOURCES), AL_LEFT | AL_EXTD_TEXT, TEXT_(T_RESOURCES), a1, TEXT_(T_HANDLES), a2, TEXT_(T_TIMERS), a3, TEXT_(T_CONNECTIONS), a4, TEXT_(T_cONNECTIONS), a5, TEXT_(T_CONNECTING), a6, TEXT_(T_tRANSFERRING), a7, TEXT_(T_KEEPALIVE), a8, TEXT_(T_MEMORY_CACHE), a9, TEXT_(T_BYTES), a10, TEXT_(T_FILES), a11, TEXT_(T_LOCKED), a12, TEXT_(T_LOADING), a13, TEXT_(T_FORMATTED_DOCUMENT_CACHE), a14, TEXT_(T_DOCUMENTS), a15, TEXT_(T_LOCKED), a16, NULL, r, 1, TEXT_(T_OK), NULL, B_ENTER | B_ESC);
@@ -151,7 +248,7 @@ void cache_inf(struct terminal *term, void *d, struct session *ses)
 	r->timer = install_timer(RESOURCE_INFO_REFRESH, (void (*)(void *))refresh, r);
 }
 
-#ifdef DEBUG
+#if 0
 
 void list_cache(struct terminal *term, void *d, struct session *ses)
 {
@@ -183,13 +280,7 @@ void list_cache(struct terminal *term, void *d, struct session *ses)
 
 #ifdef LEAK_DEBUG
 
-void memory_cld(struct terminal *term, void *d)
-{
-	last_mem_amount = mem_amount;
-}
-
 #define MSG_BUF	2000
-#define MSG_W	100
 
 void memory_info(struct terminal *term, void *d, struct session *ses)
 {
@@ -204,28 +295,11 @@ void memory_info(struct terminal *term, void *d, struct session *ses)
 	r->data = d;
 	r->timer = -1;
 	p = message;
-	sprintf(p, "%ld %s", mem_amount, _(TEXT_(T_MEMORY_ALLOCATED), term)), p += strlen(p);
-	if (last_mem_amount != -1) sprintf(p, ", %s %ld, %s %ld", _(TEXT_(T_LAST), term), last_mem_amount, _(TEXT_(T_DIFFERENCE), term), mem_amount - last_mem_amount), p += strlen(p);
+	snprint(p, MSG_BUF, mem_amount), p += strlen(p);
+	sprintf(p, " %s", _(TEXT_(T_MEMORY_ALLOCATED), term)), p += strlen(p);
 	sprintf(p, "."), p += strlen(p);
-#if 0 && defined(MAX_LIST_SIZE)
-	if (last_mem_amount != -1) {
-		long i, j;
-		int l = 0;
-		for (i = 0; i < MAX_LIST_SIZE; i++) if (memory_list[i].p && memory_list[i].p != last_memory_list[i].p) {
-			for (j = 0; j < MAX_LIST_SIZE; j++) if (last_memory_list[j].p == memory_list[i].p) goto b;
-			if (!l) sprintf(p, "\n%s: ", _(TEXT_(T_NEW_ADDRESSES), term)), p += strlen(p), l = 1;
-			else sprintf(p, ", "), p += strlen(p);
-			sprintf(p, "#%p of %d at %s:%d", memory_list[i].p, (int)memory_list[i].size, memory_list[i].file, memory_list[i].line), p += strlen(p);
-			if (p - message >= MSG_BUF - MSG_W) {
-				sprintf(p, ".."), p += strlen(p);
-				break;
-			}
-			b:;
-		}
-		if (!l) sprintf(p, "\n%s", _(TEXT_(T_NO_NEW_ADDRESSES), term)), p += strlen(p);
-		sprintf(p, "."), p += strlen(p);
-	}
-#endif
+	if (strlen(message) >= MSG_BUF)
+		internal("message buffer overflow");
 	p = stracpy(message);
 	msg_box(term, getml(p, NULL), TEXT_(T_MEMORY_INFO), AL_CENTER, p, r, 1, TEXT_(T_OK), NULL, B_ENTER | B_ESC);
 	r->win = term->windows.next;
@@ -242,7 +316,7 @@ void flush_caches(struct terminal *term, void *d, void *e)
 
 void go_backwards(struct terminal *term, void *psteps, struct session *ses)
 {
-	int steps = (int) psteps;
+	int steps = (int)(my_uintptr_t)psteps;
 
 	/*if (ses->tq_goto_position)
 		--steps;
@@ -280,7 +354,7 @@ void history_menu(struct terminal *term, void *ddd, struct session *ses)
 			if (!mi && !(mi = new_menu(3))) return;
 			url = stracpy(l->vs.url);
 			if (strchr(url, POST_CHAR)) *strchr(url, POST_CHAR) = 0;
-			add_to_menu(&mi, url, "", "", MENU_FUNC go_backwards, (void *) n, 0);
+			add_to_menu(&mi, url, "", "", MENU_FUNC go_backwards, (void *)(my_uintptr_t)n, 0);
 		}
 		n++;
 	}
@@ -327,7 +401,7 @@ void menu_toggle(struct terminal *term, void *ddd, struct session *ses)
 
 void display_codepage(struct terminal *term, void *pcp, struct session *ses)
 {
-	int cp = (int)pcp;
+	int cp = (int)(my_uintptr_t)pcp;
 	struct term_spec *t = new_term_spec(term->term);
 	if (t) t->charset = cp;
 	cls_redraw_all_terminals();
@@ -335,7 +409,7 @@ void display_codepage(struct terminal *term, void *pcp, struct session *ses)
 
 void assumed_codepage(struct terminal *term, void *pcp, struct session *ses)
 {
-	int cp = (int)pcp;
+	int cp = (int)(my_uintptr_t)pcp;
 	ses->ds.assume_cp = cp;
 	redraw_terminal_cls(term);
 }
@@ -348,16 +422,16 @@ void charset_list(struct terminal *term, void *xxx, struct session *ses)
 	if (!(mi = new_menu(1))) return;
 	for (i = 0; (n = get_cp_name(i)); i++) {
 		if (is_cp_special(i)) continue;
-		add_to_menu(&mi, get_cp_name(i), "", "", MENU_FUNC display_codepage, (void *)i, 0);
+		add_to_menu(&mi, get_cp_name(i), "", "", MENU_FUNC display_codepage, (void *)(my_uintptr_t)i, 0);
 	}
 	sel = ses->term->spec->charset;
 	if (sel < 0) sel = 0;
-	do_menu_selected(term, mi, ses, sel);
+	do_menu_selected(term, mi, ses, sel, NULL, NULL);
 }
 
 void set_val(struct terminal *term, void *ip, int *d)
 {
-	*d = (int)ip;
+	*d = (int)(my_uintptr_t)ip;
 }
 
 void charset_sel_list(struct terminal *term, struct session *ses, int *ptr)
@@ -367,11 +441,11 @@ void charset_sel_list(struct terminal *term, struct session *ses, int *ptr)
 	struct menu_item *mi;
 	if (!(mi = new_menu(1))) return;
 	for (i = 0; (n = get_cp_name(i)); i++) {
-		add_to_menu(&mi, get_cp_name(i), "", "", MENU_FUNC set_val, (void *)i, 0);
+		add_to_menu(&mi, get_cp_name(i), "", "", MENU_FUNC set_val, (void *)(my_uintptr_t)i, 0);
 	}
 	sel = *ptr;
 	if (sel < 0) sel = 0;
-	do_menu_selected(term, mi, ptr, sel);
+	do_menu_selected(term, mi, ptr, sel, NULL, NULL);
 }
 
 void terminal_options_ok(void *p)
@@ -579,12 +653,14 @@ unsigned char unrtime_str[5];
 
 void refresh_net(void *xxx)
 {
+	netcfg_stamp++;
 	/*abort_all_connections();*/
 	max_connections = atoi(max_c_str);
 	max_connections_to_host = atoi(max_cth_str);
 	max_tries = atoi(max_t_str);
 	receive_timeout = atoi(time_str);
 	unrestartable_receive_timeout = atoi(unrtime_str);
+	abort_background_connections();
 	register_bottom_half(check_queue, NULL);
 }
 
@@ -967,7 +1043,7 @@ void menu_html_options(struct terminal *term, void *xxx, struct session *ses)
 
 void menu_set_language(struct terminal *term, void *pcp, struct session *ses)
 {
-	set_language((int)pcp);
+	set_language((int)(my_uintptr_t)pcp);
 	cls_redraw_all_terminals();
 }
 
@@ -979,10 +1055,10 @@ void menu_language_list(struct terminal *term, void *xxx, struct session *ses)
 	if (!(mi = new_menu(1))) return;
 	for (i = 0; i < n_languages(); i++) {
 		n = language_name(i);
-		add_to_menu(&mi, n, "", "", MENU_FUNC menu_set_language, (void *)i, 0);
+		add_to_menu(&mi, n, "", "", MENU_FUNC menu_set_language, (void *)(my_uintptr_t)i, 0);
 	}
 	sel = current_language;
-	do_menu_selected(term, mi, ses, sel);
+	do_menu_selected(term, mi, ses, sel, NULL, NULL);
 }
 
 unsigned char *resize_texts[] = { TEXT_(T_COLUMNS), TEXT_(T_ROWS) };
@@ -1002,10 +1078,10 @@ void do_resize_terminal(struct terminal *term)
 void dlg_resize_terminal(struct terminal *term, void *xxx, struct session *ses)
 {
 	struct dialog *d;
-	int x = term->x > 999 ? 999 : term->x;
-	int y = term->y > 999 ? 999 : term->y;
-	sprintf(x_str, "%d", x);
-	sprintf(y_str, "%d", y);
+	unsigned x = (unsigned)term->x > 999 ? 999 : term->x;
+	unsigned y = (unsigned)term->y > 999 ? 999 : term->y;
+	sprintf(x_str, "%u", x);
+	sprintf(y_str, "%u", y);
 	d = mem_alloc(sizeof(struct dialog) + 5 * sizeof(struct dialog_item));
 	memset(d, 0, sizeof(struct dialog) + 5 * sizeof(struct dialog_item));
 	d->title = TEXT_(T_RESIZE_TERMINAL);
@@ -1120,7 +1196,7 @@ void do_file_menu(struct terminal *term, void *xxx, struct session *ses)
 		e++;
 		x = 0;
 	}
-	if (can_resize_window(term->environment)) {
+	if (can_resize_window(term)) {
 		e->text = TEXT_(T_RESIZE_TERMINAL);
 		e->rtext = "";
 		e->hotkey = TEXT_(T_HK_RESIZE_TERMINAL);
@@ -1189,6 +1265,11 @@ struct menu_item ext_menu[] = {
 	{ NULL, NULL, 0, NULL, NULL, 0, 0 },
 };
 
+static void menu_write_config(struct terminal *term, void *xxx, void *yyy)
+{
+	write_config(term);
+}
+
 struct menu_item setup_menu[] = {
 	{ TEXT_(T_LANGUAGE), ">", TEXT_(T_HK_LANGUAGE), MENU_FUNC menu_language_list, NULL, 1, 0 },
 	{ TEXT_(T_CHARACTER_SET), ">", TEXT_(T_HK_CHARACTER_SET), MENU_FUNC charset_list, (void *)1, 1, 0 },
@@ -1199,7 +1280,7 @@ struct menu_item setup_menu[] = {
 	{ TEXT_(T_ASSOCIATIONS), ">", TEXT_(T_HK_ASSOCIATIONS), MENU_FUNC do_menu, assoc_menu, 1, 0 },
 	{ TEXT_(T_FILE_EXTENSIONS), ">", TEXT_(T_HK_FILE_EXTENSIONS), MENU_FUNC do_menu, ext_menu, 1, 0 },
 	{ "", "", M_BAR, NULL, NULL, 0, 0 },
-	{ TEXT_(T_SAVE_OPTIONS), "", TEXT_(T_HK_SAVE_OPTIONS), MENU_FUNC write_config, NULL, 0, 0 },
+	{ TEXT_(T_SAVE_OPTIONS), "", TEXT_(T_HK_SAVE_OPTIONS), MENU_FUNC menu_write_config, NULL, 0, 0 },
 	{ NULL, NULL, 0, NULL, NULL, 0, 0 },
 };
 
@@ -1259,10 +1340,13 @@ void query_file(struct session *ses, unsigned char *url, void (*std)(struct sess
 	int dfl = 0;
 	int l;
 	get_filename_from_url(url, &file, &l);
+	file = memacpy(file, l);
+	check_filename(&file);
 	def = init_str();
 	add_to_str(&def, &dfl, download_dir);
 	if (*def && !dir_sep(def[strlen(def) - 1])) add_chr_to_str(&def, &dfl, '/');
-	add_bytes_to_str(&def, &dfl, file, l);
+	add_to_str(&def, &dfl, file);
+	mem_free(file);
 	input_field(ses->term, NULL, TEXT_(T_DOWNLOAD), TEXT_(T_SAVE_TO_FILE), TEXT_(T_OK), TEXT_(T_CANCEL), ses, &file_history, MAX_INPUT_URL_LEN, def, 0, 0, NULL, (void (*)(void *, unsigned char *))std, (void (*)(void *))cancel);
 	mem_free(def);
 }
