@@ -6,9 +6,14 @@
 #define SYS_ATHEOS	6
 #define SYS_SPAD	7
 #define SYS_INTERIX	8
+#define SYS_OPENVMS	9
 
 /* hardcoded limit of 10 OSes in default.c */
 
+#ifdef UNIX
+#undef UNIX
+#endif
+
 #if defined(__EMX__)
 #define OS2
 #elif defined(_WIN32) || defined(__CYGWIN__)
@@ -23,16 +28,12 @@
 #define ATHEOS
 #elif defined(__SPAD__)
 #define SPAD
+#elif defined(__VMS)
+#define OPENVMS
 #else
 #define UNIX
 #endif
 
-#if defined(OS2) || defined(WIN32) || defined(INTERIX) || defined(BEOS) || defined(RISCOS) || defined(SPAD)
-#ifdef UNIX
-#undef UNIX
-#endif
-#endif
-
 #if defined(UNIX)
 
 static inline int dir_sep(char x) { return x == '/'; }
@@ -66,6 +67,7 @@ static inline int dir_sep(char x) { return x == '/' || x == '\\'; }
 #define NO_FG_EXEC
 #define NO_CTRL_Z
 #define DOS_FS
+#define DOS_FS_8_3
 #define NO_FILE_SECURITY
 #define NO_FORK_ON_EXIT
 #define ASSOC_CONS_XWIN
@@ -85,9 +87,19 @@ static inline int dir_sep(char x) { return x == '/' || x == '\\'; }
 #define DOS_FS
 #define SET_WINDOW_TITLE_UTF_8
 #define ASSOC_CONS_XWIN
+#ifdef _UWIN
+#define DISABLE_SMB
+#endif
+#ifdef __CYGWIN__
 #define OS_BAD_SIGNALS
+#endif
 #ifndef HAVE_PTHREADS
-  #define HAVE_PTHREADS
+#define HAVE_PTHREADS
+#endif
+#if defined(HAVE_SYS_UN_H) && !defined(_UWIN)
+#define USE_AF_UNIX
+#else
+#define DONT_USE_AF_UNIX
 #endif
 
 #elif defined(INTERIX)
@@ -195,6 +207,26 @@ static inline int dir_sep(char x) { return x == '/'; }
 #define ASSOC_CONS_XWIN
 #define NO_FORK_ON_EXIT
 
+#elif defined(OPENVMS)
+
+static inline int dir_sep(unsigned char x) { return x == '/'; }
+#define NEWLINE "\n"
+#define FS_UNIX_RIGHTS
+#define SYSTEM_ID SYS_OPENVMS
+#define SYSTEM_NAME "VMS"
+#define DEFAULT_SHELL "/DCL"
+#define GETSHELL NULL
+#define SHARED_CONFIG_DIR "/etc/"
+#ifndef HAVE_PTHREADS
+#define HAVE_PTHREADS
+#endif
+#define DONT_USE_AF_UNIX
+#define ASSOC_BLOCK
+#define ASSOC_CONS_XWIN
+#define NO_CTRL_Z
+#define NO_FORK_ON_EXIT
+#define DISABLE_SMB
+
 #endif
 
 #ifdef FS_UNIX_USERS
@@ -203,7 +235,3 @@ static inline int dir_sep(char x) { return x == '/'; }
 #endif
 #endif
 
-#if !defined(HAVE_BEGINTHREAD) && !defined(BEOS) && !defined(HAVE_PTHREADS)
-#define THREAD_SAFE_LOOKUP
-#endif
-
