@@ -4,6 +4,8 @@ tcount connection_count = 0;
 
 int active_connections = 0;
 
+tcount netcfg_stamp = 0;
+
 struct list_head queue = {&queue, &queue};
 
 struct h_conn {
@@ -17,7 +19,9 @@ struct list_head h_conns = {&h_conns, &h_conns};
 
 struct list_head keepalive_connections = {&keepalive_connections, &keepalive_connections};
 
-long connect_info(int type)
+void check_keepalive_connections();
+
+unsigned long connect_info(int type)
 {
 	int i = 0;
 	struct connection *ce;
@@ -33,12 +37,13 @@ long connect_info(int type)
 			foreach(ce, queue) i += ce->state == S_TRANS;
 			return i;
 		case CI_KEEP:
+			check_keepalive_connections();
 			foreach(cee, keepalive_connections) i++;
 			return i;
 		case CI_LIST:
 			return (long) &queue;
 		default:
-			internal("cache_info: bad request");
+			internal("connect_info: bad request");
 	}
 	return 0;
 }
@@ -73,18 +78,19 @@ void stat_timer(struct connection *c)
 	r->loaded = c->received;
 	if ((r->size = c->est_length) < (r->pos = c->from) && r->size != -1)
 		r->size = c->from;
-	r->dis_b += a = get_time() - r->last_time;
+	a = (uttime)get_time() - (uttime)r->last_time;
+	r->dis_b += (uttime)a;
 	while (r->dis_b >= SPD_DISP_TIME * CURRENT_SPD_SEC) {
 		r->cur_loaded -= r->data_in_secs[0];
-		memmove(r->data_in_secs, r->data_in_secs + 1, sizeof(int) * (CURRENT_SPD_SEC - 1));
+		memmove(r->data_in_secs, r->data_in_secs + 1, sizeof(off_t) * (CURRENT_SPD_SEC - 1));
 		r->data_in_secs[CURRENT_SPD_SEC - 1] = 0;
-		r->dis_b -= SPD_DISP_TIME;
+		r->dis_b -= (uttime)SPD_DISP_TIME;
 	}
 	r->data_in_secs[CURRENT_SPD_SEC - 1] += r->loaded - r->last_loaded;
-	r->cur_loaded += r->loaded - r->last_loaded;
-	r->last_loaded = r->loaded;
-	r->last_time += a;
-	r->elapsed += a;
+	r->cur_loaded += (uttime)r->loaded - (uttime)r->last_loaded;
+	r->last_loaded = (uttime)r->loaded;
+	r->last_time += (uttime)a;
+	r->elapsed += (uttime)a;
 	r->timer = install_timer(SPD_DISP_TIME, (void (*)(void *))stat_timer, c);
 	if (!st_r) send_connection_info(c);
 }
@@ -137,12 +143,13 @@ struct k_conn *is_host_on_keepalive_list(struct connection *c)
 	return NULL;
 }
 
-int get_keepalive_socket(struct connection *c)
+int get_keepalive_socket(struct connection *c, int *protocol_data)
 {
 	struct k_conn *k;
 	int cc;
 	if (!(k = is_host_on_keepalive_list(c))) return -1;
 	cc = k->conn;
+	if (protocol_data) *protocol_data = k->protocol_data;
 	del_from_list(k);
 	mem_free(k->host);
 	mem_free(k);
@@ -150,12 +157,14 @@ int get_keepalive_socket(struct connection *c)
 	return 0;
 }
 
-void check_keepalive_connections();
-
 void abort_all_keepalive_connections()
 {
 	struct k_conn *k;
-	foreach(k, keepalive_connections) mem_free(k->host), close(k->conn);
+	int rs;
+	foreach(k, keepalive_connections) {
+		mem_free(k->host);
+		EINTRLOOP(rs, close(k->conn));
+	}
 	free_list(keepalive_connections);
 	check_keepalive_connections();
 }
@@ -163,13 +172,14 @@ void abort_all_keepalive_connections()
 void free_connection_data(struct connection *c)
 {
 	struct h_conn *h;
+	int rs;
 	if (c->sock1 != -1) set_handlers(c->sock1, NULL, NULL, NULL, NULL);
 	if (c->sock2 != -1) set_handlers(c->sock2, NULL, NULL, NULL, NULL);
 	close_socket(&c->sock2);
 	if (c->pid) {
-		kill(c->pid, SIGINT);
-		kill(c->pid, SIGTERM);
-		kill(c->pid, SIGKILL);
+		EINTRLOOP(rs, kill(c->pid, SIGINT));
+		EINTRLOOP(rs, kill(c->pid, SIGTERM));
+		EINTRLOOP(rs, kill(c->pid, SIGKILL));
 		c->pid = 0;
 	}
 	if (!c->running) {
@@ -222,6 +232,11 @@ void del_connection(struct connection *c)
 {
 	del_from_list(c);
 	send_connection_info(c);
+	if (c->detached) {
+		struct cache_entry *ce = c->cache;
+		if (ce && !ce->url[0] && !is_entry_used(ce) && !ce->refcount)
+			delete_cache_entry(ce);
+	}
 	mem_free(c->url);
 	mem_free(c);
 }
@@ -230,16 +245,20 @@ void del_connection(struct connection *c)
 void check_queue_bugs();
 #endif
 
-void add_keepalive_socket(struct connection *c, ttime timeout)
+void add_keepalive_socket(struct connection *c, ttime timeout, int protocol_data)
 {
 	struct k_conn *k;
+	int rs;
 	free_connection_data(c);
 	if (c->sock1 == -1) {
 		internal("keepalive connection not connected");
 		goto del;
 	}
 	k = mem_alloc(sizeof(struct k_conn));
-	if ((k->port = get_port(c->url)) == -1 || !(k->protocol = get_protocol_handle(c->url)) || !(k->host = get_host_and_pass(c->url))) {
+	if (c->netcfg_stamp != netcfg_stamp ||
+	    (k->port = get_port(c->url)) == -1 ||
+	    !(k->protocol = get_protocol_handle(c->url)) ||
+	    !(k->host = get_host_and_pass(c->url))) {
 		mem_free(k);
 		del_connection(c);
 		goto close;
@@ -247,6 +266,7 @@ void add_keepalive_socket(struct connection *c, ttime timeout)
 	k->conn = c->sock1;
 	k->timeout = timeout;
 	k->add_time = get_time();
+	k->protocol_data = protocol_data;
 	add_to_list(keepalive_connections, k);
 	del:
 	del_connection(c);
@@ -256,7 +276,7 @@ void add_keepalive_socket(struct connection *c, ttime timeout)
 	register_bottom_half(check_queue, NULL);
 	return;
 	close:
-	close(c->sock1);
+	EINTRLOOP(rs, close(c->sock1));
 #ifdef DEBUG
 	check_queue_bugs();
 #endif
@@ -265,16 +285,15 @@ void add_keepalive_socket(struct connection *c, ttime timeout)
 
 void del_keepalive_socket(struct k_conn *kc)
 {
+	int rs;
 	del_from_list(kc);
-	close(kc->conn);
+	EINTRLOOP(rs, close(kc->conn));
 	mem_free(kc->host);
 	mem_free(kc);
 }
 
 int keepalive_timeout = -1;
 
-void check_keepalive_connections();
-
 void keepalive_timer(void *x)
 {
 	keepalive_timeout = -1;
@@ -287,7 +306,7 @@ void check_keepalive_connections()
 	ttime ct = get_time();
 	int p = 0;
 	if (keepalive_timeout != -1) kill_timer(keepalive_timeout), keepalive_timeout = -1;
-	foreach(kc, keepalive_connections) if (can_read(kc->conn) || ct - kc->add_time > kc->timeout) {
+	foreach(kc, keepalive_connections) if (can_read(kc->conn) || (uttime)ct - (uttime)kc->add_time > (uttime)kc->timeout) {
 		kc = kc->prev;
 		del_keepalive_socket(kc->next);
 	} else p++;
@@ -326,12 +345,13 @@ void sort_queue()
 void interrupt_connection(struct connection *c)
 {
 #ifdef HAVE_SSL
-	if (c->ssl == (void *)-1) c->ssl = 0;
-	if(c->ssl) {
+	if (c->ssl == (void *)-1) c->ssl = NULL;
+	if (c->ssl) {
 		SSL_free(c->ssl);
-		c->ssl=NULL;
+		c->ssl = NULL;
 	}
 #endif
+	if (c->sock1 != -1) set_handlers(c->sock1, NULL, NULL, NULL, NULL);
 	close_socket(&c->sock1);
 	free_connection_data(c);
 }
@@ -460,14 +480,16 @@ void check_queue_bugs()
 	foreach(d, queue) {
 		int q = getpri(d);
 		cc += d->running;
-		if (q < p) if (!ps) {
-			internal("queue is not sorted");
-			sort_queue();
-			ps = 1;
-			goto again;
-		} else {
-			internal("queue is not sorted even after sort_queue!");
-			break;
+		if (q < p) {
+			if (!ps) {
+				internal("queue is not sorted");
+				sort_queue();
+				ps = 1;
+				goto again;
+			} else {
+				internal("queue is not sorted even after sort_queue!");
+				break;
+			}
 		} else p = q;
 		if (d->state < 0) {
 			internal("interrupted connection on queue (conn %s, state %d)", d->url, d->state);
@@ -567,7 +589,7 @@ int load_url(unsigned char *url, struct status *stat, int pri, int no_cache)
 		}
 		if (stat) {
 			stat->ce = e;
-			stat->state = S_OK;
+			stat->state = S__OK;
 			if (stat->end) stat->end(stat, stat->data);
 		}
 		e->refcount--;
@@ -611,7 +633,18 @@ int load_url(unsigned char *url, struct status *stat, int pri, int no_cache)
 	c->url = u;
 	c->running = 0;
 	c->prev_error = 0;
-	c->from = no_cache >= NC_IF_MOD || !e || e->frag.next == &e->frag || ((struct fragment *)e->frag.next)->offset ? 0 : ((struct fragment *)e->frag.next)->length;
+	if (no_cache >= NC_IF_MOD || !e) {
+		c->from = 0;
+	} else {
+		struct fragment *frag;
+		c->from = 0;
+		foreach(frag, e->frag) {
+			if (frag->offset != c->from)
+				break;
+			c->from += frag->length;
+		}
+		
+	}
 	memset(c->pri, 0, sizeof c->pri);
 	c->pri[pri] = 1;
 	c->no_cache = no_cache;
@@ -622,6 +655,7 @@ int load_url(unsigned char *url, struct status *stat, int pri, int no_cache)
 	c->newconn = NULL;
 	c->cache = NULL;
 	c->tries = 0;
+	c->netcfg_stamp = netcfg_stamp;
 	init_list(c->statuss);
 	c->est_length = -1;
 	c->unrestartable = 0;
@@ -698,8 +732,8 @@ void detach_connection(struct status *stat, off_t pos)
 	off_t l;
 	if (stat->state < 0) return;
 	c = stat->c;
-	if (c->detached) goto detach_done;
 	if (!c->cache) return;
+	if (c->detached) goto detach_done;
 	if (c->est_length == -1) l = c->from;
 	else l = c->est_length;
 	if (l < (longlong)memory_cache_size * MAX_CACHED_OBJECT) return;
@@ -718,8 +752,7 @@ void connection_timeout(struct connection *c)
 {
 	c->timer = -1;
 	setcstate(c, S_TIMEOUT);
-	if (c->dnsquery) abort_connection(c);
-	else retry_connection(c);
+	retry_connection(c);
 }
 
 void connection_timeout_1(struct connection *c)
@@ -828,7 +861,7 @@ struct s_msg_dsc msg_dsc[] = {
 	{ S_PROC,		TEXT_(T_SERVER_IS_PROCESSING_REQUEST) },
 	{ S_TRANS,		TEXT_(T_TRANSFERRING) },
 
-	{ S_OK,			TEXT_(T_OK) },
+	{ S__OK,		TEXT_(T_OK) },
 	{ S_INTERRUPTED,	TEXT_(T_INTERRUPTED) },
 	{ S_EXCEPT,		TEXT_(T_SOCKET_EXCEPTION) },
 	{ S_INTERNAL,		TEXT_(T_INTERNAL_ERROR) },
