@@ -35,14 +35,31 @@ struct list_head timers = {&timers, &timers};
 ttime get_time()
 {
 	struct timeval tv;
-	gettimeofday(&tv, NULL);
-	return tv.tv_sec * 1000 + tv.tv_usec / 1000;
+	int rs;
+	EINTRLOOP(rs, gettimeofday(&tv, NULL));
+	if (rs) {
+		error("gettimeofday failed: %d", errno);
+		fatal_tty_exit();
+		exit(RET_FATAL);
+	}
+	return (uttime)tv.tv_sec * 1000 + tv.tv_usec / 1000;
 }
 
-long select_info(int type)
+#ifndef OPENVMS
+void portable_sleep(unsigned msec)
+{
+	struct timeval tv;
+	int rs;
+	tv.tv_sec = msec / 1000;
+	tv.tv_usec = msec % 1000 * 1000;
+	EINTRLOOP(rs, select(0, NULL, NULL, NULL, &tv));
+}
+#endif
+
+unsigned long select_info(int type)
 {
 	int i = 0, j;
-	struct cache_entry *ce;
+	struct timer *ce;
 	switch (type) {
 		case CI_FILES:
 			for (j = 0; j < FD_SETSIZE; j++)
@@ -52,7 +69,7 @@ long select_info(int type)
 			foreach(ce, timers) i++;
 			return i;
 		default:
-			internal("cache_info: bad request");
+			internal("select_info: bad request");
 	}
 	return 0;
 }
@@ -66,15 +83,25 @@ struct bottom_half {
 
 struct list_head bottom_halves = { &bottom_halves, &bottom_halves };
 
-int register_bottom_half(void (*fn)(void *), void *data)
+void register_bottom_half(void (*fn)(void *), void *data)
 {
 	struct bottom_half *bh;
-	foreach(bh, bottom_halves) if (bh->fn == fn && bh->data == data) return 0;
-	if (!(bh = mem_alloc(sizeof(struct bottom_half)))) return -1;
+	foreach(bh, bottom_halves) if (bh->fn == fn && bh->data == data) return;
+	bh = mem_alloc(sizeof(struct bottom_half));
 	bh->fn = fn;
 	bh->data = data;
 	add_to_list(bottom_halves, bh);
-	return 0;
+}
+
+void unregister_bottom_half(void (*fn)(void *), void *data)
+{
+	struct bottom_half *bh;
+	retry:
+	foreach(bh, bottom_halves) if (bh->fn == fn && bh->data == data) {
+		del_from_list(bh);
+		mem_free(bh);
+		goto retry;
+	}
 }
 
 void check_bottom_halves()
@@ -101,7 +128,7 @@ void check_timers()
 {
 	ttime interval = get_time() - last_time;
 	struct timer *t;
-	foreach(t, timers) t->interval -= interval;
+	foreach(t, timers) t->interval -= (uttime)interval;
 	ch:
 	foreach(t, timers) if (t->interval <= 0) {
 		struct timer *tt = t;
@@ -111,13 +138,13 @@ void check_timers()
 		CHK_BH;
 		goto ch;
 	} else break;
-	last_time += interval;
+	last_time += (uttime)interval;
 }
 
 int install_timer(ttime t, void (*func)(void *), void *data)
 {
 	struct timer *tm, *tt;
-	if (!(tm = mem_alloc(sizeof(struct timer)))) return -1;
+	tm = mem_alloc(sizeof(struct timer));
 	tm->interval = t;
 	tm->func = func;
 	tm->data = data;
@@ -151,7 +178,7 @@ void *get_handler(int fd, int tp)
 	if (fd < 0)
 		internal("get_handler: handle %d", fd);
 	if (fd >= (int)FD_SETSIZE) {
-		error("too big handle %d");
+		error("too big handle %d", fd);
 		fatal_tty_exit();
 		exit(RET_FATAL);
 		return NULL;
@@ -168,10 +195,14 @@ void *get_handler(int fd, int tp)
 
 void set_handlers(int fd, void (*read_func)(void *), void (*write_func)(void *), void (*error_func)(void *), void *data)
 {
+#ifdef __GNU__
+	/* GNU Hurd has bugs w.r.t. exceptions */
+	error_func = NULL;
+#endif
 	if (fd < 0)
 		internal("set_handlers: handle %d", fd);
 	if (fd >= (int)FD_SETSIZE) {
-		error("too big handle %d");
+		error("too big handle %d", fd);
 		fatal_tty_exit();
 		exit(RET_FATAL);
 		return;
@@ -206,7 +237,29 @@ void set_handlers(int fd, void (*read_func)(void *), void (*write_func)(void *),
 	}
 }
 
+void clear_events(int h, int blocking)
+{
+#if !defined(O_NONBLOCK) && !defined(FIONBIO)
+	blocking = 1;
+#endif
+	while (blocking ? can_read(h) : 1) {
+		unsigned char c[64];
+		int rd;
+		EINTRLOOP(rd, (int)read(h, c, sizeof c));
+		if (rd != sizeof c) break;
+	}
+}
+
+static void clear_events_ptr(void *handle)
+{
+	clear_events((int)(my_intptr_t)handle, 0);
+}
+
+#if defined(NSIG) && NSIG > 32
+#define NUM_SIGNALS	NSIG
+#else
 #define NUM_SIGNALS	32
+#endif
 
 struct signal_handler {
 	void (*fn)(void *);
@@ -214,20 +267,19 @@ struct signal_handler {
 	int critical;
 };
 
-int signal_mask[NUM_SIGNALS];
-struct signal_handler signal_handlers[NUM_SIGNALS];
+volatile int signal_mask[NUM_SIGNALS];
+volatile struct signal_handler signal_handlers[NUM_SIGNALS];
 
 int signal_pipe[2];
 
-void signal_break(void *data)
-{
-	char c;
-	while (can_read(signal_pipe[0])) read(signal_pipe[0], &c, 1);
-}
-
 SIGNAL_HANDLER void got_signal(int sig)
 {
 	int sv_errno = errno;
+
+#if !defined(HAVE_SIGACTION)
+	do_signal(sig, got_signal);
+#endif
+
 	if (sig >= NUM_SIGNALS || sig < 0) {
 		/*error("ERROR: bad signal number: %d", sig);*/
 		goto ret;
@@ -239,14 +291,22 @@ SIGNAL_HANDLER void got_signal(int sig)
 	}
 	signal_mask[sig] = 1;
 	ret:
-	if (can_write(signal_pipe[1])) write(signal_pipe[1], "x", 1);
+	if (can_write(signal_pipe[1])) {
+		int wr;
+		EINTRLOOP(wr, write(signal_pipe[1], "", 1));
+	}
 	errno = sv_errno;
 }
 
+#ifdef HAVE_SIGACTION
 struct sigaction sa_zero;
+#endif
 
 void install_signal_handler(int sig, void (*fn)(void *), void *data, int critical)
 {
+#if defined(OPENVMS)
+#elif defined(HAVE_SIGACTION)
+	int rs;
 	struct sigaction sa = sa_zero;
 	if (sig >= NUM_SIGNALS || sig < 0) {
 		internal("bad signal number: %d", sig);
@@ -254,35 +314,81 @@ void install_signal_handler(int sig, void (*fn)(void *), void *data, int critica
 	}
 	if (!fn) sa.sa_handler = SIG_IGN;
 	else sa.sa_handler = (void *)got_signal;
-#ifdef HAVE_SIGFILLSET
 	sigfillset(&sa.sa_mask);
-#else
-	memset(&sa.sa_mask, -1, sizeof sa.sa_mask);
-#endif
 	sa.sa_flags = SA_RESTART;
-	if (!fn) sigaction(sig, &sa, NULL);
+	if (!fn)
+		EINTRLOOP(rs, sigaction(sig, &sa, NULL));
 	signal_handlers[sig].fn = fn;
 	signal_handlers[sig].data = data;
 	signal_handlers[sig].critical = critical;
-	if (fn) sigaction(sig, &sa, NULL);
+	if (fn)
+		EINTRLOOP(rs, sigaction(sig, &sa, NULL));
+#else
+	if (!fn) do_signal(sig, SIG_IGN);
+	signal_handlers[sig].fn = fn;
+	signal_handlers[sig].data = data;
+	signal_handlers[sig].critical = critical;
+	if (fn) do_signal(sig, got_signal);
+#endif
 }
 
 void interruptible_signal(int sig, int in)
 {
+#if defined(HAVE_SIGACTION) && !defined(OPENVMS)
 	struct sigaction sa = sa_zero;
+	int rs;
 	if (sig >= NUM_SIGNALS || sig < 0) {
 		internal("bad signal number: %d", sig);
 		return;
 	}
 	if (!signal_handlers[sig].fn) return;
 	sa.sa_handler = (void *)got_signal;
-#ifdef HAVE_SIGFILLSET
 	sigfillset(&sa.sa_mask);
-#else
-	memset(&sa.sa_mask, -1, sizeof sa.sa_mask);
-#endif
 	if (!in) sa.sa_flags = SA_RESTART;
-	sigaction(sig, &sa, NULL);
+	EINTRLOOP(rs, sigaction(sig, &sa, NULL));
+#endif
+}
+
+static sigset_t sig_old_mask;
+static int sig_unblock = 0;
+
+void block_signals(int except1, int except2)
+{
+	int rs;
+	sigset_t mask;
+	sigfillset(&mask);
+#ifdef HAVE_SIGDELSET
+	if (except1) sigdelset(&mask, except1);
+	if (except2) sigdelset(&mask, except2);
+#ifdef SIGILL
+	sigdelset(&mask, SIGILL);
+#endif
+#ifdef SIGABRT
+	sigdelset(&mask, SIGABRT);
+#endif
+#ifdef SIGFPE
+	sigdelset(&mask, SIGFPE);
+#endif
+#ifdef SIGSEGV
+	sigdelset(&mask, SIGSEGV);
+#endif
+#ifdef SIGBUS
+	sigdelset(&mask, SIGBUS);
+#endif
+#else
+	if (except1 || except2) return;
+#endif
+	EINTRLOOP(rs, do_sigprocmask(SIG_BLOCK, &mask, &sig_old_mask));
+	if (!rs) sig_unblock = 1;
+}
+
+void unblock_signals(void)
+{
+	int rs;
+	if (sig_unblock) {
+		EINTRLOOP(rs, do_sigprocmask(SIG_SETMASK, &sig_old_mask, NULL));
+		sig_unblock = 0;
+	}
 }
 
 int check_signals()
@@ -298,12 +404,16 @@ int check_signals()
 	return r;
 }
 
+#ifdef SIGCHLD
 void sigchld(void *p)
 {
+	pid_t pid;
 #ifndef WNOHANG
-	wait(NULL);
+	EINTRLOOP(pid, wait(NULL));
 #else
-	while (waitpid(-1, NULL, WNOHANG) > 0) ;
+	do {
+		EINTRLOOP(pid, waitpid(-1, NULL, WNOHANG));
+	} while (pid > 0);
 #endif
 }
 
@@ -311,33 +421,43 @@ void set_sigcld()
 {
 	install_signal_handler(SIGCHLD, sigchld, NULL, 1);
 }
+#else
+void set_sigcld()
+{
+}
+#endif
 
-int terminate = 0;
+int terminate_loop = 0;
 
 void select_loop(void (*init)())
 {
 	struct stat st;
-	if (stat(".", &st) && getenv("HOME")) chdir(getenv("HOME"));
+	int rs;
+	EINTRLOOP(rs, stat(".", &st));
+	if (rs && getenv("HOME"))
+		EINTRLOOP(rs, chdir(getenv("HOME")));
+#ifdef HAVE_SIGACTION
 	memset(&sa_zero, 0, sizeof sa_zero);
-	memset(signal_mask, 0, sizeof signal_mask);
-	memset(signal_handlers, 0, sizeof signal_handlers);
+#endif
+	memset((void *)signal_mask, 0, sizeof signal_mask);
+	memset((void *)signal_handlers, 0, sizeof signal_handlers);
 	FD_ZERO(&w_read);
 	FD_ZERO(&w_write);
 	FD_ZERO(&w_error);
 	w_max = 0;
 	last_time = get_time();
-	signal(SIGPIPE, SIG_IGN);
+	ignore_signals();
 	if (c_pipe(signal_pipe)) {
 		error("ERROR: can't create pipe for signal handling");
 		fatal_tty_exit();
 		exit(RET_FATAL);
 	}
-	fcntl(signal_pipe[0], F_SETFL, O_NONBLOCK);
-	fcntl(signal_pipe[1], F_SETFL, O_NONBLOCK);
-	set_handlers(signal_pipe[0], signal_break, NULL, NULL, NULL);
+	set_nonblock(signal_pipe[0]);
+	set_nonblock(signal_pipe[1]);
+	set_handlers(signal_pipe[0], clear_events_ptr, NULL, NULL, (void *)(my_intptr_t)signal_pipe[0]);
 	init();
 	CHK_BH;
-	while (!terminate) {
+	while (!terminate_loop) {
 		int n, i;
 		struct timeval tv;
 		struct timeval *tm = NULL;
@@ -352,13 +472,13 @@ void select_loop(void (*init)())
 		tm = &tv;
 #endif
 		if (!list_empty(timers)) {
-			ttime tt = ((struct timer *)timers.next)->interval + 1;
+			ttime tt = (uttime)((struct timer *)timers.next)->interval + 1;
 			if (tt < 0) tt = 0;
 #ifdef OS_BAD_SIGNALS
 			if (tt < 1000)
 #endif
 			{
-				tv.tv_sec = tt / 1000;
+				tv.tv_sec = tt / 1000 < MAXINT ? (int)(tt / 1000) : MAXINT;
 				tv.tv_usec = (tt % 1000) * 1000;
 				tm = &tv;
 			}
@@ -367,7 +487,7 @@ void select_loop(void (*init)())
 		memcpy(&x_write, &w_write, sizeof(fd_set));
 		memcpy(&x_error, &w_error, sizeof(fd_set));
 		/*rep_sel:*/
-		if (terminate) break;
+		if (terminate_loop) break;
 		if (!w_max && list_empty(timers)) break;
 		if (check_signals()) {
 			continue;
