@@ -1,5 +1,7 @@
 #include "links.h"
 
+void undisplay_download(struct download *down);
+
 struct list_head downloads = {&downloads, &downloads};
 
 int are_there_downloads()
@@ -29,10 +31,10 @@ void free_strerror_buf()
 
 int get_error_from_errno(int errn)
 {
-	if (errn > 0 && (errn < -S_OK || errn > -S_MAX))
+	if (errn > 0 && (errn < -S__OK || errn > -S_MAX))
 		return -errn;
 #ifdef BEOS
-	if (-errn > 0 && (-errn < -S_OK || -errn > -S_MAX))
+	if (-errn > 0 && (-errn < -S__OK || -errn > -S_MAX))
 		return errn;
 #endif
 	return S_UNKNOWN_ERROR;
@@ -42,7 +44,7 @@ unsigned char *get_err_msg(int state)
 {
 	unsigned char *e;
 	struct strerror_val *s;
-	if ((state >= S_MAX && state <= S_OK) || state >= S_WAIT) {
+	if ((state >= S_MAX && state <= S__OK) || state >= S_WAIT) {
 		int i;
 		for (i = 0; msg_dsc[i].msg; i++)
 			if (msg_dsc[i].n == state) return msg_dsc[i].msg;
@@ -80,8 +82,7 @@ void add_time_to_str(unsigned char **s, int *l, ttime t)
 {
 	unsigned char q[64];
 	if (t < 0) t = 0;
-	t &= 0xffffffff;
-	if (t >= 86400) sprintf(q, "%dd ", (int)(t / 86400)), add_to_str(s, l, q);
+	if (t >= 86400) sprintf(q, "%ud ", (unsigned)(t / 86400)), add_to_str(s, l, q);
 	if (t >= 3600) t %= 86400, sprintf(q, "%d:%02d", (int)(t / 3600), (int)(t / 60 % 60)), add_to_str(s, l, q);
 	else sprintf(q, "%d", (int)(t / 60)), add_to_str(s, l, q);
 	sprintf(q, ":%02d", (int)(t % 60)), add_to_str(s, l, q);
@@ -116,11 +117,11 @@ void print_screen_status(struct session *ses)
 	struct terminal *term = ses->term;
 	struct status *stat = NULL;
 	unsigned char *m;
-	fill_area(term, 0, 0, term->x, 1, COLOR_TITLE_BG);
-	fill_area(term, 0, term->y - 1, term->x, 1, COLOR_STATUS_BG);
+	fill_area(term, 0, 0, term->x, 1, COLOR_TITLE_BG | ' ');
+	fill_area(term, 0, term->y - 1, term->x, 1, COLOR_STATUS_BG | ' ');
 	if (ses->wtd) stat = &ses->loading;
 	else if (!list_empty(ses->history)) stat = &cur_loc(ses)->stat;
-	if (stat && stat->state == S_OK) {
+	if (stat && stat->state == S__OK) {
 		struct file_to_load *ftl;
 		foreach(ftl, ses->more_files) {
 			if (ftl->req_sent && ftl->stat.state >= 0) {
@@ -130,7 +131,7 @@ void print_screen_status(struct session *ses)
 		}
 	}
 	if (stat) {
-		if (stat->state == S_OK) if ((m = print_current_link(ses))) goto p;
+		if (stat->state == S__OK) if ((m = print_current_link(ses))) goto p;
 		if ((m = get_stat_msg(stat, term))) {
 			p:
 			print_text(term, 0, term->y - 1, strlen(m), m, COLOR_STATUS);
@@ -155,7 +156,6 @@ void print_screen_status(struct session *ses)
 void print_error_dialog(struct session *ses, struct status *stat, unsigned char *title)
 {
 	unsigned char *t = get_err_msg(stat->state);
-	if (!t) return;
 	msg_box(ses->term, NULL, title, AL_CENTER, t, ses, 1, TEXT_(T_CANCEL), NULL, B_ENTER | B_ESC/*, _("Retry"), NULL, 0 !!! FIXME: retry */);
 }
 
@@ -278,8 +278,8 @@ void ses_imgmap(struct session *ses)
 	d_opt = &fd->f_data->opt;
 	if (get_image_map(ce->head, fr->data, fr->data + fr->length, ses->goto_position, &menu, &ml, ses->imgmap_href_base, ses->imgmap_target_base, ses->term->spec->charset, ses->ds.assume_cp, ses->ds.hard_assume))
 		return;
-	add_empty_window(ses->term, (void (*)(void *))freeml, ml);
-	do_menu(ses->term, menu, ses);
+	/*add_empty_window(ses->term, (void (*)(void *))freeml, ml);*/
+	do_menu_selected(ses->term, menu, ses, 0, (void (*)(void *))freeml, ml);
 }
 
 void map_selected(struct terminal *term, struct link_def *ld, struct session *ses)
@@ -318,13 +318,19 @@ struct session *get_download_ses(struct download *down)
 
 void abort_download(struct download *down)
 {
+	int rs;
+	unregister_bottom_half((void (*)(void *))abort_download, down);
+	unregister_bottom_half((void (*)(void *))undisplay_download, down);
 	if (down->win) delete_window(down->win);
 	if (down->ask) delete_window(down->ask);
 	if (down->stat.state >= 0) change_connection(&down->stat, NULL, PRI_CANCEL);
 	mem_free(down->url);
-	if (down->handle != -1) prealloc_truncate(down->handle, down->last_pos), close(down->handle);
+	if (down->handle != -1) {
+		prealloc_truncate(down->handle, down->last_pos);
+		EINTRLOOP(rs, close(down->handle));
+	}
 	if (down->prog) {
-		unlink(down->file);
+		EINTRLOOP(rs, unlink(down->file));
 		mem_free(down->prog);
 	}
 	mem_free(down->file);
@@ -361,6 +367,40 @@ void download_abort_function(struct dialog_data *dlg)
 	down->win = NULL;
 }
 
+int test_percentage(struct status *stat)
+{
+	return stat->prg->size > 0;
+}
+
+static int download_meter(int size, struct status *stat)
+{
+	int m;
+	if (!stat->prg->size) return 0;
+	m = (int)((double)size * (double)stat->prg->pos / (double)stat->prg->size);
+	if (m < 0) m = 0;
+	if (m > size) m = size;
+	return m;
+}
+
+unsigned char *download_percentage(struct download *down, int pad)
+{
+	unsigned char *s;
+	int l;
+	int perc;
+	struct status *stat = &down->stat;
+	if (stat->state != S_TRANS || !test_percentage(stat)) return stracpy("");
+	s = init_str();
+	l = 0;
+	perc = download_meter(100, stat);
+	if (pad) {
+		if (perc < 10) add_chr_to_str(&s, &l, ' ');
+		if (perc < 100) add_chr_to_str(&s, &l, ' ');
+	}
+	add_num_to_str(&s, &l, perc);
+	add_chr_to_str(&s, &l, '%');
+	return s;
+}
+
 void download_window_function(struct dialog_data *dlg)
 {
 	struct download *down = dlg->dlg->udata;
@@ -368,6 +408,7 @@ void download_window_function(struct dialog_data *dlg)
 	int max = 0, min = 0;
 	int w, x, y;
 	int t = 0;
+	int show_percentage = 0;
 	unsigned char *m, *u;
 	struct status *stat = &down->stat;
 	redraw_below_window(dlg->win);
@@ -395,15 +436,17 @@ void download_window_function(struct dialog_data *dlg)
 		add_to_str(&m, &l, "\n");
 		add_to_str(&m, &l, _(TEXT_(T_ELAPSED_TIME), term));
 		add_to_str(&m, &l, " ");
-		add_time_to_str(&m, &l, stat->prg->elapsed / 1000);
+		add_time_to_str(&m, &l, (uttime)stat->prg->elapsed / 1000);
 		if (stat->prg->size >= 0 && stat->prg->loaded > 0) {
 			add_to_str(&m, &l, ", ");
 			add_to_str(&m, &l, _(TEXT_(T_ESTIMATED_TIME), term));
 			add_to_str(&m, &l, " ");
 			/*add_time_to_str(&m, &l, stat->prg->elapsed / 1000 * stat->prg->size / stat->prg->loaded * 1000 - stat->prg->elapsed);*/
-			add_time_to_str(&m, &l, (stat->prg->size - stat->prg->pos) / ((longlong)stat->prg->loaded * 10 / (stat->prg->elapsed / 100)));
+			/*add_time_to_str(&m, &l, (stat->prg->size - stat->prg->pos) / ((longlong)stat->prg->loaded * 10 / (stat->prg->elapsed / 100)));*/
+			add_time_to_str(&m, &l, (stat->prg->size - stat->prg->pos) / ((double)stat->prg->loaded * 1000 / stat->prg->elapsed));
 		}
 	} else m = stracpy(_(get_err_msg(stat->state), term));
+	show_percentage = t && test_percentage(stat);
 	u = stracpy(down->url);
 	if (strchr(u, POST_CHAR)) *strchr(u, POST_CHAR) = 0;
 	max_text_width(term, u, &max);
@@ -415,7 +458,7 @@ void download_window_function(struct dialog_data *dlg)
 	w = dlg->win->term->x * 9 / 10 - 2 * DIALOG_LB;
 	if (w < min) w = min;
 	if (w > dlg->win->term->x - 2 * DIALOG_LB) w = dlg->win->term->x - 2 * DIALOG_LB;
-	if (t && stat->prg->size >= 0) {
+	if (show_percentage) {
 		if (w < DOWN_DLG_MIN) w = DOWN_DLG_MIN;
 	} else {
 		if (w > max) w = max;
@@ -424,7 +467,7 @@ void download_window_function(struct dialog_data *dlg)
 	y = 0;
 	dlg_format_text(NULL, term, u, 0, &y, w, NULL, COLOR_DIALOG_TEXT, AL_LEFT);
 	y++;
-	if (t && stat->prg->size >= 0) y += 2;
+	if (show_percentage) y += 2;
 	dlg_format_text(NULL, term, m, 0, &y, w, NULL, COLOR_DIALOG_TEXT, AL_LEFT);
 	y++;
 	dlg_format_buttons(NULL, term, dlg->items, dlg->n, 0, &y, w, NULL, AL_CENTER);
@@ -435,15 +478,16 @@ void download_window_function(struct dialog_data *dlg)
 	y = dlg->y + DIALOG_TB + 1;
 	x = dlg->x + DIALOG_LB;
 	dlg_format_text(term, term, u, x, &y, w, NULL, COLOR_DIALOG_TEXT, AL_LEFT);
-	if (t && stat->prg->size >= 0) {
-		unsigned char q[64];
+	if (show_percentage) {
+		unsigned char *q;
 		int p = w - 6;
 		y++;
 		set_only_char(term, x, y, '[');
 		set_only_char(term, x + w - 5, y, ']');
-		fill_area(term, x + 1, y, (int)((longlong)p * (longlong)stat->prg->pos / (longlong)stat->prg->size), 1, COLOR_DIALOG_METER);
-		sprintf(q, "%3d%%", (int)((longlong)100 * (longlong)stat->prg->pos / (longlong)stat->prg->size));
+		fill_area(term, x + 1, y, download_meter(p, stat), 1, COLOR_DIALOG_METER | ' ');
+		q = download_percentage(down, 1);
 		print_text(term, x + w - 4, y, strlen(q), q, COLOR_DIALOG_TEXT);
+		mem_free(q);
 		y++;
 	}
 	y++;
@@ -484,6 +528,7 @@ void display_download(struct terminal *term, struct download *down, struct sessi
 
 time_t parse_http_date(const char *date)	/* this functions is bad !!! */
 {
+#ifdef HAVE_MKTIME
 	const char *months[12] =
 		{"Jan", "Feb", "Mar", "Apr", "May", "Jun",
 		 "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};
@@ -508,12 +553,16 @@ time_t parse_http_date(const char *date)	/* this functions is bad !!! */
 	t = mktime(&tm);
 	if (t == (time_t) - 1) return 0;
 	else return t;
+#else
+	return 0;
+#endif
 }
 
 void download_data(struct status *stat, struct download *down)
 {
 	struct cache_entry *ce;
 	struct fragment *frag;
+	int rs;
 	if (stat->state >= S_WAIT && stat->state < S_TRANS) goto end_store;
 	if (!(ce = stat->ce)) goto end_store;
 	if (ce->last_modified)
@@ -546,24 +595,24 @@ void download_data(struct status *stat, struct download *down)
 		}*/
 	}
 	x:
-	foreach(frag, ce->frag) while (frag->offset <= down->last_pos && frag->offset + frag->length > down->last_pos) {
+	foreachback(frag, ce->frag) if (frag->offset <= down->last_pos) goto have_frag;
+	foreach(frag, ce->frag) have_frag: while (frag->offset <= down->last_pos && frag->offset + frag->length > down->last_pos) {
 		int w;
 #ifdef HAVE_OPEN_PREALLOC
 		if (!down->last_pos && (!down->stat.prg || down->stat.prg->size > 0) && can_prealloc(down->file)) {
 			struct stat st;
-			if (fstat(down->handle, &st) || !S_ISREG(st.st_mode)) goto skip_prealloc;
-			close(down->handle);
-			down->handle = open_prealloc(down->file, O_CREAT|O_WRONLY|O_TRUNC, 0666, down->stat.prg ? down->stat.prg->size : ce->length);
-			if (down->handle == -1) goto error;
+			EINTRLOOP(rs, fstat(down->handle, &st));
+			if (rs || !S_ISREG(st.st_mode)) goto skip_prealloc;
+			EINTRLOOP(rs, close(down->handle));
+			down->handle = open_prealloc(down->file, O_CREAT|O_NOCTTY|O_WRONLY|O_TRUNC, 0666, down->stat.prg ? down->stat.prg->size : ce->length);
+			if (down->handle == -1) goto download_error;
 			set_bin(down->handle);
 			skip_prealloc:;
 		}
 #endif
-		w = write(down->handle, frag->data + (down->last_pos - frag->offset), frag->length - (down->last_pos - frag->offset));
+		w = hard_write(down->handle, frag->data + (down->last_pos - frag->offset), frag->length - (down->last_pos - frag->offset));
 		if (w == -1) {
-#ifdef HAVE_OPEN_PREALLOC
-			error:
-#endif
+			download_error:
 			if (!list_empty(sessions)) {
 				unsigned char *emsg = stracpy(strerror(errno));
 				unsigned char *msg = stracpy(down->file);
@@ -578,23 +627,30 @@ void download_data(struct status *stat, struct download *down)
 	detach_connection(stat, down->last_pos);
 	end_store:;
 	if (stat->state < 0) {
-		if (stat->state != S_OK) {
+		if (stat->state != S__OK) {
 			unsigned char *t = get_err_msg(stat->state);
-			if (t) {
-				unsigned char *tt = stracpy(down->url);
-				if (strchr(tt, POST_CHAR)) *strchr(tt, POST_CHAR) = 0;
-				msg_box(get_download_ses(down)->term, getml(tt, NULL), TEXT_(T_DOWNLOAD_ERROR), AL_CENTER | AL_EXTD_TEXT, TEXT_(T_ERROR_DOWNLOADING), " ", tt, ":\n\n", t, NULL, get_download_ses(down), 1, TEXT_(T_CANCEL), NULL, B_ENTER | B_ESC/*, TEXT_(T_RETRY), NULL, 0 !!! FIXME: retry */);
-			}
+			unsigned char *tt = stracpy(down->url);
+			if (strchr(tt, POST_CHAR)) *strchr(tt, POST_CHAR) = 0;
+			msg_box(get_download_ses(down)->term, getml(tt, NULL), TEXT_(T_DOWNLOAD_ERROR), AL_CENTER | AL_EXTD_TEXT, TEXT_(T_ERROR_DOWNLOADING), " ", tt, ":\n\n", t, NULL, get_download_ses(down), 1, TEXT_(T_CANCEL), NULL, B_ENTER | B_ESC/*, TEXT_(T_RETRY), NULL, 0 !!! FIXME: retry */);
 		} else {
+			prealloc_truncate(down->handle, down->last_pos);
+			EINTRLOOP(rs, close(down->handle));
+			down->handle = -1;
+			if (rs) goto download_error;
 			if (down->prog) {
-				prealloc_truncate(down->handle, down->last_pos);
-				close(down->handle), down->handle = -1;
 				exec_on_terminal(get_download_ses(down)->term, down->prog, down->file, !!down->prog_flags);
 				mem_free(down->prog), down->prog = NULL;
 			} else if (down->remotetime && download_utime) {
+#ifdef HAVE_UTIMES
+				struct timeval utv[2];
+				utv[0].tv_usec = utv[1].tv_usec = 0;
+				utv[0].tv_sec = utv[1].tv_sec = down->remotetime;
+				EINTRLOOP(rs, utimes(down->file, utv));
+#elif defined(HAVE_UTIME)
 				struct utimbuf foo;
 				foo.actime = foo.modtime = down->remotetime;
-				utime(down->file, &foo);
+				EINTRLOOP(rs, utime(down->file, &foo));
+#endif
 			}
 		}
 		abort_download(down);
@@ -632,7 +688,7 @@ int create_download_file(struct terminal *term, unsigned char *fi, int safe)
 			}
 		}
 	}
-	h = open(file, O_CREAT|O_WRONLY|O_TRUNC|(safe?O_EXCL:0), sf ? 0600 : 0666);
+	EINTRLOOP(h, open(file, O_CREAT|O_NOCTTY|O_WRONLY|O_TRUNC|(safe?O_EXCL:0), sf ? 0600 : 0666));
 	if (h == -1) {
 		unsigned char *msg, *msge;
 		int errn = errno;
@@ -672,7 +728,7 @@ unsigned char *get_temp_name(unsigned char *url)
 #endif
 	nm = tempnam(directory, "links");
 	if (!nm) return NULL;
-#ifdef OS2
+#ifdef DOS_FS_8_3
 	if (strlen(nm) > 4 && !strcasecmp(nm + strlen(nm) - 4, ".tmp")) nm[strlen(nm) - 4] = 0;
 #endif
 	name = init_str();
@@ -974,7 +1030,7 @@ int do_move(struct session *ses, struct status **stat)
 		internal("no ses->loading_url");
 		return 0;
 	}
-	if (!(ce = (*stat)->ce) || (ses->wtd == WTD_IMGMAP && (*stat)->state >= 0)) {
+	if (!(ce = (*stat)->ce) || ((*stat)->state >= S_WAIT && (*stat)->state < S_TRANS) || (ses->wtd == WTD_IMGMAP && (*stat)->state >= 0)) {
 		return 0;
 	}
 	if (ce->redirect && ses->redirect_cnt++ < MAX_REDIRECTS) {
@@ -1095,7 +1151,7 @@ void display_timer(struct session *ses)
 	ttime t = get_time();
 	html_interpret(ses);
 	draw_formatted(ses);
-	t = (get_time() - t) * DISPLAY_TIME;
+	t = ((uttime)get_time() - (uttime)t) * DISPLAY_TIME;
 	if (t < DISPLAY_TIME_MIN) t = DISPLAY_TIME_MIN;
 	ses->display_timer = install_timer(t, (void (*)(void *))display_timer, ses);
 	load_frames(ses, ses->screen);
@@ -1121,7 +1177,7 @@ void end_load(struct status *stat, struct session *ses)
 		}
 		if (d == 1) doc_end_load(stat, ses);
 	}
-	if (stat->state < 0 && stat->state != S_OK && d != 2 && d != 1) {
+	if (stat->state < 0 && stat->state != S__OK && d != 2 && d != 1) {
 		print_error_dialog(ses, stat, TEXT_(T_ERROR));
 		if (!d) reload(ses, NC_CACHE);
 	}
@@ -1136,7 +1192,10 @@ void doc_end_load(struct status *stat, struct session *ses)
 		draw_formatted(ses);
 		load_frames(ses, ses->screen);
 		process_file_requests(ses);
-		if (stat->state != S_OK) print_error_dialog(ses, stat, TEXT_(T_ERROR));
+		if (stat->state != S__OK) print_error_dialog(ses, stat, TEXT_(T_ERROR));
+#ifdef LINKS_TESTMODE_AUTO_EXIT
+		terminate_loop = 1;
+#endif
 	} else if (ses->display_timer == -1) display_timer(ses);
 	print_screen_status(ses);
 }
@@ -1319,7 +1378,7 @@ int read_session_info(int fd, struct session *ses, void *data, int len)
 	}
 	if (sz) {
 		char *u, *uu;
-		if (len < 2 * (int)sizeof(int) + sz) return 0;
+		if (len < 2 * (int)sizeof(int) + sz) return -1;
 		if ((unsigned)sz >= MAXINT) overalloc();
 		u = mem_alloc(sz + 1);
 		memcpy(u, (int *)data + 2, sz);
@@ -1549,7 +1608,7 @@ void win_func(struct window *win, struct event *ev, int fw)
 				register_bottom_half((void (*)(void *))destroy_terminal, win->term);
 				return;
 			}
-			/*make_request(ses, 0);*/
+			/*-fallthrough*/
 		case EV_RESIZE:
 			html_interpret(ses);
 			draw_formatted(ses);
