@@ -10,9 +10,12 @@
 #define LEAK_DEBUG_LIST
 #endif
 
-#define LINKS_MANUAL_URL		"http://links.sourceforge.net/docs/manual-0.90-en/"
+#define LINKS_MANUAL_URL		(!strcmp(language_name(current_language),"Czech") || !strcmp(language_name(current_language),"Slovak")?\
+					"http://links.twibright.com/user.html":\
+					"http://links.twibright.com/user_en.html")
 
 #define LINKS_SOCK_NAME			"socket"
+#define LINKS_ANONYMOUS_SOCK_SUFFIX	"-a"
 #define LINKS_PORT			23456
 #define MAX_BIND_TRIES			3
 
@@ -38,7 +41,7 @@
 
 #define COL(x)				((x)*0x100)
 
-#define COLOR_MENU			COL(070)
+#define COLOR_MENU_TEXT			COL(070)
 #define COLOR_MENU_FRAME		COL(070)
 #define COLOR_MENU_SELECTED		COL(007)
 #define COLOR_MENU_HOTKEY		COL(007)
@@ -89,6 +92,7 @@
 #define HTML_FRAME_CHAR_HEIGHT		16
 #define HTML_TABLE_2ND_PASS
 #define HTML_DEFAULT_INPUT_SIZE		20
+#define HTML_MINIMAL_TEXTAREA_WIDTH	6
 #define HTML_DEFAULT_TEXTAREA_WIDTH	40
 #define HTML_DEFAULT_TEXTAREA_HEIGHT	7
 
