@@ -2,12 +2,15 @@
 
 int hard_write(int fd, unsigned char *p, int l)
 {
-	int w = 1;
 	int t = 0;
-	while (l > 0 && w) {
-		if ((w = write(fd, p, l)) < 0) {
-			if (errno == EINTR) continue;
+	while (l > 0) {
+		int w;
+		EINTRLOOP(w, write(fd, p, l));
+		if (w < 0)
 			return -1;
+		if (!w) {
+			errno = ENOSPC;
+			break;
 		}
 		t += w;
 		p += w;
@@ -21,11 +24,9 @@ int hard_read(int fd, unsigned char *p, int l)
 	int r = 1;
 	int t = 0;
 	while (l > 0 && r) {
-		if ((r = read(fd, p, l)) < 0) {
-			if (errno == EINTR) continue;
+		EINTRLOOP(r, read(fd, p, l));
+		if (r < 0)
 			return -1;
-		}
-		/*{int ww;for(ww=0;ww<r;ww++)fprintf(stderr," %02x",(int)p[ww]);fflush(stderr);}*/
 		t += r;
 		p += r;
 		l -= r;
@@ -37,11 +38,12 @@ unsigned char *get_cwd()
 {
 	int bufsize = 128;
 	unsigned char *buf;
+	unsigned char *gcr;
 	while (1) {
 		buf = mem_alloc(bufsize);
-		if (getcwd(buf, bufsize)) return buf;
+		ENULLLOOP(gcr, getcwd(buf, bufsize));
+		if (gcr) return buf;
 		mem_free(buf);
-		if (errno == EINTR) continue;
 		if (errno != ERANGE) return NULL;
 		if ((unsigned)bufsize > MAXINT - 128) overalloc();
 		bufsize += 128;
@@ -51,7 +53,9 @@ unsigned char *get_cwd()
 
 void set_cwd(unsigned char *path)
 {
-	if (path) while (chdir(path) && errno == EINTR) ;
+	int rs;
+	if (path)
+		EINTRLOOP(rs, chdir(path));
 }
 
 struct list_head terminals = {&terminals, &terminals};
@@ -59,6 +63,8 @@ struct list_head terminals = {&terminals, &terminals};
 void alloc_term_screen(struct terminal *term, int x, int y)
 {
 	unsigned *s, *t;
+	if (x < 0) x = 0;
+	if (y < 0) y = 0;
 	if (x && (unsigned)x * (unsigned)y / (unsigned)x != (unsigned)y) overalloc();
 	if ((unsigned)x * (unsigned)y > MAXINT / sizeof(unsigned)) overalloc();
 	s = mem_realloc(term->screen, x * y * sizeof(unsigned));
@@ -78,7 +84,7 @@ void check_if_no_terminal();
 
 void clear_terminal(struct terminal *term)
 {
-	fill_area(term, 0, 0, term->x, term->y, ' ');
+	fill_area(term, 0, 0, term->x, term->y, ' ' | COL(007));
 	set_cursor(term, 0, 0, 0, 0);
 }
 
@@ -344,7 +350,8 @@ void in_term(struct terminal *term)
 	if ((unsigned)term->qlen + ALLOC_GR > MAXINT) overalloc();
 	iq = mem_realloc(term->input_queue, term->qlen + ALLOC_GR);
 	term->input_queue = iq;
-	if ((r = read(term->fdin, iq + term->qlen, ALLOC_GR)) <= 0) {
+	EINTRLOOP(r, read(term->fdin, iq + term->qlen, ALLOC_GR));
+	if (r <= 0) {
 		if (r == -1 && errno != ECONNRESET) error("ERROR: error %d on terminal: could not read event", errno);
 		destroy_terminal(term);
 		return;
@@ -355,7 +362,7 @@ void in_term(struct terminal *term)
 	ev = (struct event *)iq;
 	r = sizeof(struct event);
 	if (ev->ev != EV_INIT && ev->ev != EV_RESIZE && ev->ev != EV_REDRAW && ev->ev != EV_KBD && ev->ev != EV_MOUSE && ev->ev != EV_ABORT) {
-		error("ERROR: error on terminal: bad event %d", ev->ev);
+		error("ERROR: error on terminal: bad event %d", (int)ev->ev);
 		goto mm;
 	}
 	if (ev->ev == EV_INIT) {
@@ -368,7 +375,7 @@ void in_term(struct terminal *term)
 		memcpy(term->cwd, iq + sizeof(struct event) + MAX_TERM_LEN, MAX_CWD_LEN);
 		term->cwd[MAX_CWD_LEN - 1] = 0;
 		term->environment = *(int *)(iq + sizeof(struct event) + MAX_TERM_LEN + MAX_CWD_LEN);
-		ev->b = (long)(iq + sizeof(struct event) + MAX_TERM_LEN + MAX_CWD_LEN + sizeof(int));
+		ev->b = (my_intptr_t)(iq + sizeof(struct event) + MAX_TERM_LEN + MAX_CWD_LEN + sizeof(int));
 		r = sizeof(struct event) + MAX_TERM_LEN + MAX_CWD_LEN + 2 * sizeof(int) + init_len;
 		sync_term_specs();
 	}
@@ -480,7 +487,7 @@ unsigned char frame_restrict[48] = {
 		if (attrib & 0100) add_to_str(&a, &l, ";1");				\
 		add_to_str(&a, &l, "m");						\
 	}										\
-	if (c >= ' ' && c != 127/* && c != 155*/) add_chr_to_str(&a, &l, c);		\
+	if (c >= ' ' && c != 127 && (c != 155 || cp2u(155, term->spec->charset) != -1)) add_chr_to_str(&a, &l, c);\
 	else if (!c || c == 1) add_chr_to_str(&a, &l, ' ');				\
 	else add_chr_to_str(&a, &l, '.');						\
 	cx++;										\
@@ -504,11 +511,16 @@ void redraw_screen(struct terminal *term)
 	if (!term->dirty || (term->master && is_blocked())) return;
 	a = init_str();
 	s = term->spec;
-	for (y = 0; y < term->y; y++)
+	for (y = 0; y < term->y; y++) {
+		if (!memcmp(&term->screen[p], &term->last_screen[p], sizeof(unsigned) * term->x)) {
+			p += term->x;
+			continue;
+		}
 		for (x = 0; x < term->x; x++, p++) {
 			if (y == term->y - 1 && x == term->x - 1) break;
 			if (term->screen[p] == term->last_screen[p]) continue;
-			if ((term->screen[p] & 0x3800) == (term->last_screen[p] & 0x3800) && ((term->screen[p] & 0xff) == 0 || (term->screen[p] & 0xff) == 1 || (term->screen[p] & 0xff) == ' ') && ((term->last_screen[p] & 0xff) == 0 || (term->last_screen[p] & 0xff) == 1 || (term->last_screen[p] & 0xff) == ' ') && (x != term->cx || y != term->cy)) continue;
+			/*if ((term->screen[p] & 0x3800) == (term->last_screen[p] & 0x3800) && ((term->screen[p] & 0xff) == 0 || (term->screen[p] & 0xff) == 1 || (term->screen[p] & 0xff) == ' ') && ((term->last_screen[p] & 0xff) == 0 || (term->last_screen[p] & 0xff) == 1 || (term->last_screen[p] & 0xff) == ' ') && (x != term->cx || y != term->cy)) continue;*/
+			/*fprintf(stderr, "%d.%d : %04x -> %04x\n", x, y, term->last_screen[p] & 0xffff, term->screen[p] & 0xffff);*/
 			term->last_screen[p] = term->screen[p];
 			if (cx == x && cy == y) goto pc;/*PRINT_CHAR(p)*/
 			else if (cy == y && x - cx < 10 && x - cx > 0) {
@@ -525,6 +537,7 @@ void redraw_screen(struct terminal *term)
 				PRINT_CHAR(p);
 			}
 		}
+	}
 	if (l) {
 		if (s->col) add_to_str(&a, &l, "\033[37;40m");
 		add_to_str(&a, &l, "\033[0m");
@@ -551,6 +564,8 @@ void redraw_screen(struct terminal *term)
 
 void destroy_terminal(struct terminal *term)
 {
+	int rs;
+	unregister_bottom_half((void (*)(void *))destroy_terminal, term);
 	while ((term->windows.next) != &term->windows) delete_window(term->windows.next);
 	/*if (term->cwd) mem_free(term->cwd);*/
 	if (term->title) mem_free(term->title);
@@ -559,19 +574,22 @@ void destroy_terminal(struct terminal *term)
 	set_handlers(term->fdin, NULL, NULL, NULL, NULL);
 	mem_free(term->input_queue);
 	if (term->blocked != -1) {
-		close(term->blocked);
+		EINTRLOOP(rs, close(term->blocked));
 		set_handlers(term->blocked, NULL, NULL, NULL, NULL);
 	}
 	del_from_list(term);
-	close(term->fdin);
-	if (term->fdout != 1) {
-		if (term->fdout != term->fdin) close(term->fdout);
+	EINTRLOOP(rs, close(term->fdin));
+	if (!term->master) {
+		if (term->fdout != term->fdin)
+			EINTRLOOP(rs, close(term->fdout));
 	} else {
 		unhandle_terminal_signals(term);
 		free_all_itrms();
 #ifndef NO_FORK_ON_EXIT
 		if (!list_empty(terminals)) {
-			if (fork() > 0) _exit(0);
+			pid_t rp;
+			EINTRLOOP(rp, fork());
+			if (rp > 0) _exit(0);
 		}
 #endif
 	}
@@ -588,7 +606,7 @@ void destroy_all_terminals()
 void check_if_no_terminal()
 {
 	if (!list_empty(terminals)) return;
-	terminate = 1;
+	terminate_loop = 1;
 }
 
 void set_char(struct terminal *t, int x, int y, unsigned c)
@@ -599,6 +617,7 @@ void set_char(struct terminal *t, int x, int y, unsigned c)
 
 unsigned get_char(struct terminal *t, int x, int y)
 {
+	if (!t->x || !t->y) return 070 * 256 + ' ';
 	if (x >= t->x) x = t->x - 1;
 	if (x < 0) x = 0;
 	if (y >= t->y) y = t->y - 1;
@@ -636,11 +655,23 @@ void set_line_color(struct terminal *t, int x, int y, int l, unsigned c)
 
 void fill_area(struct terminal *t, int x, int y, int xw, int yw, unsigned c)
 {
-	int i,j;
+	int i;
+	unsigned *p, *ps;
+	if (x < 0) xw += x, x = 0;
+	if (x + xw > t->x) xw = t->x - x;
+	if (xw <= 0) return;
+	if (y < 0) yw += y, y = 0;
+	if (y + yw > t->y) yw = t->y - y;
+	if (yw <= 0) return;
 	t->dirty = 1;
-	for (j = y >= 0 ? 0 : -y; j < yw && y+j < t->y; j++)
-		for (i = x >= 0 ? 0 : -x; i < xw && x+i < t->x; i++) 
-			t->screen[x+i + t->x*(y+j)] = c;
+	p = ps = &t->screen[x + t->x * y];
+	for (i = 0; i < xw; i++)
+		*p++ = c;
+	p = ps;
+	for (i = 1; i < yw; i++) {
+		p += t->x;
+		memcpy(p, ps, xw * sizeof(unsigned));
+	}
 }
 
 int p1[] = { 218, 191, 192, 217, 179, 196 };
@@ -679,32 +710,39 @@ void set_cursor(struct terminal *term, int x, int y, int altx, int alty)
 
 void exec_thread(unsigned char *path, int p)
 {
-#if defined(HAVE_SETPGID) && !defined(BEOS) && !defined(HAVE_BEGINTHREAD)
-	if (path[0] == 2) setpgid(0, 0);
+	int rs;
+#if defined(HAVE_SETPGID) && !defined(EXEC_IN_THREADS)
+	if (path[0] == 2)
+		EINTRLOOP(rs, setpgid(0, 0));
 #endif
 	exe(path + 1);
-	/*close(p);*/
-	if (path[1 + strlen(path + 1) + 1]) unlink(path + 1 + strlen(path + 1) + 1);
+	if (path[1 + strlen(path + 1) + 1])
+		EINTRLOOP(rs, unlink(path + 1 + strlen(path + 1) + 1));
 }
 
 void close_handle(void *p)
 {
-	int h = (int)p;
-	close(h);
+	int h = (int)(my_uintptr_t)p;
+	int rs;
+	EINTRLOOP(rs, close(h));
 	set_handlers(h, NULL, NULL, NULL, NULL);
 }
 
 void unblock_terminal(struct terminal *term)
 {
-	close_handle((void *)term->blocked);
+	close_handle((void *)(my_uintptr_t)term->blocked);
 	term->blocked = -1;
 	set_handlers(term->fdin, (void (*)(void *))in_term, NULL, (void (*)(void *))destroy_terminal, term);
 	unblock_itrm(term->fdin);
-	redraw_terminal_cls(term);
+	/* clear the dirty flag because unblock_itrm queued a resize
+	   event - so avoid double redraw */
+	term->dirty = 0;
+	/*redraw_terminal_cls(term);*/
 }
 
 void exec_on_terminal(struct terminal *term, unsigned char *path, unsigned char *delete, int fg)
 {
+	int rs;
 	if (path && !*path) return;
 	if (!path) path="";
 #ifdef NO_FG_EXEC
@@ -716,7 +754,8 @@ void exec_on_terminal(struct terminal *term, unsigned char *path, unsigned char
 			int blockh;
 			unsigned char *param;
 			if (is_blocked() && fg) {
-				unlink(delete);
+				if (*delete)
+					EINTRLOOP(rs, unlink(delete));
 				return;
 			}
 			param = mem_alloc(strlen(path) + strlen(delete) + 3);
@@ -736,7 +775,7 @@ void exec_on_terminal(struct terminal *term, unsigned char *path, unsigned char
 				set_handlers(term->fdin, NULL, NULL, (void (*)(void *))destroy_terminal, term);
 				/*block_itrm(term->fdin);*/
 			} else {
-				set_handlers(blockh, close_handle, NULL, close_handle, (void *)blockh);
+				set_handlers(blockh, close_handle, NULL, close_handle, (void *)(my_uintptr_t)blockh);
 			}
 		}
 	} else {
@@ -769,10 +808,7 @@ void do_terminal_function(struct terminal *term, unsigned char code, unsigned ch
 
 void set_terminal_title(struct terminal *term, unsigned char *title)
 {
-	int i;
-	for (i = 0; i < 10000; i++) if (!title[i]) goto s;
-	title[10000] = 0;
-	s:
+	if (strlen(cast_const_char title) > 10000) title[10000] = 0;
 	if (strchr(title, 1)) {
 		unsigned char *a, *b;
 		for (a = title, b = title; *a; a++) if (*a != 1) *b++ = *a;
