@@ -7,9 +7,9 @@ struct list_head mms_prog = { &mms_prog, &mms_prog };
 
 struct list_head assoc = { &assoc, &assoc };
 
-tcount get_assoc_cnt()
+unsigned get_assoc_cnt()
 {
-	static tcount assoc_cnt = 0;
+	static unsigned assoc_cnt = 0;
 	if (!++assoc_cnt) assoc_cnt = 1;
 	return assoc_cnt;
 }
@@ -55,12 +55,15 @@ unsigned char *get_content_type(unsigned char *head, unsigned char *url)
 	struct assoc *a;
 	unsigned char *ct, *ext, *exxt;
 	int extl, el;
+	int code, vers;
 	if (head && (ct = parse_http_header(head, "Content-Type", NULL))) {
 		unsigned char *s;
 		if ((s = strchr(ct, ';'))) *s = 0;
 		while (*ct && ct[strlen(ct) - 1] <= ' ') ct[strlen(ct) - 1] = 0;
 		return ct;
 	}
+	if (!get_http_code(head, &code, &vers) && code >= 300)
+		return stracpy("text/html");
 	ext = NULL, extl = 0;
 	for (ct = url; *ct && !end_of_dir(*ct); ct++)
 		if (*ct == '.') ext = ct + 1;
@@ -217,7 +220,7 @@ void update_assoc(struct assoc *new)
 
 void really_del_ct(void *fcp)
 {
-	tcount fc = (int)fcp;
+	unsigned fc = (unsigned)(my_uintptr_t)fcp;
 	struct assoc *del;
 	foreach(del, assoc) if (del->cnt == fc) goto ok;
 	return;
@@ -229,7 +232,7 @@ void menu_del_ct(struct terminal *term, void *fcp, void *xxx2)
 {
 	unsigned char *str;
 	int l;
-	tcount fc = (int)fcp;
+	unsigned fc = (unsigned)(my_uintptr_t)fcp;
 	struct assoc *del;
 	foreach(del, assoc) if (del->cnt == fc) goto ok;
 	return;
@@ -244,7 +247,7 @@ void menu_del_ct(struct terminal *term, void *fcp, void *xxx2)
 void menu_add_ct(struct terminal *term, void *fcp, void *xxx2)
 {
 	int p;
-	tcount fc = (int)fcp;
+	unsigned fc = (unsigned)(my_uintptr_t)fcp;
 	struct assoc *new, *from;
 	unsigned char *label;
 	unsigned char *ct;
@@ -334,7 +337,7 @@ void menu_list_assoc(struct terminal *term, void *fn, void *xxx)
 	int n = 0;
 	foreachback(a, assoc) if (a->system == SYSTEM_ID) {
 		if (!mi && !(mi = new_menu(7))) return;
-		add_to_menu(&mi, stracpy(a->label), stracpy(a->ct), "", MENU_FUNC fn, (void *)a->cnt, 0), n++;
+		add_to_menu(&mi, stracpy(a->label), stracpy(a->ct), "", MENU_FUNC fn, (void *)(my_uintptr_t)a->cnt, 0), n++;
 	}
 	if (!mi) do_menu(term, mi_no_assoc, xxx);
 	else do_menu(term, mi, xxx);
@@ -411,7 +414,7 @@ void update_ext(struct extension *new)
 
 void really_del_ext(void *fcp)
 {
-	tcount fc = (int)fcp;
+	unsigned fc = (unsigned)(my_uintptr_t)fcp;
 	struct extension *del;
 	foreach(del, extensions) if (del->cnt == fc) goto ok;
 	return;
@@ -423,7 +426,7 @@ void menu_del_ext(struct terminal *term, void *fcp, void *xxx2)
 {
 	unsigned char *str;
 	int l;
-	tcount fc = (int)fcp;
+	unsigned fc = (unsigned)(my_uintptr_t)fcp;
 	struct extension *del;
 	foreach(del, extensions) if (del->cnt == fc) goto ok;
 	return;
@@ -437,7 +440,7 @@ void menu_del_ext(struct terminal *term, void *fcp, void *xxx2)
 
 void menu_add_ext(struct terminal *term, void *fcp, void *xxx2)
 {
-	tcount fc = (int)fcp;
+	unsigned fc = (unsigned)(my_uintptr_t)fcp;
 	struct extension *new, *from;
 	unsigned char *ext;
 	unsigned char *ct;
@@ -494,7 +497,7 @@ void menu_list_ext(struct terminal *term, void *fn, void *xxx)
 	int n = 0;
 	foreachback(a, extensions) {
 		if (!mi && !(mi = new_menu(7))) return;
-		add_to_menu(&mi, stracpy(a->ext), stracpy(a->ct), "", MENU_FUNC fn, (void *)a->cnt, 0), n++;
+		add_to_menu(&mi, stracpy(a->ext), stracpy(a->ct), "", MENU_FUNC fn, (void *)(my_uintptr_t)a->cnt, 0), n++;
 	}
 	if (!mi) do_menu(term, mi_no_ext, xxx);
 	else do_menu(term, mi, xxx);
