@@ -179,7 +179,8 @@ int get_port(unsigned char *url)
 	if (parse_url(url, NULL, NULL, NULL, NULL, NULL, NULL, NULL, &h, &hl, NULL, NULL, NULL)) return -1;
 	if (h) {
 		n = strtol(h, NULL, 10);
-		if (n && n < MAXINT) return n;
+		if (n > 0 && n < 65536) return n;
+		return -1;
 	}
 	if ((h = get_protocol_name(url))) {
 		int nn = -1;	/* against warning */
@@ -320,6 +321,11 @@ unsigned char *join_urls(unsigned char *base, unsigned char *rel)
 			translate_directories(n);
 			return n;
 		}
+		add_to_strn(&n, cast_uchar "/");
+		if (!parse_url(n, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL)) {
+			translate_directories(n);
+			return n;
+		}
 		mem_free(n);
 	}
 	if (!casecmp("proxy://", rel, 8)) goto prx;
@@ -459,33 +465,61 @@ void get_filename_from_url(unsigned char *url, unsigned char **s, int *l)
 	*l = url - *s;
 }
 
-#define accept_char(x)	((x) != '"' && (x) != '\'' && (x) != '&' && (x) != '<' && (x) != '>')
-#define special_char(x)	((x) == '%' || (x) == '#')
+int url_not_saveable(unsigned char *url)
+{
+	int p, palen;
+	unsigned char *u = translate_url(url, cast_uchar "/");
+	if (!u)
+		return 1;
+	p = parse_url(u, NULL, NULL, NULL, NULL, &palen, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
+	mem_free(u);
+	return p || palen;
+}
+
+#define accept_char(x)	((x) != 10 && (x) != 13 && (x) != '"' && (x) != '\'' && (x) != '&' && (x) != '<' && (x) != '>')
+#define special_char(x)	((x) < ' ' || (x) == '%' || (x) == '#' || (x) >= 127)
+
+/*
+ * -2 percent to raw
+ * -1 percent to html
+ *  0 raw to html
+ *  1 raw to percent
+ */
 
 void add_conv_str(unsigned char **s, int *l, unsigned char *b, int ll, int encode_special)
 {
-	for (; ll; ll--, b++) {
-		if ((unsigned char)*b < ' ') continue;
-		if (special_char(*b) && encode_special == 1) {
+	for (; ll > 0; ll--, b++) {
+		unsigned char chr = *b;
+		if (!chr) continue;
+		if (special_char(chr) && encode_special == 1) {
 			unsigned char h[4];
-			sprintf(h, "%%%02X", (unsigned)*b & 0xff);
+			sprintf(cast_char h, "%%%02X", (unsigned)chr & 0xff);
 			add_to_str(s, l, h);
-		} else if (*b == '%' && encode_special <= -1 && ll > 2 && ((b[1] >= '0' && b[1] <= '9') || (b[1] >= 'A' && b[1] <= 'F') || (b[1] >= 'a' && b[1] <= 'f'))) {
-			unsigned char h = 0;
+			continue;
+		}
+		if (chr == '%' && encode_special <= -1 && ll > 2 &&
+		    ((b[1] >= '0' && b[1] <= '9') || (b[1] >= 'A' && b[1] <= 'F') || (b[1] >= 'a' && b[1] <= 'f')) &&
+		    ((b[2] >= '0' && b[2] <= '9') || (b[2] >= 'A' && b[2] <= 'F') || (b[2] >= 'a' && b[2] <= 'f'))) {
 			int i;
+			chr = 0;
 			for (i = 1; i < 3; i++) {
-				if (b[i] >= '0' && b[i] <= '9') h = h * 16 + b[i] - '0';
-				if (b[i] >= 'A' && b[i] <= 'F') h = h * 16 + b[i] - 'A' + 10;
-				if (b[i] >= 'a' && b[i] <= 'f') h = h * 16 + b[i] - 'a' + 10;
+				if (b[i] >= '0' && b[i] <= '9') chr = chr * 16 + b[i] - '0';
+				if (b[i] >= 'A' && b[i] <= 'F') chr = chr * 16 + b[i] - 'A' + 10;
+				if (b[i] >= 'a' && b[i] <= 'f') chr = chr * 16 + b[i] - 'a' + 10;
 			}
-			if (h >= ' ') add_chr_to_str(s, l, h);
 			ll -= 2;
 			b += 2;
-		} else if (accept_char(*b) || encode_special == -2) {
-			add_chr_to_str(s, l, *b);
+			if (!chr)
+				continue;
+		}
+		if (chr == ' ' && (!encode_special || encode_special == -1)) {
+			add_to_str(s, l, cast_uchar "&nbsp;");
+		} else if (accept_char(chr) || encode_special == -2) {
+			add_chr_to_str(s, l, chr);
+		} else if (chr == 10 || chr == 13) {
 		} else {
-			add_to_str(s, l, "&#");
-			add_num_to_str(s, l, (int)*b);
+			add_to_str(s, l, cast_uchar "&#");
+			add_num_to_str(s, l, (int)chr);
 			add_chr_to_str(s, l, ';');
 		}
 	}
