@@ -546,7 +546,7 @@ struct link *get_last_link(struct f_data_c *f)
 	int i;
 	struct link *l = NULL;
 	for (i = f->vs->view_pos; i < f->vs->view_pos + f->yw; i++)
-		if (i >= 0 && i < f->f_data->y && f->f_data->lines2[i] > l)
+		if (i >= 0 && i < f->f_data->y && f->f_data->lines2[i] && (!l || f->f_data->lines2[i] > l))
 			l = f->f_data->lines2[i];
 	return l;
 }
@@ -798,11 +798,11 @@ void draw_doc(struct terminal *t, struct f_data_c *scr, int active)
 		set_window_ptr(get_root_window(t), xp, yp);
 	}
 	if (!scr->vs) {
-		fill_area(t, xp, yp, xw, yw, scr->f_data->y ? scr->f_data->bg : ' ');
+		fill_area(t, xp, yp, xw, yw, (scr->f_data->y ? scr->f_data->bg : COL(007)) | ' ');
 		return;
 	}
 	if (scr->f_data->frame) {
-	 	fill_area(t, xp, yp, xw, yw, scr->f_data->y ? scr->f_data->bg : ' ');
+	 	fill_area(t, xp, yp, xw, yw, (scr->f_data->y ? scr->f_data->bg : COL(007)) | ' ');
 		draw_frame_lines(t, scr->f_data->frame_desc, xp, yp);
 		if (scr->vs && scr->vs->current_link == -1) scr->vs->current_link = 0, scr->vs->orig_link = scr->vs->current_link;
 		return;
@@ -861,7 +861,7 @@ void draw_doc(struct terminal *t, struct f_data_c *scr, int active)
 	free_link(scr);
 	scr->xl = vx;
 	scr->yl = vy;
-	fill_area(t, xp, yp, xw, yw, scr->f_data->y ? scr->f_data->bg : ' ');
+	fill_area(t, xp, yp, xw, yw, (scr->f_data->y ? scr->f_data->bg : COL(007)) | ' ');
 	if (!scr->f_data->y) return;
 	while (vs->view_pos >= scr->f_data->y) vs->view_pos -= yw ? yw : 1;
 	if (vs->view_pos < 0) vs->view_pos = 0;
@@ -905,7 +905,7 @@ void draw_formatted(struct session *ses)
 {
 	if (!ses->screen || !ses->screen->f_data) {
 		/*internal("document not formatted");*/
-		fill_area(ses->term, 0, 1, ses->term->x, ses->term->y - 2, ' ');
+		fill_area(ses->term, 0, 1, ses->term->x, ses->term->y - 2, COL(007) | ' ');
 		return;
 	}
 	if (!ses->screen->vs && !list_empty(ses->history))
@@ -966,7 +966,7 @@ int dump_to_file(struct f_data *fd, int h)
 				if (fc->type == FC_RESET) add_to_str(&s, &l, "Reset form");
 				else if (!fc->action) add_to_str(&s, &l, "Button");
 				else {
-					if (!fc->method == FM_GET) add_to_str(&s, &l, "Submit form: ");
+					if (fc->method == FM_GET) add_to_str(&s, &l, "Submit form: ");
 					else add_to_str(&s, &l, "Post form: ");
 					add_to_str(&s, &l, fc->action);
 				}
@@ -1396,6 +1396,7 @@ void encode_multipart(struct session *ses, struct list_head *l, unsigned char **
 	int i, j;
 	int flg = 0;
 	unsigned char *p;
+	int rs;
 	struct conv_table *convert_table = get_translation_table(cp_from, cp_to);
 	memset(bound, 'x', BL);
 	*len = 0;
@@ -1450,23 +1451,29 @@ void encode_multipart(struct session *ses, struct list_head *l, unsigned char **
 				goto error;
 			}*/
 			if (*sv->value) {
+				unsigned char *wd;
 				if (anonymous) {
 					goto not_allowed;
 				}
-				if ((fh = open(sv->value, O_RDONLY)) == -1) {
+				wd = get_cwd();
+				set_cwd(ses->term->cwd);
+				EINTRLOOP(fh, open(sv->value, O_RDONLY | O_NOCTTY));
+				if (fh == -1) {
 					errn = errno;
+					if (wd) set_cwd(wd), mem_free(wd);
 					goto error;
 				}
+				if (wd) set_cwd(wd), mem_free(wd);
 				set_bin(fh);
 				do {
-					if ((rd = read(fh, buffer, F_BUFLEN)) == -1) {
+					if ((rd = hard_read(fh, buffer, F_BUFLEN)) == -1) {
 						errn = errno;
-						close(fh);
+						EINTRLOOP(rs, close(fh));
 						goto error;
 					}
 					if (rd) add_bytes_to_str(data, len, buffer, rd);
 				} while (rd);
-				close(fh);
+				EINTRLOOP(rs, close(fh));
 			}
 		}
 		add_to_str(data, len, "\r\n");
@@ -1481,11 +1488,12 @@ void encode_multipart(struct session *ses, struct list_head *l, unsigned char **
 	again:
 	for (i = 0; i <= *len - BL; i++) {
 		for (j = 0; j < BL; j++) if ((*data)[i + j] != bound[j]) goto nb;
-		for (j = BL - 1; j >= 0; j--)
+		for (j = BL - 1; j >= 0; j--) {
 			if (bound[j] < '0') bound[j] = '0' - 1;
 			if (bound[j]++ >= '9') bound[j] = '0';
 			else goto again;
-		internal("Counld not assing boundary");
+		}
+		internal("Could not assign boundary");
 		nb:;
 	}
 	for (i = 0; i < nbound_ptrs; i++) memcpy(*data + bound_ptrs[i], bound, BL);
@@ -1621,8 +1629,8 @@ int enter(struct session *ses, struct f_data_c *f, int a)
 			struct form_control *fc;
 			foreach(fc, f->f_data->forms)
 				if (fc->form_num == link->form->form_num && fc->type == FC_RADIO && !xstrcmp(fc->name, link->form->name)) {
-					struct form_state *ffs = find_form_state(f, fc);
-					if (ffs) ffs->state = 0;
+					struct form_state *fffs = find_form_state(f, fc);
+					if (fffs) fffs->state = 0;
 				}
 			fs = find_form_state(f, link->form);
 			fs->state = 1;
@@ -1651,7 +1659,7 @@ void back(struct session *ses, struct f_data_c *f, int a)
 
 void selected_item(struct terminal *term, void *pitem, struct session *ses)
 {
-	int item = (int)pitem;
+	int item = (int)(my_uintptr_t)pitem;
 	struct f_data_c *f = current_frame(ses);
 	struct link *l;
 	struct form_state *fs;
@@ -1809,8 +1817,12 @@ int field_op(struct session *ses, struct f_data_c *f, struct link *l, struct eve
 				fs->state = 0;
 				break;
 			case ACT_PASTE_CLIPBOARD: {
-				char *clipboard = get_clipboard_text();
+				unsigned char *clipboard = get_clipboard_text();
 				if (!clipboard) break;
+				if (form->type != FC_TEXTAREA) {
+					unsigned char *nl = clipboard;
+					while ((nl = strchr(nl, '\n'))) *nl = ' ';
+				}
 				if (!form->ro && strlen(clipboard) <= (size_t)form->maxlength) {
 					unsigned char *v;
 					v = mem_realloc(fs->value, strlen(clipboard) +1);
@@ -2009,21 +2021,7 @@ void rep_ev(struct session *ses, struct f_data_c *fd, void (*f)(struct session *
 
 struct link *choose_mouse_link(struct f_data_c *f, struct event *ev)
 {
-	struct link *l1 = f->f_data->links + f->f_data->nlinks;
-	struct link *l2 = f->f_data->links;
-	struct link *l;
-	int i;
-	if (!f->f_data->nlinks) return NULL;
-	if (ev->x < 0 || ev->y < 0 || ev->x >= f->xw || ev->y >= f->yw) return NULL;
-	for (i = f->vs->view_pos; i < f->f_data->y && i < f->vs->view_pos + f->yw; i++) {
-		if (f->f_data->lines1[i] && f->f_data->lines1[i] < l1) l1 = f->f_data->lines1[i];
-		if (f->f_data->lines2[i] && f->f_data->lines2[i] > l2) l2 = f->f_data->lines2[i];
-	}
-	for (l = l1; l <= l2; l++) {
-		int i;
-		for (i = 0; i < l->n; i++) if (l->pos[i].x - f->vs->view_posx == ev->x && l->pos[i].y - f->vs->view_pos == ev->y) return l;
-	}
-	return NULL;
+	return get_link_at_location(f->f_data, ev->x + f->vs->view_posx, ev->y + f->vs->view_pos);
 }
 
 void goto_link_number(struct session *ses, unsigned char *num)
@@ -2044,7 +2042,7 @@ void goto_link_number(struct session *ses, unsigned char *num)
 
 void frm_download(struct session *, struct f_data_c *);
 void send_image(struct terminal *term, void *xxx, struct session *ses);
-    
+
 int frame_ev(struct session *ses, struct f_data_c *fd, struct event *ev)
 {
 	int x = 1;
@@ -2286,7 +2284,7 @@ void send_event(struct session *ses, struct event *ev)
 				goto x;
 			case ACT_QUIT:
 		  		quit:
-				exit_prog(ses->term, (void *)(ev->x == KBD_CTRL_C), ses);
+				exit_prog(ses->term, (void *)(my_uintptr_t)(ev->x == KBD_CTRL_C), ses);
 				goto x;
 			case ACT_DOCUMENT_INFO:
 				state_msg(ses);
@@ -2389,11 +2387,11 @@ void send_download(struct terminal *term, void *xxx, struct session *ses)
 /* open a link in a new xterm */
 void send_open_in_new_xterm(struct terminal *term, void (*open_window)(struct terminal *term, unsigned char *, unsigned char *), struct session *ses)
 {
-        struct f_data_c *fd = current_frame(ses);
-        if (!fd) return;
-        if (fd->vs->current_link == -1) return;
-        if (ses->dn_url) mem_free(ses->dn_url);
-        if ((ses->dn_url = get_link_url(ses, fd, &fd->f_data->links[fd->vs->current_link]))) {
+	struct f_data_c *fd = current_frame(ses);
+	if (!fd) return;
+	if (fd->vs->current_link == -1) return;
+	if (ses->dn_url) mem_free(ses->dn_url);
+	if ((ses->dn_url = get_link_url(ses, fd, &fd->f_data->links[fd->vs->current_link]))) {
 		unsigned char *enc_url = encode_url(ses->dn_url);
 		unsigned char *path = escape_path(path_to_exe);
 		open_window(term, path, enc_url);
@@ -2406,10 +2404,10 @@ void send_open_new_xterm(struct terminal *term, void (*open_window)(struct termi
 {
 	int l;
 	unsigned char *path;
-        if (ses->dn_url) mem_free(ses->dn_url);
+	if (ses->dn_url) mem_free(ses->dn_url);
 	ses->dn_url = init_str();
 	l = 0;
-        add_to_str(&ses->dn_url, &l, "-base-session ");
+	add_to_str(&ses->dn_url, &l, "-base-session ");
 	add_num_to_str(&ses->dn_url, &l, ses->id);
 	path = escape_path(path_to_exe);
 	open_window(term, path, ses->dn_url);
@@ -2483,11 +2481,12 @@ void save_as(struct terminal *term, void *xxx, struct session *ses)
 void save_formatted(struct session *ses, unsigned char *file)
 {
 	int h;
+	int rs;
 	struct f_data_c *f;
 	if (!(f = current_frame(ses)) || !f->f_data) return;
 	if ((h = create_download_file(ses->term, file, 0)) < 0) return;
 	if (dump_to_file(f->f_data, h)) msg_box(ses->term, NULL, TEXT_(T_SAVE_ERROR), AL_CENTER, TEXT_(T_ERROR_WRITING_TO_FILE), NULL, 1, TEXT_(T_CANCEL), NULL, B_ENTER | B_ESC);
-	close(h);
+	EINTRLOOP(rs, close(h));
 }
 
 void menu_save_formatted(struct terminal *term, void *xxx, struct session *ses)
@@ -2712,7 +2711,8 @@ void loc_msg(struct terminal *term, struct location *lo, struct f_data_c *frame)
 #ifdef HAVE_SSL
 		if (ce->ssl_info) {
 			add_to_str(&s, &l, "\n");
-			add_to_str(&s, &l, "SSL cipher: ");
+			add_to_str(&s, &l, _(TEXT_(T_SSL_CIPHER), term));
+			add_to_str(&s, &l, ": ");
 			add_to_str(&s, &l, ce->ssl_info);
 		}
 #endif
@@ -2751,6 +2751,10 @@ void head_msg(struct session *ses)
 			while ((ss = strstr(s, "\r\n"))) memmove(ss, ss + 1, strlen(ss));
 			while (*s && s[strlen(s) - 1] == '\n') s[strlen(s) - 1] = 0;
 		}
-		msg_box(ses->term, getml(s, NULL), TEXT_(T_HEADER_INFO), AL_LEFT, s, NULL, 1, TEXT_(T_OK), NULL, B_ENTER | B_ESC);
+		if (*s && *s != '\n') {
+			msg_box(ses->term, getml(s, NULL), TEXT_(T_HEADER_INFO), AL_LEFT, s, NULL, 1, TEXT_(T_OK), NULL, B_ENTER | B_ESC);
+		} else {
+			msg_box(ses->term, getml(s, NULL), TEXT_(T_HEADER_INFO), AL_CENTER, TEXT_(T_NO_HEADER), NULL, 1, TEXT_(T_OK), NULL, B_ENTER | B_ESC);
+		}
 	}
 }
